(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
header {* Term Rewrite Systems *}
theory Trs
imports
  Term
  "$AFP/Abstract-Rewriting/Relative_Rewriting"
  "~~/src/HOL/Library/Infinite_Set"
begin

text {* A theory on first-order term rewrite systems (TRSs). *}

text {* A rewrite rule is a pair of terms. A TRS is a set of
rewrite rules. *}
type_synonym ('f, 'v) rule = "('f, 'v) term \<times> ('f, 'v) term"
type_synonym ('f, 'v) trs  = "('f, 'v) rule set"

subsection {* Variables of Rules *}

definition
  vars_rule :: "('f, 'v) rule \<Rightarrow> 'v set"
where
  "vars_rule r = vars_term (fst r) \<union> vars_term (snd r)"

lemma finite_vars_rule:
  "finite (vars_rule r)"
  by (auto simp: vars_rule_def)

subsection {* Variables of TRSs *}

definition vars_trs :: "('f, 'v) trs \<Rightarrow> 'v set" where
  "vars_trs R = (\<Union>r\<in>R. vars_rule r)"

lemma vars_trs_union: "vars_trs (R \<union> S) = vars_trs R \<union> vars_trs S" 
  unfolding vars_trs_def by auto

lemma finite_trs_has_finite_vars:
  assumes "finite R" shows "finite (vars_trs R)"
  using assms unfolding vars_trs_def vars_rule_def [abs_def] by simp

lemmas vars_defs = vars_trs_def vars_rule_def

subsection {* Function Symbols of Rules *}

definition funs_rule :: "('f, 'v) rule \<Rightarrow> 'f set" where
  "funs_rule r = funs_term (fst r) \<union> funs_term (snd r)"


text {* The same including arities. *}
definition funas_rule :: "('f, 'v) rule \<Rightarrow> 'f sig" where
  "funas_rule r = funas_term (fst r) \<union> funas_term (snd r)"


subsection {* Function Symbols of TRSs *}

definition funs_trs :: "('f, 'v) trs \<Rightarrow> 'f set" where
  "funs_trs R = (\<Union>r\<in>R. funs_rule r)"

definition funas_trs :: "('f, 'v) trs \<Rightarrow> 'f sig" where
  "funas_trs R = (\<Union>r\<in>R. funas_rule r)"

lemma finite_funas_rule: "finite (funas_rule lr)"
  unfolding funas_rule_def
  using finite_funas_term by auto

lemma finite_funas_trs: assumes "finite R"
  shows "finite (funas_trs R)"
  unfolding funas_trs_def
  using assms finite_funas_rule by auto

lemma funas_empty[simp]: "funas_trs {} = {}" unfolding funas_trs_def by simp

lemma funas_trs_union[simp]: "funas_trs (R \<union> S) = funas_trs R \<union> funas_trs S"
  unfolding funas_trs_def by blast

definition funas_args_rule :: "('f, 'v) rule \<Rightarrow> 'f sig" where
  "funas_args_rule r = funas_args_term (fst r) \<union> funas_args_term (snd r)"

definition funas_args_trs :: "('f, 'v) trs \<Rightarrow> 'f sig" where
  "funas_args_trs R = (\<Union>r\<in>R. funas_args_rule r)"

lemmas funas_args_defs =
  funas_args_trs_def funas_args_rule_def funas_args_term_def

definition funas_root_rule :: "('f, 'v) rule \<Rightarrow> 'f sig" where
  "funas_root_rule r = Option.set (root (fst r)) \<union> Option.set (root (snd r))"

definition funas_root_trs :: "('f, 'v) trs \<Rightarrow> 'f sig" where
  "funas_root_trs R = (\<Union>r\<in>R. funas_root_rule r)"

lemmas funas_root_defs =
  funas_root_trs_def funas_root_rule_def

definition funas_head :: "('f, 'v) trs \<Rightarrow> ('f, 'v) trs \<Rightarrow> 'f sig" where
  "funas_head P R = funas_trs P - (funas_trs R \<union> funas_args_trs P)"

lemmas funs_defs = funs_trs_def funs_rule_def
lemmas funas_defs =
  funas_trs_def funas_rule_def
  funas_args_defs
  funas_head_def
  funas_root_defs

text {* A function symbol is said to be \emph{defined} (w.r.t.\ to a given
TRS) if it occurs as root of some left-hand side. *}
definition
  defined :: "('f, 'v) trs \<Rightarrow> ('f \<times> nat) \<Rightarrow> bool"
where
  "defined R fn \<longleftrightarrow> (\<exists>l r. (l, r) \<in> R \<and> root l = Some fn)"

text {* The set of all defined symbols. *}
definition
  "defs" :: "('f, 'v) trs \<Rightarrow> ('f \<times> nat) set"
where
  "defs R = {(f, n). defined R (f, n)}"

lemma defined_funas_trs: assumes d: "defined R fn" shows "fn \<in> funas_trs R"
proof -
  from d [unfolded defined_def] obtain l r
    where "(l, r) \<in> R" and "root l = Some fn" by auto
  thus ?thesis
    unfolding funas_trs_def funas_rule_def [abs_def] by (cases l) force+
qed

text {* The remaining symbols are called \emph{constructor symbols}, since
they are only used to construct data. *}
definition cons :: "('f, 'v) trs \<Rightarrow> ('f \<times> nat) set" where
  "cons R = funas_trs R - defs R"

text {* Closure under contexts of a binary relation on terms. *}
inductive_set
  ctxt_closure :: "('f, 'v) trs \<Rightarrow> ('f, 'v) trs" ("CC")
  for R :: "('f, 'v) trs"
where
  ctxt_closureI[intro]: "(s, t) \<in> R \<Longrightarrow> (C\<langle>s\<rangle>,C\<langle>t\<rangle>) \<in> CC R"

lemma subset_ctxt_closure: "R \<subseteq> CC R"
proof
  fix s t assume "(s, t) \<in> R"
  from ctxt_closureI[OF this, of \<box>] show "(s, t) \<in> CC R" by simp
qed

definition
  ctxt_closed :: "('f, 'v) trs \<Rightarrow> bool"
where
  "ctxt_closed R = (CC R \<subseteq> R)"

lemma ctxt_closedI[intro]:
  assumes "\<And>s t C. (s, t) \<in> R \<Longrightarrow> (C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> R"
  shows "ctxt_closed R"
proof -
  have "CC R \<subseteq> R"
  proof
    fix s t assume "(s, t) \<in> CC R" thus "(s, t) \<in> R"
      by (cases) (auto intro: assms)
  qed
  thus ?thesis by (simp add: ctxt_closed_def)
qed

lemma ctxt_closedE[elim]:
  assumes "ctxt_closed R" and "(s, t) \<in> R"
  shows "(C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> R"
  using assms unfolding ctxt_closed_def by auto

lemma ctxt_closed_trancl:
  assumes "ctxt_closed R"
  shows "ctxt_closed (R^+)"
proof (unfold ctxt_closed_def, rule subrelI)
  fix s t
  assume "(s, t) \<in> CC (R\<^sup>+)"
  thus "(s, t) \<in> R\<^sup>+"
  proof (cases rule: ctxt_closure.cases)
    case (ctxt_closureI u v C)
    note IH = this
    from IH(3,1,2) show "(s, t) \<in> R\<^sup>+"
    proof (induct u v arbitrary: t rule: trancl.induct)
      case (r_into_trancl u v) thus ?case
        using assms[unfolded ctxt_closed_def] by auto
    next
      case (trancl_into_trancl u v w)
      hence "(s, C\<langle>v\<rangle>) \<in> R\<^sup>+" by simp
      moreover from trancl_into_trancl(3) and assms
        have "(C\<langle>v\<rangle>, C\<langle>w\<rangle>) \<in> R" unfolding ctxt_closed_def by auto
      ultimately show ?case unfolding trancl_into_trancl by auto
    qed
  qed
qed

lemma ctxt_closed_rel_pow: assumes ctxt: "ctxt_closed R"
  shows "(a,b) \<in> R ^^ n \<Longrightarrow> (C\<langle>a\<rangle>, C\<langle>b\<rangle>) \<in> R ^^ n"
proof (induct n arbitrary: b)
  case (Suc n c)
  from Suc(2) obtain b where ab: "(a,b) \<in> R ^^ n" and bc: "(b,c) \<in> R" 
    by (rule relpow_Suc_E)
  from Suc(1)[OF ab] ctxt_closedE[OF ctxt bc, of C] show ?case by auto
qed auto

lemma ctxt_closed_correctness: "ctxt_closed r \<Longrightarrow> r = CC r"
proof -
  assume "ctxt_closed r"
  hence a: "CC r \<subseteq> r" by (unfold ctxt_closed_def)
  from subset_ctxt_closure have b: "r \<subseteq> CC r" .
  from a b show "r = CC r" by simp
qed

lemma ctxt_closed_imp_ctxt:
 assumes "ctxt_closed R" and "(s,t) \<in> R" shows "(C\<langle>s\<rangle>,C\<langle>t\<rangle>) \<in> R"
proof -
 from `ctxt_closed R` have "R = CC R" by (rule ctxt_closed_correctness)
 from `(s,t) \<in> R` have "(C\<langle>s\<rangle>,C\<langle>t\<rangle>) \<in> CC R" by (rule ctxt_closureI)
 with `R = CC R` show ?thesis by simp
qed

lemma ctxt_closed_R_imp_supt_R_distr:
  assumes "ctxt_closed R" and "s \<rhd> t" and "(t,u) \<in> R" shows "\<exists>t. (s,t) \<in> R \<and> t \<rhd> u"
proof -
  from `s \<rhd> t` obtain C where "C \<noteq> \<box>" and "C\<langle>t\<rangle> = s" by auto
  from `ctxt_closed R` and `(t,u) \<in> R`
    have RCtCu: "(C\<langle>t\<rangle>,C\<langle>u\<rangle>) \<in> R" by (rule ctxt_closed_imp_ctxt)
  from `C \<noteq> \<box>` have "C\<langle>u\<rangle> \<rhd> u" by auto
  from RCtCu have "(s,C\<langle>u\<rangle>) \<in> R" unfolding `C\<langle>t\<rangle> = s` .
  from this and `C\<langle>u\<rangle> \<rhd> u` show ?thesis by auto
qed

lemma ctxt_closed_supt_R_imp_R_supt:
   assumes "ctxt_closed R" and "s \<rhd> t" and "(t,u) \<in> R" shows "\<exists>t. (s,t) \<in> R \<and> t \<rhd> u"
proof -
  from `s \<rhd> t` obtain "C" where "C \<noteq> \<box>" and "C\<langle>t\<rangle> = s" by auto
  from `ctxt_closed R` and `(t,u) \<in> R`
    have RCtCu: "(C\<langle>t\<rangle>,C\<langle>u\<rangle>) \<in> R" by (rule ctxt_closed_imp_ctxt)
  from `C \<noteq> \<box>` have "C\<langle>u\<rangle> \<rhd> u" by (rule nectxt_imp_supt_ctxt)
  from `(C\<langle>t\<rangle>,C\<langle>u\<rangle>) \<in> R` have "(s,C\<langle>u\<rangle>) \<in> R" unfolding `C\<langle>t\<rangle> = s` by simp
  from this and `C\<langle>u\<rangle> \<rhd> u` show ?thesis by auto
qed

lemma ctxt_closed_R_imp_R_qc_supt:
  assumes "ctxt_closed R" shows "quasi_commute R (op \<rhd>)" (is "quasi_commute ?r ?s")
proof -
  have "?s O ?r \<subseteq> ?r O (?r \<union> ?s)^*"
  proof (rule subrelI)
    fix x y assume "(x,y) \<in> ?s O ?r"
    then obtain z where "x \<rhd> z" and "(z,y) \<in> R" by best
    with `ctxt_closed R` obtain w where "(x,w) \<in> R" and "w \<rhd> y"
      using ctxt_closed_supt_R_imp_R_supt[of R x z y] by auto
    hence "(w,y) \<in> ?s" by simp
    hence "(w,y) \<in> ?r \<union> ?s" by auto
    hence "(w,y) \<in> (?r \<union> ?s)^*" by auto
    with `(x,w) \<in> R` show "(x,y) \<in> ?r O (?r \<union> ?s)^*" by auto
  qed
  thus ?thesis unfolding quasi_commute_def by force
qed

lemma ctxt_closure_idemp[simp]: "CC (CC R) = (CC R)"
proof
 show "CC(CC R) \<subseteq> CC R"
 proof (rule subrelI)
  fix s t assume "(s,t) \<in> CC(CC R)"
  thus "(s,t) \<in> CC R"
  proof
   fix s' t' C assume s:"s = C\<langle>s'\<rangle>" and t: "t = C\<langle>t'\<rangle>" and "(s',t') \<in> CC R"
   from `(s',t') \<in> CC R` show "(s,t) \<in> CC R"
   proof
    fix s'' t'' D assume s': "s' = D\<langle>s''\<rangle>" and t': "t' = D\<langle>t''\<rangle>" and "(s'',t'') \<in> R"
    hence Rst'': "(s'',t'') \<in> R" by auto
    with s t s' t' have st: "s = (C \<circ>c D)\<langle>s''\<rangle>" "t = (C \<circ>c D)\<langle>t''\<rangle>" by auto
    from Rst'' show ?thesis unfolding st by (rule ctxt_closureI)
   qed
  qed
 qed
next
 show "CC R \<subseteq> CC(CC R)"
 proof (rule subrelI)
  fix s t assume "(s,t) \<in> CC R"
  hence "(\<box>\<langle>s\<rangle>,\<box>\<langle>t\<rangle>) \<in> CC(CC R)" by (rule ctxt_closureI)
  thus "(s,t) \<in> CC(CC R)" by auto
 qed
qed

text {* Let~$R$ be a relation on terms that is closed under contexts. If~$R$
is well-founded then $R \cup \rhd$ is well-founed. *}
lemma SN_imp_SN_union_supt:
  assumes "SN R" and "ctxt_closed R"
  shows "SN (R \<union> op \<rhd>)"
proof -
  from `ctxt_closed R` have "quasi_commute R (op \<rhd>)" by (rule ctxt_closed_R_imp_R_qc_supt)
  have "SN (op \<rhd>)" by (rule SN_supt)
  from `SN R` and `SN (op \<rhd>)` and `quasi_commute R (op \<rhd>)`
  show ?thesis by (rule quasi_commute_imp_SN)
qed

inductive_set
  subst_closure :: "('f, 'v) trs \<Rightarrow> ('f, 'v) trs" ("CS")
  for R
where
 subst_closureI[intro]: "(s, t) \<in> R \<Longrightarrow> (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> CS R"


lemma subst_closureI2: "(l,r) \<in> R \<Longrightarrow> s = l \<cdot> \<sigma> \<Longrightarrow> t = r \<cdot> \<sigma> \<Longrightarrow> (s,t) \<in> CS R"
  using subst_closure.subst_closureI by blast

lemma subset_subst_closure: "R \<subseteq> CS R"
proof
  fix s t assume "(s, t) \<in> R"
  from subst_closureI[OF this, of Var] show "(s, t) \<in> CS R" by simp
qed

definition
  subst_closed :: "('f, 'v) trs \<Rightarrow> bool"
where 
  "subst_closed R = (CS R \<subseteq> R)"

lemma subst_closedI[intro]:
  assumes "\<And>s t \<sigma>. (s, t) \<in> R \<Longrightarrow> (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> R"
  shows "subst_closed R"
proof -
  have "CS R \<subseteq> R"
  proof
    fix s t assume "(s, t) \<in> CS R" thus "(s, t) \<in> R"
      by (cases) (auto intro: assms)
  qed
  thus ?thesis by (simp add: subst_closed_def)
qed

lemma subst_closedE[elim]:
  assumes "subst_closed R" and "(s, t) \<in> R"
  shows "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> R"
  using assms unfolding subst_closed_def by auto

lemma subst_closed_correctness: assumes "subst_closed R" shows "R = CS R"
using assms subset_subst_closure unfolding subst_closed_def by auto

lemma subst_closed_union: 
  assumes r: "subst_closed r"
  and s: "subst_closed s"
  shows "subst_closed (r \<union> s)"
unfolding subst_closed_def
proof (rule subrelI)
  fix x y	
  assume "(x,y) \<in> CS (r \<union> s)"
  thus "(x,y) \<in> r \<union> s"
  proof
    fix xx yy sig
    assume x: "x = xx \<cdot> sig" and y: "y = yy \<cdot> sig" and mem: "(xx,yy) \<in> r \<union> s"
    from mem show ?thesis
    proof
      assume "(xx,yy) \<in> r"
      hence "(x,y) \<in> CS r" by (simp add: x y, rule subst_closureI)
      thus ?thesis using r unfolding subst_closed_def by auto
    next
      assume "(xx,yy) \<in> s"
      hence "(x,y) \<in> CS s" by (simp add: x y, rule subst_closureI)
      thus ?thesis using s unfolding subst_closed_def by auto
    qed
  qed
qed

lemma subst_closed_comp: 
  assumes r: "subst_closed r"
  and s: "subst_closed s"
  shows "subst_closed (r O s)"
unfolding subst_closed_def
proof (rule subrelI)
  fix x y	
  assume "(x,y) \<in> CS (r O s)"
  thus "(x,y) \<in> r O s"
  proof
    fix xx yy sig
    assume x: "x = xx \<cdot> sig" and y: "y = yy \<cdot> sig" and mem: "(xx,yy) \<in> r O s"    
    from mem obtain zz where xx: "(xx,zz) \<in> r" and yy: "(zz,yy) \<in> s" by auto
    let ?z = "zz \<cdot> sig"
    from xx have "(x,?z) \<in> CS r" by (simp add: x, rule subst_closureI)
    hence x: "(x,?z) \<in> r" using r unfolding subst_closed_def by auto
    from yy have "(?z,y) \<in> CS s" 
      by (simp add: y, rule subst_closureI)
    hence y: "(?z,y) \<in> s" using s unfolding subst_closed_def by auto
    from x y show "(x,y) \<in> r O s" by auto
  qed
qed

lemma subst_closed_rtrancl: 
  assumes r: "subst_closed r"
  shows "subst_closed (r^*)"
unfolding subst_closed_def
proof (rule subrelI)
  fix x y	
  assume "(x,y) \<in> CS (r^*)"
  thus "(x,y) \<in> r^*"
  proof
    fix xx yy sig
    assume x: "x = xx \<cdot> sig" and y: "y = yy \<cdot> sig" and mem: "(xx,yy) \<in> r^*"
    from mem have "(xx \<cdot> sig, yy \<cdot> sig) \<in> r^*"
    proof (induct, simp)
      case (step yy zz)
      hence "(yy \<cdot> sig, zz \<cdot> sig) \<in> CS r" by auto
      hence "(yy \<cdot> sig, zz \<cdot> sig) \<in> r" using r unfolding subst_closed_def by auto
      with step show ?case by auto
    qed
    with x y show "(x,y) \<in> r^*" by simp
  qed
qed



lemma subst_closed_trancl: 
  assumes r: "subst_closed r"
  shows "subst_closed (r^+)"
unfolding subst_closed_def
proof (rule subrelI)
  fix x y	
  assume "(x,y) \<in> CS (r^+)"
  thus "(x,y) \<in> r^+"
  proof
    fix xx yy sig
    assume x: "x = xx \<cdot> sig" and y: "y = yy \<cdot> sig" and mem: "(xx,yy) \<in> r^+"
    from mem have "(xx \<cdot> sig, yy \<cdot> sig) \<in> r^+"
    proof (induct)
      case (step yy zz)
      hence "(yy \<cdot> sig, zz \<cdot> sig) \<in> CS r" by auto
      hence "(yy \<cdot> sig, zz \<cdot> sig) \<in> r" using r unfolding subst_closed_def by auto
      with step show ?case by auto
    next
      case (base yy)
      hence "(xx \<cdot> sig, yy \<cdot> sig) \<in> CS r" by auto
      hence "(xx \<cdot> sig, yy \<cdot> sig) \<in> r" using r unfolding subst_closed_def by auto
      thus ?case by auto
    qed
    with x y show "(x,y) \<in> r^+" by simp
  qed
qed

lemma subst_closed_decompose:
  assumes "(s,t) \<in> (CS R)" shows "\<exists>\<sigma> u v. (s = u\<cdot>\<sigma>) \<and> (t = v\<cdot>\<sigma>) \<and> (u,v) \<in> R"
using assms by (cases rule: subst_closure.cases) force


lemma stable_loop_imp_not_SN: assumes stable: "subst_closed r" and steps: "(s,s \<cdot> sig) \<in> r^+"
  shows "\<not> SN_on r {s}"
proof -
  let ?f =  "\<lambda> i. s \<cdot> (power.power Var op \<circ>s sig i)"
  have main: "\<And> i. (?f i, ?f (Suc i)) \<in> r^+"
  proof -
    fix i
    show "(?f i, ?f (Suc i)) \<in> r^+"
    proof (induct i, simp add: steps)
      case (Suc i)
      from Suc subst_closed_trancl[OF stable] have step: "(?f i \<cdot> sig, ?f (Suc i) \<cdot> sig) \<in> r^+" unfolding subst_closed_def
        by auto
      let ?sigg = "power.power Var op \<circ>s sig i"
      let ?siggs = "power.power Var op \<circ>s sig (Suc i)"
      have idi: "?sigg \<circ>s sig = sig \<circ>s ?sigg" by (rule subst_monoid_mult.power_commutes)
      have idsi: "?siggs \<circ>s sig = sig \<circ>s ?siggs" by (rule subst_monoid_mult.power_commutes)
      have "?f i \<cdot> sig = s \<cdot> ?sigg  \<circ>s sig" by simp
      also have "\<dots> = ?f (Suc i)" unfolding idi by simp
      finally have one: "?f i \<cdot> sig = ?f (Suc i)" .
      have "?f (Suc i) \<cdot> sig = s \<cdot> ?siggs \<circ>s sig" by simp
      also have "\<dots> = ?f (Suc (Suc i))" unfolding idsi by simp
      finally have two: "?f (Suc i) \<cdot> sig = ?f (Suc (Suc i))" by simp
      show ?case using one two step by simp
    qed
  qed
  hence "\<not> SN_on (r^+) {?f 0}" unfolding SN_on_def by best
  thus ?thesis using SN_on_trancl by force
qed

lemma subst_subst_closed_subseteq_subst_closed:
  fixes R::"('f, 'v) trs" shows "CS (CS R) \<subseteq> CS R"
proof (rule subrelI)
  fix s t :: "('f, 'v) term" assume a: "(s,t) \<in> CS (CS R)"
  hence "\<exists>(\<sigma>::('f, 'v) subst) u v. (s = u\<cdot>\<sigma>) \<and> (t = v\<cdot>\<sigma>) \<and> (u,v) \<in> CS R"
    using subst_closed_decompose by auto
  hence "\<exists>(\<sigma>::('f, 'v) subst) u v. (s = u\<cdot>\<sigma>) \<and> (t = v\<cdot>\<sigma>) \<and>
    (\<exists>(\<tau>::('f, 'v) subst) l r. (u = l\<cdot>\<tau>) \<and> (v = r\<cdot>\<tau>) \<and> (l,r) \<in> R)"
    using subst_closed_decompose[of _ _ R] by best
  hence "\<exists>(\<sigma>::('f, 'v) subst) (\<tau>::('f, 'v) subst) l r.
    (s = l \<cdot> (\<tau> \<circ>s \<sigma>)) \<and> (t = r \<cdot> (\<tau> \<circ>s \<sigma>)) \<and> (l, r) \<in> R" by auto
  thus "(s,t) \<in> (CS R)" by blast
qed

lemma subst_closed_idemp[simp]: "CS (CS R) = CS R"
using subst_subst_closed_subseteq_subst_closed subset_subst_closure by auto

lemma subst_closed_closure_subst[simp]: "subst_closed (CS R)"
unfolding subst_closed_def by simp

lemma supteq_subst_closed: "subst_closed (op \<unrhd>)"
unfolding subst_closed_def
proof (rule subrelI)
  fix s t :: "('f, 'v) term" assume "(s, t) \<in> CS (op \<unrhd>)"
  then obtain u v and \<sigma> :: "('f, 'v) subst" where "u \<unrhd> v" and "u \<cdot> \<sigma> = s" and "v \<cdot> \<sigma> = t"
    by (cases rule: subst_closure.cases) auto
  with supteq_stable[of u v \<sigma>] show "s \<unrhd> t" by simp
qed

lemma supt_subst_closed: "subst_closed (op \<rhd>)"
unfolding subst_closed_def
proof (rule subrelI)
  fix s t :: "('f, 'v) term"  assume "(s, t) \<in> CS (op \<rhd>)"
  then obtain u v and \<sigma> :: "('f, 'v) subst" where s: "s = u \<cdot> \<sigma>" and t: "t = v \<cdot> \<sigma>" and "u \<rhd> v"
    by (cases rule: subst_closure.cases) auto
  thus "s \<rhd> t" unfolding s t using supt_stable by simp
qed

text {*The single-step relation induced by a TRS~@{text R} is the relation @{text R} closed under
substitutions and contexts.*}
inductive_set rstep :: "('f, 'v) trs \<Rightarrow> ('f, 'v) term rel" for R
where
  id [intro]: "(s, t) \<in> R \<Longrightarrow> (s, t) \<in> rstep R" |
  subst [intro]: "(s, t) \<in> rstep R \<Longrightarrow> (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> rstep R" |
  ctxt [intro]: "(s, t) \<in> rstep R \<Longrightarrow> (C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> rstep R"

(*TODO: move to AFP/Abstract_Rewriting*)
lemma WCR_alt_def:
  "WCR A \<longleftrightarrow> A\<inverse> O A \<subseteq> A\<^sup>\<down>"
  by (auto simp: WCR_defs)

(*TODO: move to AFP/Abstract_Rewriting*)
lemma join_sym:
  "(s, t) \<in> A\<^sup>\<down> \<Longrightarrow> (t, s) \<in> A\<^sup>\<down>"
  by auto

lemma rstep_relcomp_idemp1 [simp]:
  "rstep (rstep R O rstep S) = rstep R O rstep S"
proof -
  { fix s t
    assume "(s, t) \<in> rstep (rstep R O rstep S)"
    then have "(s, t) \<in> rstep R O rstep S"
      by (induct) blast+ }
  then show ?thesis by auto
qed

lemma rstep_relcomp_idemp2 [simp]:
  "rstep (rstep R O rstep S O rstep T) = rstep R O rstep S O rstep T"
proof -
  { fix s t
    assume "(s, t) \<in> rstep (rstep R O rstep S O rstep T)"
    then have "(s, t) \<in> rstep R O rstep S O rstep T"
      by (induct) blast+ }
  then show ?thesis by auto
qed

lemma rstep_CC_CS: "rstep R = CC (CS R)"
proof
  show "rstep R \<subseteq> CC (CS R)"
  proof
    fix s t assume "(s, t) \<in> rstep R" thus "(s, t) \<in> CC (CS R)"
    proof (induct)
      case (id s t)
      from ctxt_closureI[OF subst_closureI[OF `(s, t) \<in> R`], of \<box> Var] show ?case using id by simp
    next
      case (subst s t \<sigma>)
      from `(s, t) \<in> CC (CS R)` obtain C u v where "(u, v) \<in> CS R"
        and s: "s = C\<langle>u\<rangle>" and t: "t = C\<langle>v\<rangle>" by cases auto
      from subst_closureI[OF `(u, v) \<in> CS R`, of \<sigma>] have "(u \<cdot> \<sigma>, v \<cdot> \<sigma>) \<in> CS R" by simp
      from ctxt_closureI[OF this, of "C \<cdot>c \<sigma>"] show ?case unfolding s t by simp
    next
      case (ctxt s t C)
      from ctxt_closureI[OF `(s, t) \<in> CC (CS R)`, of C] show ?case by simp
    qed
  qed
next
  show "CC (CS R) \<subseteq> rstep R"
  proof
    fix s t assume "(s, t) \<in> CC (CS R)"
    then obtain C u v where s: "s = C\<langle>u\<rangle>" and t: "t = C\<langle>v\<rangle>" and "(u, v) \<in> CS R" by cases auto
    from `(u, v) \<in> CS R` obtain \<sigma> u' v' where u: "u = u' \<cdot> \<sigma>" and v: "v = v' \<cdot> \<sigma>"
      and "(u', v') \<in> R" by (cases) auto
    from `(u', v') \<in> R` have "(u', v') \<in> rstep R" ..
    hence "(u, v) \<in> rstep R" unfolding u v ..
    thus "(s, t) \<in> rstep R" unfolding s t ..
  qed
qed

lemma ctxt_closed_rstep: "ctxt_closed (rstep R)"
unfolding ctxt_closed_def rstep_CC_CS ctxt_closure_idemp ..

lemma ctxt_closed_rstep_seq: "ctxt_closed ((rstep R)^*)" unfolding ctxt_closed_def
proof (rule subrelI)
 fix s t assume "(s,t) \<in> CC((rstep R)^*)" thus "(s,t) \<in> (rstep R)^*"
 proof (induct)
  case (ctxt_closureI s t C) thus ?case
  proof (induct)
   case base thus ?case by simp
  next
   case (step y z)
   hence "(C\<langle>y\<rangle>,C\<langle>z\<rangle>) \<in> rstep R" using ctxt_closed_rstep unfolding ctxt_closed_def by blast
   with `(C\<langle>s\<rangle>,C\<langle>y\<rangle>) \<in> (rstep R)^*` show ?case by simp
  qed
 qed
qed

lemma subset_rstep: "R \<subseteq> rstep R" by auto

lemma subst_closure_rstep_subset: "CS (rstep R) \<subseteq> rstep R"
proof
  fix s t assume "(s, t) \<in> CS (rstep R)"
  then obtain \<sigma> u v
    where s: "s = u \<cdot> \<sigma>" and t: "t = v \<cdot> \<sigma>" and "(u, v) \<in> rstep R" by cases auto
  thus "(s, t) \<in> rstep R" unfolding s t by auto
qed

lemma subst_closed_rstep: "subst_closed (rstep R)"
using subst_closure_rstep_subset unfolding subst_closed_def by auto

lemma subst_closed_rstep_seq: "subst_closed ((rstep R)^*)" unfolding subst_closed_def
proof (rule subrelI)
 fix s t assume "(s,t) \<in> CS((rstep R)^*)" thus "(s,t) \<in> (rstep R)^*"
 proof (induct)
  case (subst_closureI s t \<sigma>) thus ?case
  proof (induct)
   case base thus ?case by simp
  next
   case (step y z)
   hence "(y \<cdot> \<sigma>,z \<cdot> \<sigma>) \<in> rstep R" using subst_closed_rstep unfolding subst_closed_def by blast
   with `(s \<cdot> \<sigma>,y \<cdot> \<sigma>) \<in> (rstep R)^*` show ?case by simp
  qed
 qed
qed

lemma comp_stable:
  assumes R: "\<And>x y \<sigma>. (x, y) \<in> R \<Longrightarrow> (x \<cdot> \<sigma>, y \<cdot> \<sigma>) \<in> R"
    and S: "\<And>x y \<sigma>. (x, y) \<in> S \<Longrightarrow> (x \<cdot> \<sigma>, y \<cdot> \<sigma>) \<in> S"
  shows "\<And>x y \<sigma>. (x, y) \<in> R O S \<Longrightarrow> (x \<cdot> \<sigma>, y \<cdot> \<sigma>) \<in> R O S"
proof -
 fix s t \<sigma> assume "(s,t) \<in> R O S"
 then obtain u where "(s,u) \<in> R" and "(u,t) \<in> S" by auto
 from R[OF `(s,u) \<in> R`] have "(s \<cdot> \<sigma>,u \<cdot> \<sigma>) \<in> R" by simp
 from S[OF `(u,t) \<in> S`] have "(u \<cdot> \<sigma>,t \<cdot> \<sigma>) \<in> S" by simp
 with `(s \<cdot> \<sigma>,u \<cdot> \<sigma>) \<in> R`
 show "(s \<cdot> \<sigma>,t \<cdot> \<sigma>) \<in> R O S" by auto
qed

lemma comp_subst_closed: assumes "subst_closed R" and "subst_closed S" shows "subst_closed(R O S)" unfolding subst_closed_def
proof (rule subrelI)
 fix s t assume "(s,t) \<in> CS(R O S)"
 then obtain u v \<sigma> where s: "s = u \<cdot> \<sigma>" and t: "t = v \<cdot> \<sigma>" and "(u,v) \<in> R O S" by (cases rule: subst_closure.cases) auto
 from `(u,v) \<in> R O S` obtain w where "(u,w) \<in> R" and "(w,v) \<in> S" by auto
 from `subst_closed R` and `(u,w) \<in> R` have "(s,w \<cdot> \<sigma>) \<in> R" unfolding subst_closed_def s by auto
 from `subst_closed S` and `(w,v) \<in> S` have "(w \<cdot> \<sigma>,t) \<in> S" unfolding subst_closed_def t by auto
 with `(s,w \<cdot> \<sigma>) \<in> R` show "(s,t) \<in> R O S" by auto
qed

lemma ctxt_closed_supt_subset:
  assumes "ctxt_closed R" shows "op \<rhd> O R \<subseteq> R O op \<rhd>"
proof
  fix s t assume "(s, t) \<in> op \<rhd> O R"
  then obtain u where "s \<rhd> u" and "(u, t) \<in> R" by auto
  obtain v where "(s, v) \<in> R" and "v \<rhd> t"
    using ctxt_closed_R_imp_supt_R_distr[OF assms `s \<rhd> u` `(u, t) \<in> R`] by auto
  from `v \<rhd> t` and `(s, v) \<in> R` show "(s, t) \<in> R O op \<rhd>" by auto
qed

lemmas supt_rsteps_subset = ctxt_closed_supt_subset[OF ctxt_closed_rstep_seq]

lemma supteq_rsteps_subset: "op \<unrhd> O (rstep R)^* \<subseteq> (rstep R)^* O op \<unrhd>" (is "?S \<subseteq> ?T")
proof
  fix s t assume "(s, t) \<in> ?S"
  hence "(s, t) \<in> op \<rhd> O (rstep R)^* \<union> (rstep R)^*" by (auto simp: supt_supteq_conv)
  thus "(s, t) \<in> ?T"
  proof
    assume "(s, t) \<in> op \<rhd> O (rstep R)^*" thus ?thesis
      using supt_rsteps_subset unfolding supt_supteq_set_conv by auto
  next
    assume "(s, t) \<in> (rstep R)^*" thus ?thesis by auto
  qed
qed

lemma quasi_commute_rsteps_supt:
  "quasi_commute ((rstep R)^*) (op \<rhd>)"
  unfolding quasi_commute_def using supt_rsteps_subset [of R] by auto

definition
  rstep_r_p_s :: "('f, 'v) trs \<Rightarrow> ('f, 'v) rule \<Rightarrow> pos \<Rightarrow> ('f, 'v) subst \<Rightarrow> ('f, 'v) trs"
where
  "rstep_r_p_s R r p \<sigma> = {(s, t).
    let C = ctxt_of_pos_term p s in p \<in> poss s \<and> r \<in> R \<and> (C\<langle>fst r \<cdot> \<sigma>\<rangle> = s) \<and> (C\<langle>snd r \<cdot> \<sigma>\<rangle> = t)}"

lemma rstep_r_p_s_def':
  "rstep_r_p_s R r p \<sigma> = {(s, t).
    p \<in> poss s \<and> r \<in> R \<and> s |_ p = fst r \<cdot> \<sigma> \<and> t = replace_at s p (snd r \<cdot> \<sigma>)}" (is "?l = ?r")
proof -
  { fix s t
    have "((s,t) \<in> ?l) = ((s,t) \<in> ?r)"    
      unfolding rstep_r_p_s_def Let_def 
      using ctxt_supt_id [of p s] and subt_at_ctxt_of_pos_term [of p s "fst r \<cdot> \<sigma>"] by auto }
  then show ?thesis by auto
qed

lemma parallel_steps:
  assumes "(s, t) \<in> rstep_r_p_s R (l\<^sub>1, r\<^sub>1) p\<^sub>1 \<sigma>\<^sub>1"
    and "(s, u) \<in> rstep_r_p_s R (l\<^sub>2, r\<^sub>2) p\<^sub>2 \<sigma>\<^sub>2"
    and par: "p\<^sub>1 \<bottom> p\<^sub>2"
  shows "(t, (ctxt_of_pos_term p\<^sub>1 u)\<langle>r\<^sub>1 \<cdot> \<sigma>\<^sub>1\<rangle>) \<in> rstep_r_p_s R (l\<^sub>2, r\<^sub>2) p\<^sub>2 \<sigma>\<^sub>2 \<and>
         (u, (ctxt_of_pos_term p\<^sub>1 u)\<langle>r\<^sub>1 \<cdot> \<sigma>\<^sub>1\<rangle>) \<in> rstep_r_p_s R (l\<^sub>1, r\<^sub>1) p\<^sub>1 \<sigma>\<^sub>1"
proof -
  have p1: "p\<^sub>1 \<in> poss s" and lr1: "(l\<^sub>1, r\<^sub>1) \<in> R" and \<sigma>1: "s |_ p\<^sub>1 = l\<^sub>1 \<cdot> \<sigma>\<^sub>1"
    and t: "t = replace_at s p\<^sub>1 (r\<^sub>1 \<cdot> \<sigma>\<^sub>1)"
    and p2: "p\<^sub>2 \<in> poss s" and lr2: "(l\<^sub>2, r\<^sub>2) \<in> R" and \<sigma>2: "s |_ p\<^sub>2 = l\<^sub>2 \<cdot> \<sigma>\<^sub>2"
    and u: "u = replace_at s p\<^sub>2 (r\<^sub>2 \<cdot> \<sigma>\<^sub>2)" using assms by (auto simp: rstep_r_p_s_def')

  have "replace_at t p\<^sub>2 (r\<^sub>2 \<cdot> \<sigma>\<^sub>2) = replace_at u p\<^sub>1 (r\<^sub>1 \<cdot> \<sigma>\<^sub>1)"
    using t and u and parallel_replace_at [OF `p\<^sub>1 \<bottom> p\<^sub>2` p1 p2] by auto
  moreover
  have "(t, (ctxt_of_pos_term p\<^sub>2 t)\<langle>r\<^sub>2 \<cdot> \<sigma>\<^sub>2\<rangle>) \<in> rstep_r_p_s R (l\<^sub>2, r\<^sub>2) p\<^sub>2 \<sigma>\<^sub>2"
  proof -
    have "t |_ p\<^sub>2 = l\<^sub>2 \<cdot> \<sigma>\<^sub>2" using \<sigma>2 and parallel_replace_at_subt_at [OF par p1 p2] and t by auto
    moreover have "p\<^sub>2 \<in> poss t" using parallel_poss_replace_at [OF par p1] and t and p2 by auto
    ultimately show ?thesis using lr2 and ctxt_supt_id [of p\<^sub>2 t] by (simp add: rstep_r_p_s_def)
  qed
  moreover
  have "(u, (ctxt_of_pos_term p\<^sub>1 u)\<langle>r\<^sub>1 \<cdot> \<sigma>\<^sub>1\<rangle>) \<in> rstep_r_p_s R (l\<^sub>1, r\<^sub>1) p\<^sub>1 \<sigma>\<^sub>1"
  proof -
    have par': "p\<^sub>2 \<bottom> p\<^sub>1" using parallel_pos_sym [OF par] .
    have "u |_ p\<^sub>1 = l\<^sub>1 \<cdot> \<sigma>\<^sub>1" using \<sigma>1 and parallel_replace_at_subt_at [OF par' p2 p1] and u by auto
    moreover have "p\<^sub>1 \<in> poss u" using parallel_poss_replace_at [OF par' p2] and u and p1 by auto
    ultimately show ?thesis using lr1 and ctxt_supt_id [of p\<^sub>1 u] by (simp add: rstep_r_p_s_def)
  qed
  ultimately show ?thesis by auto
qed

lemma rstepE [elim]:
  assumes "(s, t) \<in> rstep R"
    and imp: "\<And>C \<sigma> l r. \<lbrakk>(l, r) \<in> R; s = C\<langle>l \<cdot> \<sigma>\<rangle>; t = C\<langle>r \<cdot> \<sigma>\<rangle>\<rbrakk> \<Longrightarrow> P"
  shows "P"
proof -
  from assms[unfolded rstep_CC_CS]
    obtain C u v where s: "s = C\<langle>u\<rangle>" and t: "t = C\<langle>v\<rangle>" and "(u, v) \<in> CS R" by cases auto
  from `(u, v) \<in> CS R`
    obtain \<sigma> u' v' where u: "u = u' \<cdot> \<sigma>" and v: "v = v' \<cdot> \<sigma>" and "(u', v') \<in> R" by cases auto
  from imp[OF `(u', v') \<in> R`, unfolded s t u v] show "P" by simp
qed

lemma rstepI [intro]:
  assumes "(l, r) \<in> R" and "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and "t = C\<langle>r \<cdot> \<sigma>\<rangle>"
  shows "(s, t) \<in> rstep R"
  using assms by auto

lemma rstep_iff_rstep_r_p_s:
  "(s, t) \<in> rstep R \<longleftrightarrow> (\<exists>l r p \<sigma>. (s, t) \<in> rstep_r_p_s R (l, r) p \<sigma>)" (is "?lhs = ?rhs")
proof
  assume "(s, t) \<in> rstep R"
  then obtain C \<sigma> l r where s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>" and "(l, r) \<in> R" by auto
  let ?p = "hole_pos C"
  let ?C = "ctxt_of_pos_term ?p s"
  have C: "ctxt_of_pos_term ?p s = C" unfolding s by (induct C) simp_all
  have "?p \<in> poss s" unfolding s by simp
  moreover have "(l, r) \<in> R" by fact
  moreover have "?C\<langle>l \<cdot> \<sigma>\<rangle> = s" unfolding C by (simp add: s)
  moreover have "?C\<langle>r \<cdot> \<sigma>\<rangle> = t" unfolding C by (simp add: t)
  ultimately show "?rhs" unfolding rstep_r_p_s_def Let_def by auto
next
  assume "?rhs"
  then obtain l r p \<sigma> where "p \<in> poss s"
    and "(l, r) \<in> R"
    and s[symmetric]: "(ctxt_of_pos_term p s)\<langle>l \<cdot> \<sigma>\<rangle> = s"
    and t[symmetric]: "(ctxt_of_pos_term p s)\<langle>r \<cdot> \<sigma>\<rangle> = t"
    unfolding rstep_r_p_s_def Let_def by auto
  thus "?lhs" by auto
qed

lemma rstep_r_p_s_imp_rstep:
  assumes "(s, t) \<in> rstep_r_p_s R r p \<sigma>"
  shows "(s, t) \<in> rstep R"
  using assms by (cases r) (auto simp: rstep_iff_rstep_r_p_s)

text {* Rewriting steps below the root position. *}
definition
  nrrstep :: "('f, 'v) trs \<Rightarrow> ('f, 'v) trs"
where
  "nrrstep R = {(s,t). \<exists>r i ps \<sigma>. (s,t) \<in> rstep_r_p_s R r (i<#ps) \<sigma>}"

text {* An alternative characterisation of non-root rewrite steps. *}
lemma nrrstep_def':
  "nrrstep R = {(s, t). \<exists>l r C \<sigma>. (l, r) \<in> R \<and> C \<noteq> \<box> \<and> s = C\<langle>l\<cdot>\<sigma>\<rangle> \<and> t = C\<langle>r\<cdot>\<sigma>\<rangle>}"
  (is "?lhs = ?rhs")
proof
  show "?lhs \<subseteq> ?rhs"
  proof (rule subrelI)
    fix s t assume "(s, t) \<in> nrrstep R"
    then obtain l r i ps \<sigma> where step: "(s, t) \<in> rstep_r_p_s R (l, r) (i <# ps) \<sigma>"
      unfolding nrrstep_def by best
    let ?C = "ctxt_of_pos_term (i <# ps) s"
    from step have"i <# ps \<in> poss s" and "(l, r) \<in> R" and "s = ?C\<langle>l\<cdot>\<sigma>\<rangle>" and "t = ?C\<langle>r\<cdot>\<sigma>\<rangle>"
      unfolding rstep_r_p_s_def Let_def by auto
    moreover from `i <# ps \<in> poss s` have "?C \<noteq> \<box>" by (induct s) auto
    ultimately show "(s, t) \<in> ?rhs" by auto
  qed
next
  show "?rhs \<subseteq> ?lhs"
  proof (rule subrelI)
    fix s t assume "(s, t) \<in> ?rhs"
    then obtain l r C \<sigma> where in_R: "(l, r) \<in> R" and "C \<noteq> \<box>"
      and s: "s = C\<langle>l\<cdot>\<sigma>\<rangle>" and t: "t = C\<langle>r\<cdot>\<sigma>\<rangle>" by auto
    from `C \<noteq> \<box>` obtain i p where ip: "hole_pos C = i <# p" by (induct C) auto
    have "i <# p \<in> poss s" using hole_pos_poss[of C] unfolding s ip by simp
    hence C: "C = ctxt_of_pos_term (i <# p) s"
      unfolding s ip[symmetric] by simp
    from `i <# p \<in> poss s` in_R s t have "(s, t) \<in> rstep_r_p_s R (l, r) (i <# p) \<sigma>"
      unfolding rstep_r_p_s_def C by simp
    thus "(s, t) \<in> nrrstep R" unfolding nrrstep_def by best
  qed
qed

lemma nrrstepI: "(l,r) \<in> R \<Longrightarrow> s = C\<langle>l\<cdot>\<sigma>\<rangle> \<Longrightarrow> t = C\<langle>r \<cdot> \<sigma>\<rangle> \<Longrightarrow> C \<noteq> \<box> \<Longrightarrow> (s,t) \<in> nrrstep R" unfolding nrrstep_def' by auto

lemma nrrstep_union: "nrrstep (R \<union> S) = nrrstep R \<union> nrrstep S"
  unfolding nrrstep_def' by blast

text {* Rewriting step at the root position. *}
definition
  rrstep :: "('f, 'v) trs \<Rightarrow> ('f, 'v) trs"
where
  "rrstep R = {(s,t). \<exists>r \<sigma>. (s,t) \<in> rstep_r_p_s R r \<epsilon> \<sigma>}"

text {* An alternative characterisation of root rewrite steps. *}
lemma rrstep_def': "rrstep R = {(s, t). \<exists>l r \<sigma>. (l, r) \<in> R \<and> s = l\<cdot>\<sigma> \<and> t = r\<cdot>\<sigma>}"
  (is "_ = ?rhs")
by (auto simp: rrstep_def rstep_r_p_s_def)

lemma rrstep_union: "rrstep (R \<union> S) = rrstep R \<union> rrstep S" unfolding rrstep_def' by blast

lemma rrstep_empty[simp]: "rrstep {} = {}"
  unfolding rrstep_def' by auto


lemma subst_closed_rrstep: "subst_closed (rrstep R)"
unfolding subst_closed_def
proof
  fix ss ts
  assume "(ss,ts) \<in> CS (rrstep R)"
  thus "(ss,ts) \<in> rrstep R"
  proof
    fix s t \<sigma>
    assume ss: "ss = s \<cdot> \<sigma>" and ts: "ts = t \<cdot> \<sigma>" and step: "(s,t) \<in> rrstep R"
    from step obtain l r \<delta> where lr: "(l,r) \<in> R" and s: "s = l \<cdot> \<delta>" and t: "t = r \<cdot> \<delta>" unfolding rrstep_def' by auto
    obtain sig where "sig = \<delta> \<circ>s \<sigma>" by auto
    with ss s ts t have "ss = l \<cdot> sig" and "ts = r \<cdot> sig" by simp+
    with lr show "(ss,ts) \<in> rrstep R" unfolding rrstep_def' by (auto simp: Let_def)
  qed
qed

lemma rstep_iff_rrstep_or_nrrstep: "rstep R = (rrstep R \<union> nrrstep R)"
proof
 show "rstep R \<subseteq> rrstep R \<union> nrrstep R"
 proof (rule subrelI)
  fix s t assume "(s,t) \<in> rstep R"
  then obtain l r p \<sigma> where rstep_rps: "(s,t) \<in> rstep_r_p_s R (l,r) p \<sigma>"
   by (auto simp: rstep_iff_rstep_r_p_s)
  thus "(s,t) \<in> rrstep R \<union> nrrstep R" unfolding rrstep_def nrrstep_def by (cases p) auto
 qed
next
 show "rrstep R \<union> nrrstep R \<subseteq> rstep R"
 proof (rule subrelI)
  fix s t assume "(s,t) \<in> rrstep R \<union> nrrstep R"
  thus "(s,t) \<in> rstep R" by (auto simp: rrstep_def nrrstep_def rstep_iff_rstep_r_p_s)
 qed
qed

lemma rstep_i_pos_imp_rstep_arg_i_pos:
 assumes nrrstep: "(Fun f ss,t) \<in> rstep_r_p_s R (l,r) (i<#ps) \<sigma>"
 shows "(ss!i,t|_<i>) \<in> rstep_r_p_s R (l,r) ps \<sigma>"
proof -
 from nrrstep obtain C where C:"C = ctxt_of_pos_term (i<#ps) (Fun f ss)"
  and pos: "(i<#ps) \<in> poss (Fun f ss)"
  and Rlr: "(l,r) \<in> R"
  and Fun: "C\<langle>l\<cdot>\<sigma>\<rangle> = Fun f ss"
  and t: "C\<langle>r\<cdot>\<sigma>\<rangle> = t" unfolding rstep_r_p_s_def Let_def by auto
 then obtain D where C':"C = More f (take i ss) D (drop (Suc i) ss)" by auto
 hence CFun: "C\<langle>l\<cdot>\<sigma>\<rangle> = Fun f (take i ss @ (D\<langle>l\<cdot>\<sigma>\<rangle>) # drop (Suc i) ss)" by auto
 from pos have len: "i < length ss" by auto
 from len
 have "(take i ss @ (D\<langle>l\<cdot>\<sigma>\<rangle>) # drop (Suc i) ss)!i = D\<langle>l\<cdot>\<sigma>\<rangle>" by (auto simp: nth_append)
 with C Fun CFun have ssi: "ss!i = D\<langle>l\<cdot>\<sigma>\<rangle>" by auto
 from C' t have t': "t = Fun f (take i ss @ (D\<langle>r\<cdot>\<sigma>\<rangle>) # drop (Suc i) ss)" by auto
 from len
 have "(take i ss @ (D\<langle>r\<cdot>\<sigma>\<rangle>) # drop (Suc i) ss)!i = D\<langle>r\<cdot>\<sigma>\<rangle>" by (auto simp: nth_append)
 with t' have "t|_<i> = (D\<langle>r\<cdot>\<sigma>\<rangle>)|_\<epsilon>" by auto
 hence subt_at: "t|_<i> = D\<langle>r\<cdot>\<sigma>\<rangle>" by simp
 from C C' have "D = ctxt_of_pos_term ps (ss!i)" by auto
 with pos Rlr ssi[symmetric] subt_at[symmetric]
 show ?thesis unfolding rstep_r_p_s_def Let_def by auto
qed

lemma ctxt_closure_is_closed_under_ctxts: "ctxt_closed (CC R)"
unfolding ctxt_closed_def by auto

lemma ctxt_closed_rstep_id[simp]: "CC (rstep R) = rstep R" by (simp add: rstep_CC_CS)

lemma subst_closed_rstep_equal_rstep: "CS (rstep R) = rstep R" 
using subst_closure_rstep_subset subst_closed_correctness unfolding subst_closed_def by blast

lemma supt_rstep_subset: "op \<rhd> O rstep R \<subseteq> rstep R O op \<rhd>"
proof (rule subrelI)
  fix s t assume "(s,t) \<in> op \<rhd> O rstep R" thus "(s,t) \<in> rstep R O op \<rhd>"
  proof (induct s)
    case (Var x)
    hence "\<exists>u. Var x \<rhd> u \<and> (u,t) \<in> rstep R" by auto
    then obtain u where "Var x \<rhd> u" and "(u,t) \<in> rstep R" by auto
    from `Var x \<rhd> u` show ?case by (cases rule: supt.cases)
  next
    case (Fun f ss)
    then obtain u where "Fun f ss \<rhd> u" and "(u,t) \<in> rstep R" by auto
    from `Fun f ss \<rhd> u` obtain C
      where "C \<noteq> \<box>" and "C\<langle>u\<rangle> = Fun f ss" by auto
    from `C \<noteq> \<box>` have "C\<langle>t\<rangle> \<rhd> t" by (rule nectxt_imp_supt_ctxt)
    from `(u,t) \<in> rstep R` have "(C\<langle>u\<rangle>,C\<langle>t\<rangle>) \<in> rstep R" ..
    hence "(Fun f ss,C\<langle>t\<rangle>) \<in> rstep R" unfolding `C\<langle>u\<rangle> = Fun f ss` .
    with `C\<langle>t\<rangle> \<rhd> t` show ?case by auto
  qed
qed

lemma ne_rstep_seq_imp_list_of_terms:
  assumes "(s,t) \<in> (rstep R)^+"
  shows "\<exists>ts. length ts > 1 \<and> ts!0 = s \<and> ts!(length ts - 1) = t \<and>
         (\<forall>i<length ts - 1. (ts!i,ts!(Suc i)) \<in> (rstep R))" (is "\<exists>ts. _ \<and> _ \<and> _ \<and> ?P ts")
using assms proof (induct rule: trancl.induct)
  case (r_into_trancl x y)
  let ?ts = "[x,y]"
  have "length ?ts > 1" by simp
  moreover have "?ts!0 = x" by simp
  moreover have "?ts!(length ?ts - 1) = y" by simp
  moreover from r_into_rtrancl r_into_trancl have "?P ?ts" by auto
  ultimately show ?case by fast
next
  case (trancl_into_trancl x y z)
  then obtain ts where "length ts > 1" and "ts!0 = x"
    and last1: "ts!(length ts - 1) = y" and "?P ts" by auto
  let ?ts = "ts@[z]"
  have len: "length ?ts = length ts + 1" by simp
  from `length ts > 1` have "length ?ts > 1" by auto
  moreover with `ts!0 = x` have "?ts!0 = x" by (induct ts) auto
  moreover have last2: "?ts!(length ?ts - 1) = z" by simp
  moreover have "?P ?ts"
  proof (intro allI impI)
    fix i assume A: "i < length ?ts - 1"
    show "(?ts!i,?ts!(Suc i)) \<in> rstep R"
    proof (cases "i < length ts - 1")
      case True with `?P ts` A show ?thesis unfolding len unfolding nth_append by auto
    next
      case False with A have "i = length ts - 1" by simp
      with last1 `length ts > 1` have "?ts!i = y" unfolding nth_append by auto
      have "Suc i = length ?ts - 1" using `i = length ts - 1` using `length ts > 1` by auto
      with last2 have "?ts!(Suc i) = z" by auto
      from `(y,z) \<in> rstep R` show ?thesis unfolding `?ts!(Suc i) = z` `?ts!i = y` .
    qed
  qed
  ultimately show ?case by fast
qed

definition restrict_SN_supt :: "('f, 'v) trs \<Rightarrow> ('f, 'v) trs" where
  "restrict_SN_supt R = (restrict_SN R R) \<union> (restrict_SN (op \<rhd>) R)"

lemma restrict_SN_imp_SN_subt:
  assumes "ctxt_closed R" and "SN_on R {s}" and "s \<unrhd> t"
  shows "SN_on R {t}"
proof (rule ccontr)
  assume "\<not> SN_on R {t}"
  then obtain A where "A 0 = t" and "\<forall>i. (A i,A (Suc i)) \<in> R"
    unfolding SN_on_def by best
  from `s \<unrhd> t` obtain C where "s = C\<langle>t\<rangle>" by auto
  let ?B = "\<lambda>i. C\<langle>A i\<rangle>"
  have "\<forall>i. (?B i,?B(Suc i)) \<in> R"
  proof
    fix i
    from `\<forall>i. (A i,A(Suc i)) \<in> R` have "(?B i,?B(Suc i)) \<in> CC(R)" using ctxt_closureI by fast
    thus "(?B i,?B(Suc i)) \<in> R" using `ctxt_closed R` unfolding ctxt_closed_def by auto
  qed
  with `A 0 = t` have "?B 0 = s \<and> (\<forall>i. (?B i,?B(Suc i)) \<in> R)" unfolding `s = C\<langle>t\<rangle>` by auto
  hence "\<not> SN_on R {s}" unfolding SN_on_def by auto
  with assms show "False" by simp
qed

lemma ctxt_closed_R_imp_supt_restrict_SN_distr:
  assumes "ctxt_closed R"
    and "(s,t) \<in> (restrict_SN (op \<rhd>) R)"
    and "(t,u) \<in> restrict_SN R R"
  shows "(\<exists>t. (s,t) \<in> restrict_SN R R \<and> (t,u) \<in> restrict_SN (op \<rhd>) R)" (is "\<exists> t. _ \<and> (t,u) \<in> ?snSub")
proof -
  from `(s,t) \<in> ?snSub` have "SN_on R {s}" and "s \<rhd> t" unfolding restrict_SN_def supt_def by auto
  from `s \<rhd> t` obtain C where "C \<noteq> Hole" and "C\<langle>t\<rangle> = s" by best
  from `(t,u) \<in> restrict_SN R R` have "(t,u) \<in> R" unfolding restrict_SN_def by auto
  from `ctxt_closed R` and this have RCtCu: "(C\<langle>t\<rangle>,C\<langle>u\<rangle>) \<in> R" by (rule ctxt_closed_imp_ctxt)
  with `C\<langle>t\<rangle> = s` `SN_on R {s}` have one: "SN_on R {C\<langle>u\<rangle>}"
    using step_preserves_SN_on[of "C\<langle>t\<rangle>" "C\<langle>u\<rangle>" R] by blast
  from `C \<noteq> \<box>` have "C\<langle>u\<rangle> \<rhd> u" by auto
  with one have "(C\<langle>u\<rangle>,u) \<in> ?snSub" unfolding restrict_SN_def supt_def by auto
  from RCtCu have "(s,C\<langle>u\<rangle>) \<in> R" unfolding `C\<langle>t\<rangle> = s` .
  from this and `SN_on R {s}` and `(C\<langle>u\<rangle>,u) \<in> ?snSub` show ?thesis unfolding restrict_SN_def by auto
qed

lemma ctxt_closed_R_imp_restrict_SN_qc_supt:
  assumes "ctxt_closed R"
  shows "quasi_commute (restrict_SN R R) (restrict_SN supt R)" (is "quasi_commute ?r ?s")
proof -
  have "?s O ?r \<subseteq> ?r O (?r \<union> ?s)^*" 
  proof (rule subrelI)
    fix x y 
    assume "(x,y) \<in> ?s O ?r"
    from this obtain z where "(x,z) \<in> ?s" and "(z,y) \<in> ?r" by best
    hence "x \<rhd> z" and "(z,y) \<in> R" and "SN_on R {x}" unfolding restrict_SN_def supt_def by auto
    with `ctxt_closed R` obtain w where "(x,w) \<in> R" and "w \<rhd> y" using ctxt_closed_supt_R_imp_R_supt[where R = R and s = x and t = z and u = y] by auto
    from this and `SN_on R {x}` have "SN_on R {w}" using step_preserves_SN_on
      unfolding supt_supteq_conv by auto
    with `w \<rhd> y` have "(w,y) \<in> ?s" unfolding restrict_SN_def supt_def by auto
    with `(x,w) \<in> R` `SN_on R {x}` show "(x,y) \<in> ?r O (?r \<union> ?s)^*" unfolding restrict_SN_def by auto
  qed
  thus ?thesis unfolding quasi_commute_def .
qed

lemma ctxt_closed_imp_SN_restrict_SN_supt:
 assumes "ctxt_closed R"
 shows "SN (restrict_SN_supt R)"
proof -
  let ?r = "restrict_SN R R"
  let ?s = "restrict_SN (op \<rhd>) R"
  from `ctxt_closed R` have "quasi_commute ?r ?s"
    by (rule ctxt_closed_R_imp_restrict_SN_qc_supt)
  have "SN (op \<rhd>)" by (rule SN_supt)
  hence "SN ?s" unfolding restrict_SN_def SN_on_def by auto
  have "SN ?r" by (rule SN_restrict_SN_idemp)
  from `SN ?r` and `SN ?s` and `quasi_commute ?r ?s`
  show ?thesis unfolding restrict_SN_supt_def by  (rule quasi_commute_imp_SN)
qed

lemma SN_restrict_SN_supt_rstep: 
  shows "SN (restrict_SN_supt (rstep R))"
proof -
  have "ctxt_closed (rstep R)" by (rule ctxt_closed_rstep)
  thus ?thesis by (rule ctxt_closed_imp_SN_restrict_SN_supt)
qed

lemma nrrstep_imp_pos_term:
  "(Fun f ss,t) \<in> nrrstep R \<Longrightarrow>
    \<exists>i s. t = Fun f (ss[i:=s]) \<and> (ss!i,s) \<in> rstep R \<and> i < length ss"
proof -
 assume "(Fun f ss,t) \<in> nrrstep R"
 then obtain l r i ps \<sigma> where rstep_rps:"(Fun f ss,t) \<in> rstep_r_p_s R (l,r) (i<#ps) \<sigma>"
  unfolding nrrstep_def by auto
 then obtain C
 where "(l,r) \<in> R"
  and pos: "(i<#ps) \<in> poss (Fun f ss)"
  and C: "C = ctxt_of_pos_term (i<#ps) (Fun f ss)"
  and "C\<langle>l\<cdot>\<sigma>\<rangle> = Fun f ss"
  and t: "C\<langle>r\<cdot>\<sigma>\<rangle> = t"
 unfolding rstep_r_p_s_def Let_def by auto
 then obtain D where "C = More f (take i ss) D (drop (Suc i) ss)" by auto
 with t have t': "t = Fun f (take i ss @ (D\<langle>r\<cdot>\<sigma>\<rangle>) # drop (Suc i) ss)" by auto
 from rstep_rps have "(ss!i,t|_<i>) \<in> rstep_r_p_s R (l,r) ps \<sigma>"
  by (rule rstep_i_pos_imp_rstep_arg_i_pos)
 hence rstep:"(ss!i,t|_<i>) \<in> rstep R" by (auto simp: rstep_iff_rstep_r_p_s)
 hence "(C\<langle>ss!i\<rangle>,C\<langle>t|_<i>\<rangle>) \<in> rstep R" ..
 from pos have len: "i < length ss" by auto
 from len
 have "(take i ss @ (D\<langle>r\<cdot>\<sigma>\<rangle>) # drop (Suc i) ss)!i = D\<langle>r\<cdot>\<sigma>\<rangle>" by (auto simp: nth_append)
 with C t' have "t|_<i> = D\<langle>r\<cdot>\<sigma>\<rangle>" by auto
 with t' have "t = Fun f (take i ss @ (t|_<i>) # drop (Suc i) ss)" by auto
 with len have "t = Fun f (ss[i:=t|_<i>])" by (auto simp: upd_conv_take_nth_drop)
 with rstep len show "\<exists>i s. t = Fun f (ss[i:=s]) \<and> (ss!i,s) \<in> rstep R \<and> i < length ss" by auto
qed

lemma rstep_cases[consumes 1, case_names root nonroot]:
 "\<lbrakk>(s,t) \<in> rstep R; (s,t) \<in> rrstep R \<Longrightarrow> P; (s,t) \<in> nrrstep R \<Longrightarrow> P\<rbrakk> \<Longrightarrow> P"
by (auto simp: rstep_iff_rrstep_or_nrrstep)

lemma nrrstep_imp_rstep: "(s,t) \<in> nrrstep R \<Longrightarrow> (s,t) \<in> rstep R"
by (auto simp: rstep_iff_rrstep_or_nrrstep)

lemma nrrstep_imp_Fun: "(s,t) \<in> nrrstep R \<Longrightarrow> \<exists>f ss. s = Fun f ss"
proof -
 assume "(s,t) \<in> nrrstep R"
 then obtain i ps where "i<#ps \<in> poss s"
 unfolding nrrstep_def rstep_r_p_s_def Let_def by auto
 thus "\<exists>f ss. s = Fun f ss" by (cases s) auto
qed

lemma nrrstep_imp_subt_rstep:
 assumes "(s,t) \<in> nrrstep R"
 shows "\<exists>i. i < num_args s \<and> num_args s = num_args t \<and> (s|_<i>,t|_<i>) \<in> rstep R \<and> (\<forall>j. i \<noteq> j \<longrightarrow> s|_<j> = t|_<j>)"
proof -
 from `(s,t) \<in> nrrstep R` obtain f ss where "s = Fun f ss" using nrrstep_imp_Fun by blast
 with `(s,t) \<in> nrrstep R` have "(Fun f ss,t) \<in> nrrstep R" by simp
 then obtain i u where "t = Fun f (ss[i := u])" and "(ss!i,u) \<in> rstep R" and "i < length ss"
  using nrrstep_imp_pos_term by best
 from `s = Fun f ss` and `t = Fun f (ss[i := u])` have "num_args s = num_args t" by auto
 from `i < length ss` and `s = Fun f ss` have "i < num_args s" by auto
 from `s = Fun f ss` have "s|_<i> = (ss!i)" by auto
 from `t = Fun f (ss[i := u])` and `i < length ss` have "t|_<i> = u" by auto
 from `s = Fun f ss` and `t = Fun f (ss[i := u])`
 have  "\<forall>j. j \<noteq> i \<longrightarrow> s|_<j> = t|_<j>" by auto
 with `(ss!i,u) \<in> rstep R`
  and `i < num_args s`
  and `num_args s = num_args t`
  and `s|_<i> = (ss!i)`[symmetric] and `t|_<i> = u`[symmetric]
 show ?thesis by auto
qed

lemma finite_range:
  fixes f::"nat \<Rightarrow> 'a"
  assumes "finite (range f)"
  shows "\<exists>x. \<exists>\<^sub>\<infinity>i. f i = x"
proof (rule ccontr)
  assume "\<not>(\<exists>x. \<exists>\<^sub>\<infinity>i. f i = x)"
  hence "\<forall>x. \<exists>j. \<forall>i>j. f i \<noteq> x" unfolding INFM_nat by blast
  from choice[OF this] obtain j where neq: "\<forall>x. \<forall>i>j x. f i \<noteq> x" ..
  from finite_range_imageI[OF assms] have "finite (range (j \<circ> f))" by (simp add: comp_def)
  from finite_nat_bounded[OF this] obtain m where "range (j \<circ> f) \<subseteq> {..<m}" ..
  hence "j (f m) < m" by (auto simp: comp_def)
  with neq show False by blast
qed

lemma nrrstep_subt: assumes "(s, t) \<in> nrrstep R" shows "\<exists>u\<lhd>s. \<exists>v\<lhd>t. (u, v) \<in> rstep R"
proof -
  from assms obtain l r C \<sigma> where "(l, r) \<in> R" and "C \<noteq> \<box>"
    and s: "s = C\<langle>l\<cdot>\<sigma>\<rangle>" and t: "t = C\<langle>r\<cdot>\<sigma>\<rangle>" unfolding nrrstep_def' by best
  from `C \<noteq> \<box>` s have "s \<rhd> l\<cdot>\<sigma>" by auto
  moreover from `C \<noteq> \<box>` t have "t \<rhd> r\<cdot>\<sigma>" by auto
  moreover from `(l, r) \<in> R` have "(l\<cdot>\<sigma>, r\<cdot>\<sigma>) \<in> rstep R" by auto
  ultimately show ?thesis by auto
qed

lemma nrrstep_args:
  assumes "(s, t) \<in> nrrstep R"
  shows "\<exists>f ss ts. s = Fun f ss \<and> t = Fun f ts \<and> length ss = length ts
    \<and> (\<exists>j<length ss. (ss!j, ts!j) \<in> rstep R \<and> (\<forall>i<length ss. i \<noteq> j \<longrightarrow> ss!i = ts!i))"
proof -
  from assms obtain l r C \<sigma> where "(l, r) \<in> R" and "C \<noteq> \<box>"
    and s: "s = C\<langle>l\<cdot>\<sigma>\<rangle>" and t: "t = C\<langle>r\<cdot>\<sigma>\<rangle>" unfolding nrrstep_def' by best
  from `C \<noteq> \<box>` obtain f ss1 D ss2 where C: "C = More f ss1 D ss2" by (induct C) auto
  have "s = Fun f (ss1 @ D\<langle>l\<cdot>\<sigma>\<rangle> # ss2)" (is "_ = Fun f ?ss") by (simp add: s C)
  moreover have "t = Fun f (ss1 @ D\<langle>r\<cdot>\<sigma>\<rangle> # ss2)" (is "_ = Fun f ?ts") by (simp add: t C)
  moreover have "length ?ss = length ?ts" by simp
  moreover have "\<exists>j<length ?ss.
    (?ss!j, ?ts!j) \<in> rstep R \<and> (\<forall>i<length ?ss. i \<noteq> j \<longrightarrow> ?ss!i = ?ts!i)"
  proof -
    let ?j = "length ss1"
    have "?j < length ?ss" by simp
    moreover have "(?ss!?j, ?ts!?j) \<in> rstep R"
    proof -
      from `(l, r) \<in> R` have "(D\<langle>l\<cdot>\<sigma>\<rangle>, D\<langle>r\<cdot>\<sigma>\<rangle>) \<in> rstep R" by auto
      thus ?thesis by auto
    qed
    moreover have "\<forall>i<length ?ss. i \<noteq> ?j \<longrightarrow> ?ss!i = ?ts!i" (is "\<forall>i<length ?ss. _ \<longrightarrow> ?P i")
    proof (intro allI impI)
      fix i assume "i < length ?ss" and "i \<noteq> ?j"
      hence "i < length ss1 \<or> length ss1 < i" by auto
      thus "?P i"
      proof
        assume "i < length ss1" thus "?P i" by (auto simp: nth_append)
      next
        assume "i > length ss1" thus "?P i"
          using `i < length ?ss` by (auto simp: nth_Cons' nth_append)
      qed
    qed
    ultimately show ?thesis by best
  qed
  ultimately show ?thesis by auto
qed


lemma subterms_NF_imp_SN_on_nrrstep:
  assumes "\<forall>s\<lhd>t. s \<in> NF (rstep R)" shows "SN_on (nrrstep R) {t}"
proof
  fix S assume "S 0 \<in> {t}" and "\<forall>i. (S i, S (Suc i)) \<in> nrrstep R"
  hence "(t, S (Suc 0)) \<in> nrrstep R" by auto
  then obtain l r C \<sigma> where "(l, r) \<in> R" and "C \<noteq> \<box>" and t: "t = C\<langle>l\<cdot>\<sigma>\<rangle>" unfolding nrrstep_def' by auto
  then obtain f ss1 D ss2 where C: "C = More f ss1 D ss2" by (induct C) auto
  have "t \<rhd> D\<langle>l\<cdot>\<sigma>\<rangle>" unfolding t C by auto
  moreover have "D\<langle>l\<cdot>\<sigma>\<rangle> \<notin> NF (rstep R)"
  proof -
    from `(l, r) \<in> R` have "(D\<langle>l\<cdot>\<sigma>\<rangle>, D\<langle>r\<cdot>\<sigma>\<rangle>) \<in> rstep R" by auto
    thus ?thesis by auto
  qed
  ultimately show False using assms by simp
qed

lemma args_NF_imp_SN_on_nrrstep:
  assumes "\<forall>t\<in>set ts. t \<in> NF (rstep R)" shows "SN_on (nrrstep R) {Fun f ts}"
proof
  fix S assume "S 0 \<in> {Fun f ts}" and "\<forall>i. (S i, S (Suc i)) \<in> nrrstep R"
  hence "(Fun f ts, S (Suc 0)) \<in> nrrstep R"
    unfolding singletonD[OF `S 0 \<in> {Fun f ts}`, symmetric] by simp
  then obtain l r C \<sigma> where "(l, r) \<in> R" and "C \<noteq> \<box>" and "Fun f ts = C\<langle>l\<cdot>\<sigma>\<rangle>"
    unfolding nrrstep_def' by auto
  then obtain ss1 D ss2 where C: "C = More f ss1 D ss2" by (induct C) auto
  with `Fun f ts = C\<langle>l\<cdot>\<sigma>\<rangle>` have "D\<langle>l\<cdot>\<sigma>\<rangle> \<in> set ts" by auto
  moreover have "D\<langle>l\<cdot>\<sigma>\<rangle> \<notin> NF (rstep R)"
  proof -
    from `(l, r) \<in> R` have "(D\<langle>l\<cdot>\<sigma>\<rangle>, D\<langle>r\<cdot>\<sigma>\<rangle>) \<in> rstep R" by auto
    thus ?thesis by auto
  qed
  ultimately show False using assms by simp
qed

lemma rrstep_imp_rule_subst:
  assumes "(s,t) \<in> rrstep R"
  shows "\<exists>l r \<sigma>. (l,r) \<in> R \<and> (l\<cdot>\<sigma>) = s \<and> (r\<cdot>\<sigma>) = t"
proof -
  have "ctxt_of_pos_term \<epsilon> s = Hole" by auto
  obtain l r \<sigma> where "(s,t) \<in> rstep_r_p_s R (l,r) \<epsilon> \<sigma>" using assms unfolding rrstep_def by best
  hence "let C = ctxt_of_pos_term \<epsilon> s in \<epsilon> \<in> poss s \<and> (l,r) \<in> R \<and> C\<langle>l\<cdot>\<sigma>\<rangle> = s \<and> C\<langle>r\<cdot>\<sigma>\<rangle> = t" unfolding rstep_r_p_s_def by simp
  with `ctxt_of_pos_term \<epsilon> s = Hole` have "(l,r) \<in> R" and "l\<cdot>\<sigma> = s" and "r\<cdot>\<sigma> = t" unfolding Let_def by auto
  thus ?thesis by auto
qed

lemma nrrstep_preserves_root:
  assumes "(Fun f ss,t) \<in> nrrstep R" (is "(?s,t) \<in> nrrstep R") shows "\<exists>ts. t = (Fun f ts)"
using assms unfolding nrrstep_def rstep_r_p_s_def Let_def by auto

lemma nrrstep_equiv_root: assumes "(s,t) \<in> nrrstep R" shows "\<exists>f ss ts. s = Fun f ss \<and> t = Fun f ts"
proof -
 from assms obtain f ss where "s = Fun f ss" using nrrstep_imp_Fun by best
 with assms obtain ts where "t = Fun f ts" using nrrstep_preserves_root by best
 from `s = Fun f ss` and `t = Fun f ts` show ?thesis by best
qed

lemma nrrstep_reflects_root:
 assumes "(s,Fun g ts) \<in> nrrstep R" (is "(s,?t) \<in> nrrstep R")
 shows "\<exists>ss. s = (Fun g ss)"
proof -
 from assms obtain f ss ts' where "s = Fun f ss" and "Fun g ts = Fun f ts'" using nrrstep_equiv_root by best
 hence "f = g" by simp
 with `s = Fun f ss` have "s = Fun g ss" by simp
 thus ?thesis by auto
qed

lemma nrrsteps_preserve_root:
 assumes "(Fun f ss,t) \<in> (nrrstep R)^*"
 shows "\<exists>ts. t = (Fun f ts)"
using assms by induct (auto simp: nrrstep_preserves_root)

lemma nrrstep_Fun_imp_arg_rsteps:
   assumes "(Fun f ss,Fun f ts) \<in> nrrstep R" (is "(?s,?t) \<in> nrrstep R") and "i < length ss"
  shows "(ss!i,ts!i) \<in> (rstep R)^*"
proof -
 from assms have "<i> \<in> poss ?s" using ep_pos_term by simp
 from `(?s,?t) \<in> nrrstep R`
 obtain l r j ps \<sigma>
 where A: "let C = ctxt_of_pos_term (j<#ps) ?s in (j<#ps) \<in> poss ?s \<and> (l,r) \<in> R \<and> C\<langle>l\<cdot>\<sigma>\<rangle> = ?s \<and> C\<langle>r\<cdot>\<sigma>\<rangle> = ?t" unfolding nrrstep_def rstep_r_p_s_def by force
 let ?C = "ctxt_of_pos_term (j<#ps) ?s"
 from A have "(j<#ps) \<in> poss ?s" and "(l,r) \<in> R" and "?C\<langle>l\<cdot>\<sigma>\<rangle> = ?s" and "?C\<langle>r\<cdot>\<sigma>\<rangle> = ?t" using Let_def by auto
 have C: "?C = More f (take j ss) (ctxt_of_pos_term ps (ss!j)) (drop (Suc j) ss)" (is "?C = More f ?ss1 ?D ?ss2") by auto
 from `?C\<langle>l\<cdot>\<sigma>\<rangle> = ?s` have "?D\<langle>l\<cdot>\<sigma>\<rangle> = (ss!j)" by (auto simp: take_drop_imp_nth)
 from `(l,r) \<in> R` have "(l\<cdot>\<sigma>,r\<cdot>\<sigma>) \<in> (CS R)" by auto
 hence "(?D\<langle>l\<cdot>\<sigma>\<rangle>,?D\<langle>r\<cdot>\<sigma>\<rangle>) \<in> (CC(CS R))" and "(?C\<langle>l\<cdot>\<sigma>\<rangle>,?C\<langle>r\<cdot>\<sigma>\<rangle>) \<in> (CC(CS R))" by (auto simp only: ctxt_closure.ctxt_closureI)
 hence D_rstep: "(?D\<langle>l\<cdot>\<sigma>\<rangle>,?D\<langle>r\<cdot>\<sigma>\<rangle>) \<in> rstep R" and "(?C\<langle>l\<cdot>\<sigma>\<rangle>,?C\<langle>r\<cdot>\<sigma>\<rangle>) \<in> rstep R"
   unfolding rstep_CC_CS by auto
 from `?C\<langle>r\<cdot>\<sigma>\<rangle> = ?t` and C have "?t = Fun f (take j ss @ ?D\<langle>r\<cdot>\<sigma>\<rangle> # drop (Suc j) ss)" by auto
 hence ts: "ts = (take j ss @ ?D\<langle>r\<cdot>\<sigma>\<rangle> # drop (Suc j) ss)" by auto
 have "j = i \<or> j \<noteq> i" by simp
 from `j<#ps \<in> poss ?s` have "j < length ss" by simp
 hence "(take j ss @ ?D\<langle>r\<cdot>\<sigma>\<rangle> # drop (Suc j) ss) ! j = ?D\<langle>r\<cdot>\<sigma>\<rangle>" by (auto simp: nth_append_take)
 with ts have "ts!j = ?D\<langle>r\<cdot>\<sigma>\<rangle>" by auto
 have "j = i \<or> j \<noteq> i" by simp
 thus "(ss!i,ts!i) \<in> (rstep R)^*"
 proof
  assume "j = i"
  with `ts!j = ?D\<langle>r\<cdot>\<sigma>\<rangle>` and `?D\<langle>l\<cdot>\<sigma>\<rangle> = ss!j` and D_rstep show ?thesis by auto
 next
  assume "j \<noteq> i"
  with `i < length ss` and `j < length ss` have "(take j ss @ ?D\<langle>r\<cdot>\<sigma>\<rangle> # drop (Suc j) ss) ! i = ss!i" by (auto simp: nth_append_take_drop_is_nth_conv)
  with ts have "ts!i = ss!i" by auto
  thus ?thesis by auto
 qed
qed

lemma nrrstep_imp_arg_rsteps:
 assumes "(s,t) \<in> nrrstep R" and "i < num_args s" shows "(args s!i,args t!i) \<in> (rstep R)^*"
proof (cases s)
 fix x assume "s = Var x" thus ?thesis using assms by auto
next
 fix f ss assume "s = Fun f ss"
 hence "(Fun f ss,t) \<in> nrrstep R" using assms by simp
 then obtain ts where "t = Fun f ts" using nrrstep_preserves_root by best
 with `(s,t) \<in> nrrstep R` and `s = Fun f ss` have "(Fun f ss,Fun f ts) \<in> nrrstep R" by simp
 from  `s = Fun f ss` and `i < num_args s` have "i < length ss" by simp
 with `(Fun f ss,Fun f ts) \<in> nrrstep R`
   have "(ss!i,ts!i) \<in> (rstep R)^*" by (rule nrrstep_Fun_imp_arg_rsteps)
 with `s = Fun f ss` and `t = Fun f ts` show ?thesis by simp
qed

lemma nrrsteps_imp_rsteps: "(s,t) \<in> (nrrstep R)^* \<Longrightarrow> (s,t) \<in> (rstep R)^*"
proof (induct rule: rtrancl.induct)
 case (rtrancl_refl a) thus ?case by simp
next
 case (rtrancl_into_rtrancl a b c)
 hence IH: "(a,b) \<in> (rstep R)^*" and nrrstep: "(b,c) \<in> nrrstep R" by auto
 from nrrstep have "(b,c) \<in> rstep R" using nrrstep_imp_rstep by auto
 with IH show ?case by auto
qed

lemma nrrstep_Fun_preserves_num_args:
  assumes "(Fun f ss,Fun f ts) \<in> nrrstep R" (is "(?s,?t) \<in> nrrstep R")
  shows "length ss = length ts"
proof -
 from assms obtain l r i ps \<sigma>
   where "let C = ctxt_of_pos_term (i<#ps) ?s in (i<#ps) \<in> poss ?s \<and> (l,r) \<in> R \<and> C\<langle>l\<cdot>\<sigma>\<rangle> = ?s \<and> C\<langle>r\<cdot>\<sigma>\<rangle> = ?t" (is "let C = ?C in _")
 unfolding nrrstep_def rstep_r_p_s_def by force
 hence "(l,r) \<in> R" and Cl: "?C\<langle>l\<cdot>\<sigma>\<rangle> = ?s" and Cr: "?C\<langle>r\<cdot>\<sigma>\<rangle> = ?t" unfolding Let_def by auto
 have C: "?C = More f (take i ss) (ctxt_of_pos_term ps (ss!i)) (drop (Suc i) ss)" (is "?C = More f ?ss1 ?D ?ss2") by simp
 from C and Cl have s: "?s = Fun f (take i ss @ ?D\<langle>l\<cdot>\<sigma>\<rangle> # drop (Suc i) ss)" (is "?s = Fun f ?ss") by simp
 from C and Cr have t: "?t = Fun f (take i ss @ ?D\<langle>r\<cdot>\<sigma>\<rangle> # drop (Suc i) ss)" (is "?t = Fun f ?ts") by simp
 from s and t have ss: "ss = ?ss" and ts: "ts = ?ts" by auto
 have "length ?ss = length ?ts" by auto
 with ss and ts show ?thesis by simp
qed

lemma nrrstep_equiv_num_args:
 assumes "(s,t) \<in> nrrstep R" shows "num_args s = num_args t"
proof -
 from assms obtain f ss ts where s:"s = Fun f ss" and t:"t = Fun f ts" using nrrstep_equiv_root by best
 with assms have "(Fun f ss,Fun f ts) \<in> nrrstep R" by simp
 hence "length ss = length ts" by (rule nrrstep_Fun_preserves_num_args)
 with s and t show ?thesis by auto
qed

lemma nrrsteps_equiv_num_args:
 assumes "(s,t) \<in> (nrrstep R)^*" shows "num_args s = num_args t"
using assms by (induct, auto simp: nrrstep_equiv_num_args)
 
lemma nrrstep_preserves_num_args:
 assumes "(s,t) \<in> nrrstep R" and "i < num_args s" shows "i < num_args t"
proof (cases s)
 fix x assume "s = Var x" thus ?thesis using assms by auto
next
 fix f ss assume "s = Fun f ss"
 with assms obtain ts where "t = Fun f ts" using nrrstep_preserves_root by best
 from `(s,t) \<in> nrrstep R` have "length ss = length ts" unfolding `s = Fun f ss` and `t = Fun f ts` by (rule nrrstep_Fun_preserves_num_args)
 with assms and `s = Fun f ss` and `t = Fun f ts` show ?thesis by auto
qed

lemma nrrstep_reflects_num_args:
 assumes "(s,t) \<in> nrrstep R" and "i < num_args t" shows "i < num_args s"
proof (cases t)
 fix x assume "t = Var x" thus ?thesis using assms by auto
next
 fix g ts assume "t = Fun g ts"
 with assms obtain ss where "s = Fun g ss" using nrrstep_reflects_root by best
 from `(s,t) \<in> nrrstep R` have "length ss = length ts" unfolding `s = Fun g ss` and `t = Fun g ts` by (rule nrrstep_Fun_preserves_num_args)
 with assms and `s = Fun g ss` and `t = Fun g ts` show ?thesis by auto
qed

lemma nrrsteps_imp_arg_rsteps: 
 assumes "(s,t) \<in> (nrrstep R)^*" and "i < num_args s"
 shows "(args s!i,args t!i) \<in> (rstep R)^*"
using assms
proof (induct rule: rtrancl.induct)
 case (rtrancl_refl a) thus ?case by auto
next
 case (rtrancl_into_rtrancl a b c)
 hence IH: "(args a!i,args b!i) \<in> (rstep R)^*" by auto
 from `(a,b) \<in> (nrrstep R)^*` and `i < num_args a` have "i < num_args b" by induct (auto simp: nrrstep_preserves_num_args)
 with `(b,c) \<in> nrrstep R`
   have "(args b!i,args c!i) \<in> (rstep R)^*" by (rule nrrstep_imp_arg_rsteps)
 with IH show ?case by simp
qed

lemma nrrsteps_imp_eq_root_arg_rsteps:
  assumes steps: "(s,t) \<in> (nrrstep R)^*"
  shows "root s = root t \<and> (\<forall>i<num_args s. (s |_ <i>, t |_ <i> ) \<in> (rstep R)^*)" 
proof (cases s)
  case (Var x)
  have "s = t" using steps unfolding Var
  proof (induct)
    case (step y z)
    from nrrstep_imp_Fun[OF step(2)] step(3) have False by auto
    thus ?case ..
  qed simp
  thus ?thesis by auto
next
  case (Fun f ss)
  from nrrsteps_equiv_num_args[OF steps]
       nrrsteps_imp_arg_rsteps[OF steps]
       nrrsteps_preserve_root[OF steps[unfolded Fun]]
  show ?thesis unfolding Fun by auto
qed

definition
  wf_trs :: "('f, 'v) trs \<Rightarrow> bool"
where
  "wf_trs R = (\<forall>l r. (l,r) \<in> R \<longrightarrow> (\<exists>f ts. l = Fun f ts) \<and> vars_term r \<subseteq> vars_term l)"

lemma wf_trs_imp_lhs_Fun:
  "wf_trs R \<Longrightarrow> (l,r) \<in> R \<Longrightarrow> \<exists>f ts. l = Fun f ts"
unfolding wf_trs_def by blast

lemma rstep_imp_Fun:
 assumes "wf_trs R"
 shows "(s, t) \<in> rstep R \<Longrightarrow> \<exists>f ss. s = Fun f ss"
proof -
  assume "(s, t) \<in> rstep R"
  then obtain l r p \<sigma> where "(s, t) \<in> rstep_r_p_s R (l, r) p \<sigma>"
    unfolding rstep_iff_rstep_r_p_s by best
  hence "(ctxt_of_pos_term p s)\<langle>l \<cdot> \<sigma>\<rangle> = s" (is "?C\<langle>l \<cdot> \<sigma>\<rangle> = s") and "(l, r) \<in> R"
    unfolding rstep_r_p_s_def Let_def by auto
  with `wf_trs R` obtain f ss where "l = Fun f ss" unfolding wf_trs_def by best
  from `?C\<langle>l \<cdot> \<sigma>\<rangle> = s` obtain C where "C\<langle>l \<cdot> \<sigma>\<rangle> = s" by best
  hence "C\<langle>Fun f (map (\<lambda>x. x \<cdot> \<sigma>) ss)\<rangle> = s" unfolding `l = Fun f ss` by simp
  thus "\<exists>f ss. s = Fun f ss" by (induct C) auto
qed


lemma SN_on_imp_SN_on_subt:
  assumes "SN_on (rstep R) {t}" shows "\<forall>s\<unlhd>t. SN_on (rstep R) {s}"
proof (rule ccontr)
  assume "\<not>(\<forall>s\<unlhd>t. SN_on (rstep R) {s})"
  then obtain s where "t \<unrhd> s" and "\<not> SN_on (rstep R) {s}" by auto
  then obtain S where "S 0 = s" and chain: "chain (rstep R) S" by auto
  from `t \<unrhd> s` obtain C where t: "t = C\<langle>s\<rangle>" by auto
  let ?S = "\<lambda>i. C\<langle>S i\<rangle>"
  from `S 0 = s` have "?S 0 = t" by (simp add: t)
  moreover from chain have "chain (rstep R) ?S" by blast
  ultimately have "\<not> SN_on (rstep R) {t}" by best
  with assms show False by simp
qed

lemma not_SN_on_subt_imp_not_SN_on:
  assumes "\<not> SN_on (rstep R) {t}" and "s \<unrhd> t"
  shows "\<not> SN_on (rstep R) {s}"
using assms SN_on_imp_SN_on_subt by blast

lemma SN_on_instance_imp_SN_on_var: 
  assumes "SN_on (rstep R) {t \<cdot> \<sigma>}" and "x \<in> vars_term t"
  shows "SN_on (rstep R) {Var x \<cdot> \<sigma>}"
proof -
  from assms have "t \<unrhd> Var x" by auto
  hence "t\<cdot>\<sigma> \<unrhd> (Var x)\<cdot>\<sigma>" by (rule supteq_stable)
  with SN_on_imp_SN_on_subt and assms show ?thesis by best
qed

lemma var_imp_var_of_arg:
  assumes "x \<in> vars_term (Fun f ss)" (is "x \<in> vars_term ?s")
  shows "\<exists>i < num_args (Fun f ss). x \<in> vars_term (ss!i)"
proof -
  from assms have "x \<in> \<Union>(set (map vars_term ss))" by simp
  hence "x \<in> (\<Union>i<length ss. vars_term(ss!i))" unfolding set_conv_nth by auto
  hence "\<exists>i<length ss. x \<in> vars_term(ss!i)" using UN_iff by best
  then obtain i where "i < length ss" and "x \<in> vars_term(ss!i)" by auto
  hence "i < num_args ?s" by simp
  with `x \<in> vars_term(ss!i)` show ?thesis by auto
qed

lemma subt_instance_and_not_subst_imp_subt:
  "s\<cdot>\<sigma> \<unrhd> t \<Longrightarrow> \<forall>x \<in> vars_term s. \<not>((Var x)\<cdot>\<sigma> \<unrhd> t) \<Longrightarrow> \<exists>u. s \<unrhd> u \<and> t = u\<cdot>\<sigma>"
proof (induct s arbitrary: t rule: term_induct)
 case (Var x) thus ?case by auto
next
 case (Fun f ss)
 from `Fun f ss\<cdot>\<sigma> \<unrhd> t` have "(Fun f ss\<cdot>\<sigma> = t) \<or> (Fun f ss\<cdot>\<sigma> \<rhd> t)" by auto
 thus ?case
 proof
  assume "Fun f ss\<cdot>\<sigma> = t" with Fun show ?thesis by auto
 next
  assume "Fun f ss\<cdot>\<sigma> \<rhd> t"
  hence "Fun f (map (\<lambda>t. t\<cdot>\<sigma>) ss) \<rhd> t" by simp
  hence "\<exists>s \<in> set (map (\<lambda>t. t\<cdot>\<sigma>) ss). s \<unrhd> t" (is "\<exists>s \<in> set ?ss'. s \<unrhd> t") by (rule supt_Fun_imp_arg_supteq)
  then obtain s' where "s' \<in> set ?ss'" and "s' \<unrhd> t" by best
  hence "\<exists>i < length ?ss'. ?ss'!i = s'" using in_set_conv_nth[where ?x = "s'"] by best
  then obtain i where "i < length ?ss'" and "?ss'!i = s'" by best
  hence "?ss'!i = (ss!i)\<cdot>\<sigma>" by auto
  from `?ss'!i = s'` have "s' = (ss!i)\<cdot>\<sigma>" unfolding `?ss'!i = (ss!i)\<cdot>\<sigma>` by simp
  from `s' \<unrhd> t` have "(ss!i)\<cdot>\<sigma> \<unrhd> t" unfolding `s' = (ss!i)\<cdot>\<sigma>` by simp
  with `i < length ?ss'` have "(ss!i) \<in> set ss" by auto
  with `(ss!i)\<cdot>\<sigma> \<unrhd> t` have "\<exists>s \<in> set ss. s\<cdot>\<sigma> \<unrhd> t" by best
  then obtain s where "s \<in> set ss" and "s\<cdot>\<sigma> \<unrhd> t" by best
  with Fun have "\<forall>x \<in> vars_term s. \<not>((Var x)\<cdot>\<sigma> \<unrhd> t)" by force
  from Fun 
  have IH: "s \<in> set ss \<longrightarrow> (\<forall>v. s\<cdot>\<sigma> \<unrhd> v \<longrightarrow> (\<forall>x \<in> vars_term s. \<not> Var x\<cdot>\<sigma> \<unrhd> v) \<longrightarrow> (\<exists>u. s \<unrhd> u \<and> v = u\<cdot>\<sigma>))"
  by auto
  with `s \<in> set ss` have "!!v. s\<cdot>\<sigma> \<unrhd> v \<Longrightarrow> (\<forall>x \<in> vars_term s. \<not> Var x\<cdot>\<sigma> \<unrhd> v) \<longrightarrow> (\<exists>u. s \<unrhd> u \<and> v = u\<cdot>\<sigma>)"
  by simp
  with `s\<cdot>\<sigma> \<unrhd> t` have "(\<forall>x \<in> vars_term s. \<not> Var x\<cdot>\<sigma> \<unrhd> t) \<longrightarrow> (\<exists>u. s \<unrhd> u \<and> t = u\<cdot>\<sigma>)" by simp
  with `\<forall>x \<in> vars_term s. \<not> Var x\<cdot>\<sigma> \<unrhd> t` obtain u where "s \<unrhd> u" and "t = u\<cdot>\<sigma>" by best
  with `s \<in> set ss` have "Fun f ss \<unrhd> u" by auto
  with `t = u\<cdot>\<sigma>` show ?thesis by best
 qed
qed

lemma SN_imp_SN_subt_gen:
  assumes ctxt: "ctxt_closed R"
  and "SN_on R {s}" and "s \<unrhd> t"
  shows "SN_on R {t}"
  using assms by (rule restrict_SN_imp_SN_subt)

lemma SN_imp_SN_subt:
  "SN_on (rstep R) {s} \<Longrightarrow> s \<unrhd> t \<Longrightarrow> SN_on (rstep R) {t}"
  by (rule SN_imp_SN_subt_gen[OF ctxt_closed_rstep])

lemma subterm_preserves_SN_gen:
  assumes ctxt: "ctxt_closed R"
  and SN: "SN_on R {t}" and supt: "t \<rhd> s"
  shows "SN_on R {s}"
proof -
  from supt have "t \<unrhd> s" by auto
  thus ?thesis using SN_imp_SN_subt_gen[OF ctxt SN, of s] by simp
qed

lemma subterm_preserves_SN:
  "SN_on (rstep R) {t} \<Longrightarrow> t \<rhd> s \<Longrightarrow> SN_on (rstep R) {s}"
  by (rule subterm_preserves_SN_gen[OF ctxt_closed_rstep])


lemma SN_on_r_imp_SN_on_supt_union_r:
  assumes ctxt: "ctxt_closed R"
  and "SN_on R T"
  shows "SN_on (supt \<union> R) T" (is "SN_on ?S T")
proof (rule ccontr)
  assume "\<not> SN_on ?S T"
  then obtain s where ini: "s 0 : T" and chain: "chain ?S s"
    unfolding SN_on_def by auto
  have SN: "\<forall>i. SN_on R {s i}"
  proof
    fix i show "SN_on R {s i}"
    proof (induct i)
      case 0 show ?case using assms using `s 0 \<in> T` and SN_on_subset2[of "{s 0}" T R] by simp
    next
      case (Suc i)
      from chain have "(s i, s (Suc i)) \<in> ?S" by simp
      thus ?case
      proof
        assume "(s i, s (Suc i)) \<in> supt"
        from subterm_preserves_SN_gen[OF ctxt Suc this] show ?thesis .
      next
        assume "(s i, s (Suc i)) \<in> R"
        from step_preserves_SN_on[OF this Suc] show ?thesis .
      qed
    qed
  qed
  have "\<not> (\<exists>S. \<forall>i. (S i, S (Suc i)) \<in> restrict_SN_supt R)"
    using ctxt_closed_imp_SN_restrict_SN_supt[OF ctxt] unfolding SN_defs by auto
  moreover have "\<forall>i. (s i, s (Suc i)) \<in> restrict_SN_supt R"
  proof
    fix i
    from SN have SN: "SN_on R {s i}" by simp
    from chain have "(s i, s (Suc i)) \<in> supt \<union> R" by simp
    thus "(s i, s (Suc i)) \<in> restrict_SN_supt R"
      unfolding restrict_SN_supt_def restrict_SN_def using SN by auto
  qed
  ultimately show False by auto
qed

lemma SN_on_rstep_imp_SN_on_supt_union_rstep:
  "SN_on (rstep R) T \<Longrightarrow> SN_on (supt \<union> rstep R) T"
  by (rule SN_on_r_imp_SN_on_supt_union_r[OF ctxt_closed_rstep])

lemma SN_supt_r_trancl:
  assumes ctxt: "ctxt_closed R"
  and a: "SN R"
  shows "SN ((supt \<union> R)^+)"
proof -
  have "SN (supt \<union> R)"
    using SN_on_r_imp_SN_on_supt_union_r[OF ctxt, of UNIV]
    and a by force
  thus "SN ((supt \<union> R)^+)" by (rule SN_imp_SN_trancl)
qed

lemma SN_supt_rstep_trancl:
  "SN (rstep R) \<Longrightarrow> SN ((supt \<union> rstep R)^+)"
  by (rule SN_supt_r_trancl[OF ctxt_closed_rstep])

lemma SN_args_imp_SN_subt:
  assumes "\<And>i. i < num_args (Fun f ss) \<Longrightarrow> SN_on (rstep R) {ss ! i}"
  shows "\<forall>s. (Fun f ss) \<rhd> s \<longrightarrow> SN_on (rstep R) {s}"
proof (intro allI impI)
  fix s assume "Fun f ss \<rhd> s" (is "?t \<rhd> s")
  hence "\<exists>u \<in> set ss. u \<unrhd> s" using supt_Fun_imp_arg_supteq by best
  then obtain u where "u \<in> set ss" and "u \<unrhd> s" by best
  hence "\<exists>i < length ss. ss!i = u" unfolding in_set_conv_nth by simp
  then obtain i where "i < length ss" and "ss!i = u" by best
  hence "i < num_args ?t" by simp
  with assms have "SN_on (rstep R) {ss!i}" by simp
  with `u \<unrhd> s` show "SN_on (rstep R) {s}" unfolding `ss!i = u`[symmetric]
    using SN_imp_SN_subt[of R "ss!i"] by auto
qed

lemma SN_imp_SN_arg_gen:
  assumes ctxt: "ctxt_closed R"
  and SN: "SN_on R {Fun f ts}" and arg: "t \<in> set ts" shows "SN_on R {t}"
proof -
  from arg have "Fun f ts \<unrhd> t" by auto
  with SN show ?thesis by (rule SN_imp_SN_subt_gen[OF ctxt])
qed


lemma SN_imp_SN_arg:
  "SN_on (rstep R) {Fun f ts} \<Longrightarrow> t \<in> set ts \<Longrightarrow> SN_on (rstep R) {t}"
  by (rule SN_imp_SN_arg_gen[OF ctxt_closed_rstep])

lemma SNinstance_imp_SN:
  assumes "SN_on (rstep R) {t \<cdot> \<sigma>}" 
  shows "SN_on (rstep R) {t}"
proof(rule ccontr)
  assume "\<not> SN_on (rstep R) {t}"
  hence "\<not> SN_on (rstep R) {t \<cdot> \<sigma>}" 
  proof (simp add: SN_on_def)
    assume "\<exists>f. f 0 = t \<and> (\<forall> i. (f i, f (Suc i)) \<in> rstep R)"
    from this obtain f where prem: "f 0 = t \<and> (\<forall> i. (f i, f (Suc i)) \<in> rstep R)" by auto
    let ?g = "\<lambda> i. (f i) \<cdot> \<sigma>"
    from prem have "?g 0 = t \<cdot> \<sigma> \<and> (\<forall> i. (?g i, ?g (Suc i)) \<in> rstep R)" using subst_closed_rstep 
      unfolding subst_closed_def by auto 
    thus "\<exists> f. f 0 = t \<cdot> \<sigma> \<and> (\<forall> i. (f i, f (Suc i)) \<in> rstep R)" by auto
  qed
  with assms show "False" by simp
qed

lemma rstep_imp_C_s_r:
  assumes "(s,t) \<in> rstep R"
  shows "\<exists>C \<sigma> l r. (l,r) \<in> R \<and> s = C\<langle>l\<cdot>\<sigma>\<rangle> \<and> t = C\<langle>r\<cdot>\<sigma>\<rangle>"
proof -
 from assms obtain l r p \<sigma> where step:"(s,t) \<in> rstep_r_p_s R (l,r) p \<sigma>"
  unfolding rstep_iff_rstep_r_p_s by best
 let ?C = "ctxt_of_pos_term p s"
 from step have "p \<in> poss s" and "(l,r) \<in> R" and "?C\<langle>l\<cdot>\<sigma>\<rangle> = s" and "?C\<langle>r\<cdot>\<sigma>\<rangle> = t"
  unfolding rstep_r_p_s_def Let_def by auto
 hence "(l,r) \<in> R \<and> s = ?C\<langle>l\<cdot>\<sigma>\<rangle> \<and> t = ?C\<langle>r\<cdot>\<sigma>\<rangle>" by auto
 thus ?thesis by force
qed

lemma SN_Var:
 assumes "wf_trs R" shows "SN_on (rstep R) {Var x}"
proof (rule ccontr)
  assume "\<not> ?thesis"
  then obtain S where [symmetric]: "S 0 = Var x"
    and chain: "chain (rstep R) S" by auto
  hence "(Var x, S (Suc 0)) \<in> rstep R" by force
  then obtain C l r \<sigma> where "(l, r) \<in> R" and "Var x = C\<langle>l \<cdot> \<sigma>\<rangle>" by best
  hence "Var x = l \<cdot> \<sigma>" by (induct C) simp_all
  then obtain y where "l = Var y" by (induct l) simp_all
  with assms and `(l, r) \<in> R` show False unfolding wf_trs_def by blast
qed

fun map_rule :: "('f \<Rightarrow> 'g) \<Rightarrow> ('f, 'v) rule \<Rightarrow> ('g, 'v) rule"
where
  "map_rule fg lr = (map_term fg (fst lr), map_term fg (snd lr))"

fun map_trs :: "('f \<Rightarrow> 'g) \<Rightarrow> ('f, 'v) trs \<Rightarrow> ('g, 'v) trs"
where
  "map_trs fg R = map_rule fg ` R"

lemma map_trs_union: "map_trs fg (R \<union> S) = map_trs fg R \<union> map_trs fg S"
  unfolding map_trs.simps by auto

lemma wf_trs_map_trs[simp]: "wf_trs (map_trs f R) = wf_trs R"
unfolding wf_trs_def
proof (rule iffI, intro allI impI)
  fix l r
  assume "\<forall>l r. (l, r) \<in> map_trs f R \<longrightarrow> (\<exists>f ts. l = Fun f ts) \<and> vars_term r \<subseteq> vars_term l" and "(l, r) \<in> R"
  thus "(\<exists>f ts. l = Fun f ts) \<and> vars_term r \<subseteq> vars_term l" by (cases l, force+)
next
  assume ass: " \<forall>l r. (l, r) \<in> R \<longrightarrow> (\<exists>f ts. l = Fun f ts) \<and> vars_term r \<subseteq> vars_term l"
  show "\<forall>l r. (l, r) \<in> map_trs f R \<longrightarrow> (\<exists>f ts. l = Fun f ts) \<and> vars_term r \<subseteq> vars_term l"
  proof (intro allI impI)
    fix l r
    assume "(l, r) \<in> map_trs f R"
    with ass 
    show "(\<exists>f ts. l = Fun f ts) \<and> vars_term r \<subseteq> vars_term l"
      by (cases l, force+)
  qed
qed

lemma map_trs_comp: "map_trs fg (map_trs gh R) = map_trs (fg o gh) R"
proof -  
  have mr: "map_rule (fg o gh) = map_rule fg o map_rule gh"
    by (rule ext, auto simp: map_term_comp)
  thus ?thesis
    by (auto simp: map_term_comp image_compose[symmetric] mr)
qed

lemma map_trs_mono: assumes "R \<subseteq> R'" shows "map_trs fg R \<subseteq> map_trs fg R'"
  using assms by auto

lemma map_trs_power_mono: 
  fixes R R' :: "('f,'v)trs" and fg :: "'f \<Rightarrow> 'f"
  assumes "R \<subseteq> R'" shows "((map_trs fg)^^n) R \<subseteq> ((map_trs fg)^^n) R'"
using assms by (induct n, simp, auto)

declare map_trs.simps[simp del]


lemma rstep_induct[case_names IH, induct set: "rstep"]:
  assumes "(s, t) \<in> rstep R"
    and IH: "\<And>C \<sigma> l r. (l, r) \<in> R \<Longrightarrow> P C\<langle>l \<cdot> \<sigma>\<rangle> C\<langle>r \<cdot> \<sigma>\<rangle>"
  shows "P s t"
proof -
  from `(s, t) \<in> rstep R` obtain C \<sigma> l r
    where "(l, r) \<in> R" and s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>" by auto
  from IH[OF `(l, r) \<in> R`] show ?thesis unfolding s t .
qed
text {*
We use @{thm[source] rstep_induct} as the default induction rule for @{term rstep}.
*}

lemma rstep_imp_map_rstep:
  assumes "(s,t) \<in> rstep R" 
  shows "(map_term fg s, map_term fg t) \<in> rstep (map_trs fg R)"
using assms proof (induct)
  case (IH C \<sigma> l r)
  hence "(map_term fg l, map_term fg r) \<in> map_trs fg R" (is "(?l, ?r) \<in> ?R") 
    unfolding map_trs.simps by force
  hence "(?l, ?r) \<in> rstep ?R" ..
  hence "(?l \<cdot> map_subst fg \<sigma>, ?r \<cdot> map_subst fg \<sigma>) \<in> rstep ?R" ..
  thus ?case by auto
qed

lemma rsteps_imp_map_rsteps: assumes "(s,t) \<in> (rstep R)^*"
  shows "(map_term fg s, map_term fg t) \<in> (rstep (map_trs fg R))^*"
using assms
proof (induct, clarify)
  case (step y z)
  hence "(map_term fg y, map_term fg z) \<in> rstep (map_trs fg R)" using rstep_imp_map_rstep 
    by (auto simp: map_trs.simps)
  with step show ?case by auto
qed

lemma SN_map_imp_SN: 
  assumes SN: "SN_on (rstep (map_trs fg R)) {map_term fg t}"
  shows "SN_on (rstep R) {t}"
proof (rule ccontr)
  assume "\<not> SN_on (rstep R) {t}"
  from this obtain f where cond: "f 0 = t \<and> (\<forall> i. (f i, f (Suc i)) \<in> rstep R)" 
  unfolding SN_on_def by auto
  obtain g where g: "g = (\<lambda> i. map_term fg (f i))" by auto
  with cond have cond2: "g 0 = map_term fg t \<and> (\<forall> i. (g i, g (Suc i)) \<in> rstep (map_trs fg R))"
    using rstep_imp_map_rstep[where fg = fg] by blast
  from SN
    have "\<not> (\<exists> g. (g 0 = map_term fg t \<and> (\<forall> i. (g i, g (Suc i)) \<in> rstep (map_trs fg R))))" 
    unfolding SN_on_def by auto
  with cond2 show False  by auto
qed

lemma rstep_iff_map_rstep:
  assumes "inj fg"
  shows "(s, t) \<in> rstep R \<longleftrightarrow> (map_term fg s, map_term fg t) \<in> rstep (map_trs fg R)"
proof
  assume "(s,t) \<in> rstep R" 
  thus "(map_term fg s, map_term fg t) \<in> rstep(map_trs fg R)"
    by (rule rstep_imp_map_rstep)
next
  assume "(map_term fg s, map_term fg t) \<in> rstep (map_trs fg R)" 
  hence "(map_term fg s, map_term fg t) \<in> CC(CS(map_trs fg R))"
    by (simp add: rstep_CC_CS)
  then obtain C u v where "(u,v) \<in> CS(map_trs fg R)" and "C\<langle>u\<rangle> = map_term fg s"
    and "C\<langle>v\<rangle> = map_term fg t"
    by (cases rule: ctxt_closure.cases) force
  then obtain \<sigma> w x where "(w,x) \<in> map_trs fg R" and "w\<cdot>\<sigma> = u" and "x\<cdot>\<sigma> = v"
    by (cases rule: subst_closure.cases) force
  then obtain l r where "w = map_term fg l" and "x = map_term fg r"
    and "(l,r) \<in> R" by (auto simp: map_trs.simps)
  have ps: "C\<langle>map_term fg l \<cdot> \<sigma>\<rangle> = map_term fg s" and pt: "C\<langle>map_term fg r \<cdot> \<sigma>\<rangle> = map_term fg t"
    unfolding `w = map_term fg l`[symmetric] `x = map_term fg r`[symmetric] `w\<cdot>\<sigma> = u` `x\<cdot>\<sigma> = v`
  `C\<langle>u\<rangle> = map_term fg s` `C\<langle>v\<rangle> = map_term fg t` by auto
  let ?gf = "the_inv fg"
  let ?C = "map_ctxt ?gf C"
  let ?\<sigma> = "map_subst ?gf \<sigma>"
  have gffg: "?gf \<circ> fg = id" using the_inv_f_f[OF assms] by (intro ext, auto)
  from ps and pt have "s = map_term ?gf (C\<langle>map_term fg l \<cdot> \<sigma>\<rangle>)"
    and "t = map_term ?gf (C\<langle>map_term fg r \<cdot> \<sigma>\<rangle>)" by (auto simp: map_term_comp gffg)
  hence s: "s = ?C\<langle>map_term ?gf (map_term fg l \<cdot> \<sigma>)\<rangle>"
    and t: "t = ?C\<langle>map_term ?gf (map_term fg r \<cdot> \<sigma>)\<rangle>" using map_ctxt_distrib by auto
  from s have "s = ?C\<langle>l\<cdot>?\<sigma>\<rangle>" by (simp add: map_term_comp gffg)
  from t have "t = ?C\<langle>r\<cdot>?\<sigma>\<rangle>" by (simp add: map_term_comp gffg)
  from `(l, r) \<in> R` have "(l\<cdot>?\<sigma>, r\<cdot>?\<sigma>) \<in> CS R" by (rule subst_closureI)
  hence "(?C\<langle>l\<cdot>?\<sigma>\<rangle>,?C\<langle>r\<cdot>?\<sigma>\<rangle>) \<in> CC(CS R)" by (rule ctxt_closureI)
  thus "(s,t) \<in> rstep R" unfolding `s = ?C\<langle>l\<cdot>?\<sigma>\<rangle>` `t = ?C\<langle>r\<cdot>?\<sigma>\<rangle>` by (simp add: rstep_CC_CS)
qed

lemma cs_subseteq_cs_ext1:
  assumes "(s, t) \<in> CS R"
  shows "(s, t) \<in> CS (R \<union> S)"
proof -
 from assms obtain \<sigma> u v 
 where R:"(u,v) \<in> R" and s: "s = u\<cdot>\<sigma>" and t:"t = v\<cdot>\<sigma>" by (cases rule: subst_closure.cases) auto
 from R have "(u,v) \<in> (R \<union> S)" ..
 thus ?thesis unfolding s t by (rule subst_closureI)
qed

lemma cs_subseteq_cs_ext2:
 assumes "(s,t) \<in> CS S"
 shows "(s,t) \<in> CS(R \<union> S)"
proof -
 have comm:"(S \<union> R) = (R \<union> S)" by auto
 from assms have "(s,t) \<in> CS(S \<union> R)" by (rule cs_subseteq_cs_ext1)
 thus ?thesis unfolding comm .
qed

lemma cs_subseteq_cs_ext1': "CS R \<subseteq> CS (R \<union> S)"
by (intro subrelI) (rule cs_subseteq_cs_ext1)

lemma cs_union_distr: "CS(R \<union> S) = (CS R) \<union> (CS S)"
proof
 show "CS (R \<union> S) \<subseteq> ((CS R) \<union> (CS S))"
 proof (intro subrelI)
  fix s t assume "(s,t) \<in> CS(R \<union> S)"
  then obtain \<sigma> u v where RS:"(u,v) \<in> (R \<union> S)" and s:"s = u\<cdot>\<sigma>" and t:"t = v\<cdot>\<sigma>"
   by (cases rule: subst_closure.cases) auto
  from RS have "(u,v) \<in> R \<or> (u,v) \<in> S" by simp
  hence "(s,t) \<in> CS R \<or> (s,t) \<in> CS S"
  proof
   assume "(u,v) \<in> R" thus ?thesis unfolding s t using subst_closureI by auto
  next
   assume "(u,v) \<in> S" thus ?thesis unfolding s t using subst_closureI by auto
  qed
  thus "(s,t) \<in> CS R \<union> CS S" by simp
 qed
next
 show "(CS R) \<union> (CS S) \<subseteq> CS (R \<union> S)"
 proof (intro subrelI)
  fix s t assume "(s,t) \<in> CS R \<union> CS S"
  hence "(s,t) \<in> CS R \<or> (s,t) \<in> CS S" by simp
  thus "(s,t) \<in> CS(R \<union> S)"
  proof (rule disjE)
   assume "(s,t) \<in> CS R" thus ?thesis by (rule cs_subseteq_cs_ext1)
  next
   assume "(s,t) \<in> CS S" thus ?thesis by (rule cs_subseteq_cs_ext2)
  qed
 qed
qed

lemma cc_subseteq_cc_ext1:
 assumes "(s,t) \<in> CC R"
 shows "(s,t) \<in> CC(R \<union> S)"
proof -
 from assms obtain C u v
  where R:"(u,v) \<in> R" and s:"s = C\<langle>u\<rangle>" and t:"t = C\<langle>v\<rangle>" by (cases rule: ctxt_closure.cases) auto
 from R have "(u,v) \<in> (R \<union> S)" ..
 thus ?thesis unfolding s t by (rule ctxt_closureI)
qed

lemma cc_subseteq_cc_ext2:
 assumes "(s,t) \<in> CC S"
 shows "(s,t) \<in> CC(R \<union> S)"
proof -
 have comm:"(S \<union> R) = (R \<union> S)" by auto
 from assms have "(s,t) \<in> CC(S \<union> R)" by (rule cc_subseteq_cc_ext1)
 thus ?thesis unfolding comm .
qed

lemma cc_union_distr: "CC(R \<union> S) = (CC R) \<union> (CC S)"
proof
 show "CC (R \<union> S) \<subseteq> (CC R \<union> CC S)"
 proof (rule subrelI)
  fix s t assume "(s,t) \<in> CC(R \<union> S)"
  then obtain C u v where RS:"(u,v) \<in> (R \<union> S)" and s:"s = C\<langle>u\<rangle>" and t:"t = C\<langle>v\<rangle>"
   by (cases rule: ctxt_closure.cases) auto
  from RS have "(u,v) \<in> R \<or> (u,v) \<in> S" by simp
  thus "(s,t) \<in> CC R \<union> CC S"
  proof
   assume "(u,v) \<in> R" thus ?thesis unfolding s t using ctxt_closureI by auto
  next
   assume "(u,v) \<in> S" thus ?thesis unfolding s t using ctxt_closureI by auto
  qed
 qed
next
 show "(CC R) \<union> (CC S) \<subseteq> CC (R \<union> S)"
 proof (rule subrelI)
  fix s t assume "(s,t) \<in> CC R \<union> CC S"
  hence "(s,t) \<in> CC R \<or> (s,t) \<in> CC S" by simp
  thus "(s,t) \<in> CC(R \<union> S)"
  proof (rule disjE)
   assume "(s,t) \<in> CC R" thus ?thesis by (rule cc_subseteq_cc_ext1)
  next
   assume "(s,t) \<in> CC S" thus ?thesis by (rule cc_subseteq_cc_ext2)
  qed
 qed
qed

lemma CC_monotone: assumes "r \<subseteq> s" shows "CC r \<subseteq> CC s"
using assms
proof auto
  fix u v
  assume "(u,v) \<in> CC r"
  thus "(u,v) \<in> CC s"
  proof
    fix uu vv C
    assume "u = C\<langle>uu\<rangle>" and "v = C\<langle>vv\<rangle>" and "(uu,vv) \<in> r"
    with assms show ?thesis by (auto simp: ctxt_closureI)
  qed
qed

lemma CS_monotone: assumes "r \<subseteq> s" shows "CS r \<subseteq> CS s"
using assms
proof auto
  fix u v
  assume "(u,v) \<in> CS r"
  thus "(u,v) \<in> CS s"
  proof
    fix uu vv \<sigma>
    assume "u = uu \<cdot> \<sigma>" and "v = vv \<cdot> \<sigma>" and "(uu,vv) \<in> r"
    with assms show ?thesis by (auto simp: subst_closureI)
  qed
qed

lemma rstep_mono:
 assumes "rR \<subseteq> R"
 shows "rstep rR \<subseteq> rstep R"
proof-
 from assms have "CS rR \<subseteq> CS R" using CS_monotone by auto
 hence "CC (CS rR) \<subseteq> CC (CS R)" using CC_monotone by auto
 thus ?thesis by (simp add: rstep_CC_CS)
qed

lemma rstep_map_trs_power_mono: 
  fixes R R' :: "('f,'v)trs" and fg :: "'f \<Rightarrow> 'f"
  assumes subset: "R \<subseteq> R'" shows "rstep (((map_trs fg)^^n) R) \<subseteq> rstep (((map_trs fg)^^n) R')"
  by (rule rstep_mono, rule map_trs_power_mono, rule subset)

lemma cs_diff_distr[simp]: "CS A - CS B  \<subseteq> CS (A - B)" (* SW *)
proof-
 have "A = (A - B) \<union> (B \<inter> A)" by auto
 hence "CS A = CS ((A - B) \<union> (B \<inter> A))" by auto
 hence "CS A = CS (A - B) \<union> CS (B \<inter> A)" using cs_union_distr by auto
 hence "CS A - CS (B \<inter> A) = CS (A - B) \<union> CS (B \<inter> A) - CS (B \<inter> A)" by auto
 hence one:"CS A - CS (B \<inter> A) \<subseteq> CS (A - B)" by auto
 have "B \<inter> A \<subseteq> B" by auto
 from CS_monotone[OF this] have "CS (B \<inter> A) \<subseteq>  CS B" .
 hence "CS A - CS B \<subseteq> CS A - CS (B \<inter> A)" by auto
 with one show "CS A - CS B \<subseteq> CS (A - B)" by auto
qed

lemma cc_diff_distr[simp]: "CC A - CC B \<subseteq> CC (A - B)"  (* SW *)
proof-
 have "A = (A - B) \<union> (B \<inter> A)" by auto
 hence "CC A = CC ((A - B) \<union> (B \<inter> A))" by auto
 hence "CC A = CC (A - B) \<union> CC (B \<inter> A)" using cc_union_distr by auto
 hence one:"CC A - CC (B \<inter> A) = CC (A - B) \<union> CC (B \<inter> A) - CC (B \<inter> A)" by auto
 hence one:"CC A - CC (B \<inter> A) \<subseteq> CC (A - B)" by auto
 have "B \<inter> A \<subseteq> B" by auto
 from CC_monotone[OF this] have "CC (B \<inter> A) \<subseteq>  CC B" .
 hence "CC A - CC B \<subseteq> CC A - CC (B \<inter> A)" by auto
 with one show "CC A - CC B \<subseteq> CC (A - B)" by auto
qed

lemma rstep_diff:  (* SW *)
 shows "rstep R - rstep R' \<subseteq> rstep (R - R')"
proof-
 have "CC (CS R) - CC (CS R') \<subseteq> CC (CS R - CS R')" using cc_diff_distr by auto
 also have "\<dots> \<subseteq> CC (CS (R - R'))" using cs_diff_distr CC_monotone[where ?r="CS R - CS R'"] by auto
 finally show ?thesis unfolding rstep_CC_CS by auto
qed

   
lemma someI_ex2: "\<exists>x y. P (x,y) \<Longrightarrow> P (SOME (x,y). P (x,y))" using some_eq_ex by auto

lemma subsetI3:"(\<And>x y z. (x,y,z) \<in> A \<Longrightarrow> (x,y,z) \<in> B) \<Longrightarrow> A \<subseteq> B" by auto

lemma aux: "(\<Union>a\<in>P. {(x,y,z). x = fst a \<and> y = snd a \<and> Q a z}) = {(x,y,z). (x,y) \<in> P \<and> Q (x,y) z}" (is "?P = ?Q")
proof
 show "?P \<subseteq> ?Q"
 proof
  fix x assume "x \<in> ?P"
  then obtain a where "a \<in> P" and "x \<in> {(x,y,z). x = fst a \<and> y = snd a \<and> Q a z}" by auto
  then obtain b where "Q (fst x, fst (snd x)) (snd (snd x))" and "(fst x, fst (snd x)) = a" and "snd (snd x) = b" by force
  from `a \<in> P` have "(fst a,snd a) \<in> P" unfolding split_def by simp
  from `Q (fst x, fst (snd x)) (snd (snd x))` have "Q a b" unfolding `(fst x, fst (snd x)) = a` `snd (snd x) = b` .
  from `(fst a,snd a) \<in> P` and `Q a b` show "x \<in> ?Q" unfolding split_def `(fst x, fst (snd x)) = a`[symmetric] `snd (snd x) = b`[symmetric] by simp
 qed
next
 show "?Q \<subseteq> ?P"
 proof (rule subsetI3)
  fix x y z assume "(x,y,z) \<in> ?Q"
  hence "(x,y) \<in> P" and "Q (x,y) z" by auto
  hence "x = fst(x,y) \<and> y = snd(x,y) \<and> Q (x,y) z" by auto
  hence "(x,y,z) \<in> {(x,y,z). x = fst(x,y) \<and> y = snd(x,y) \<and> Q (x,y) z}" by auto
  hence "\<exists>p\<in>P. (x,y,z) \<in> {(x,y,z). x = fst p \<and> y = snd p \<and> Q p z}" using `(x,y) \<in> P` by blast
  thus "(x,y,z) \<in> ?P" unfolding UN_iff[symmetric] by simp
 qed
qed
 
lemma finiteR_imp_finiteDP': 
 assumes "finite R"
 shows "finite{(l,r,u). (\<exists>h us. u = Fun h us \<and> (l,r) \<in> R \<and> r \<unrhd> u \<and> defined R (h, length us) \<and> \<not>(l \<rhd> u))}"
proof -
 have "\<And>l r. (l,r) \<in> R \<Longrightarrow> finite{u. r \<unrhd> u}"
 proof -
  fix l r
  assume "(l,r) \<in> R"
  show "finite{u. r \<unrhd> u}" by (rule finite_subterms)
 qed
 with `finite R` have "finite(\<Union>(l,r) \<in> R. {u. r \<unrhd> u})" by auto
 have "finite(\<Union>lr\<in>R. {(l,r,u). l = fst lr \<and> r = snd lr \<and> snd lr \<unrhd> u})"
 proof (rule finite_UN_I)
  show "finite R" by (rule `finite R`)
 next
  fix lr
  assume "lr \<in> R"
  have "finite {u. snd lr \<unrhd> u}" by (rule finite_subterms)
  thus "finite {(l,r,u). l = fst lr \<and> r = snd lr \<and> snd lr \<unrhd> u}" by auto
 qed
 hence "finite {(l,r,u). (l,r) \<in> R \<and> r \<unrhd> u}" unfolding aux by auto
 have "{(l,r,u). (l,r) \<in> R \<and> r \<unrhd> u} \<supseteq> {(l,r,u). (\<exists>h us. u = Fun h us \<and> (l,r) \<in> R \<and> r \<unrhd> u \<and> defined R (h, length us) \<and> \<not>(l \<rhd> u))}" by auto
 with `finite {(l,r,u). (l,r) \<in> R \<and> r \<unrhd> u}` show ?thesis using finite_subset by fast
qed

lemma card_image_le':
 assumes "finite S"
 shows "card (\<Union>y\<in>S.{x. x = f y}) \<le> card S"
proof -
 have A:"(\<Union>y\<in>S. {x. x = f y}) = f ` S" by auto
 from assms show ?thesis unfolding A using card_image_le by auto
qed

lemma subteq_of_map_imp_map: "map_term g s \<unrhd> t \<Longrightarrow> \<exists>u. t = map_term g u"
proof (induct s arbitrary: t)
 case (Var x)
 hence "map_term g (Var x) \<rhd> t \<or> map_term g (Var x) = t" by auto
 thus ?case
 proof
  assume "map_term g (Var x) \<rhd> t" thus ?thesis by (cases rule: supt.cases) auto
 next
  assume "map_term g (Var x) = t" thus ?thesis by best
 qed
next
 case (Fun f ss)
 hence "map_term g (Fun f ss) \<rhd> t \<or> map_term g (Fun f ss) = t" by auto
 thus ?case
 proof
  assume "map_term g (Fun f ss) \<rhd> t"
  thus ?case using Fun by (cases rule: supt.cases) (auto simp: supt_supteq_conv)
 next
  assume "map_term g (Fun f ss) = t" thus ?thesis by best
 qed
qed

lemma map_term_inj:
  assumes "inj (fg :: ('f \<Rightarrow> 'f))"
  shows "inj (map_term fg)"
proof -
  {
    fix s t :: "('f,'v)term"
    assume "map_term fg s = map_term fg t"
    hence "s = t"
    proof (induct s arbitrary: t)
      case (Var x) with assms show ?case by (induct t) auto
    next
      case (Fun f ss) thus ?case
      proof (induct t)
        case (Var y) thus ?case by auto
      next
        case (Fun g ts)
        hence A: "map (map_term fg) ss = map (map_term fg) ts" by simp
        hence len_eq:"length ss = length ts" by (rule map_eq_imp_length_eq)
        from A have "!!i. i < length ss \<Longrightarrow> (map (map_term fg) ss)!i = (map (map_term fg) ts)!i" by auto
        with len_eq have eq: "!!i. i < length ss \<Longrightarrow> map_term fg  (ss!i) = map_term fg (ts!i)" using nth_map by auto
        have in_set: "!!i. i < length ss \<Longrightarrow> (ss!i) \<in> set ss" by auto
        from Fun have "\<forall>i < length ss. (ss!i) = (ts!i)" using in_set eq by auto
        with len_eq have "ss = ts" using nth_equalityI[where xs = ss and ys = ts] by simp
        have "f = g" using Fun `inj fg` unfolding inj_on_def by auto
        with `ss = ts` show ?case by simp
      qed
    qed
  }
  thus ?thesis unfolding inj_on_def by auto
qed 

lemma rsteps_closed_ctxt:
  assumes "(s, t) \<in> (rstep R)\<^sup>*"
  shows "(C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> (rstep R)\<^sup>*"
proof -
  from assms obtain n where "(s,t) \<in> (rstep R)^^n"
    using rtrancl_is_UN_relpow by auto
  thus ?thesis
  proof (induct n arbitrary: s)
    case 0 thus ?case by auto
  next
    case (Suc n)
    from relpow_Suc_D2[OF `(s,t) \<in> (rstep R)^^Suc n`] obtain u
      where "(s,u) \<in> (rstep R)" and "(u,t) \<in> (rstep R)^^n" by auto
    from `(s,u) \<in> (rstep R)` have "(C\<langle>s\<rangle>,C\<langle>u\<rangle>) \<in> (rstep R)" ..
    from Suc and `(u,t) \<in> (rstep R)^^n` have "(C\<langle>u\<rangle>,C\<langle>t\<rangle>) \<in> (rstep R)^*" by simp
    with `(C\<langle>s\<rangle>,C\<langle>u\<rangle>) \<in> (rstep R)` show ?case by auto
 qed
qed

lemma not_Nil_imp_last: "xs \<noteq> [] \<Longrightarrow> \<exists>ys y. xs = ys@[y]"
proof (induct xs)
 case Nil thus ?case by simp
next
 case (Cons x xs) show ?case
 proof (cases xs)
  assume "xs = []" with Cons show ?thesis by simp
 next
  fix x' xs' assume "xs = x'#xs'"
  hence "xs \<noteq> []" by simp
  with Cons obtain ys y where "xs = ys@[y]" by auto
  hence "x#xs = x#(ys@[y])" by simp
  hence "x#xs = (x#ys)@[y]" by simp
  thus ?thesis by auto
 qed
qed

lemma Nil_or_last: "xs = [] \<or> (\<exists>ys y. xs = ys@[y])"
proof (cases xs)
 assume "xs = []" thus ?thesis by simp
next
 fix x' xs' assume "xs = x'#xs'"
 hence "xs \<noteq> []" by simp
 thus ?thesis using not_Nil_imp_last by auto
qed


lemma ctxt_closed_one: "ctxt_closed r \<Longrightarrow> (s,t) \<in> r \<Longrightarrow> (Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> r"
  using ctxt_closed_imp_ctxt[of r s t "More f bef \<box> aft"] by auto

lemma ctxt_closed_union: "ctxt_closed r \<Longrightarrow> ctxt_closed s \<Longrightarrow> ctxt_closed (r \<union> s)"
  unfolding ctxt_closed_def cc_union_distr by auto

lemma one_imp_ctxt_closed: assumes 
  one: "\<And> f bef s t aft. (s,t) \<in> r \<Longrightarrow> (Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> r" shows "ctxt_closed r"
proof (unfold ctxt_closed_def, rule)
  fix cs ct
  assume "(cs, ct) \<in> CC r" 
  thus "(cs, ct) \<in> r"
  proof
    fix s t C 
    assume cs: "cs = C\<langle>s\<rangle>" and ct: "ct = C\<langle>t\<rangle>" and st: "(s,t) \<in> r"
    have "(C\<langle>s\<rangle>,C\<langle>t\<rangle>) \<in> r" 
    proof (induct C)
      case Hole
      thus ?case using st by simp
    next
      case (More f bef C aft)
      from one[OF More, of f bef aft]
      show ?case by simp
    qed
    with cs ct show ?thesis by simp 
  qed
qed

lemma ctxt_closed_comp: assumes 
  r: "ctxt_closed r" and s: "ctxt_closed s"
  shows "ctxt_closed (s O r)"
proof (rule one_imp_ctxt_closed)
  fix f bef u w aft
  assume "(u,w) \<in> s O r"
  then obtain v where uv: "(u,v) \<in> s" and vw: "(v,w) \<in> r" by auto
  show "(Fun f (bef @ u # aft), Fun f (bef @ w # aft)) \<in> s O r"
    using ctxt_closed_one[OF s uv, of f bef aft] ctxt_closed_one[OF r vw, of f bef aft]
    by auto
qed

lemma ctxt_closed_nrrstep: "ctxt_closed (nrrstep R)" 
proof (rule one_imp_ctxt_closed)
  fix f bef s t aft
  assume "(s,t) \<in> nrrstep R"
  from this[unfolded nrrstep_def'] obtain l r C \<sigma>
    where lr: "(l,r) \<in> R" and C: "C \<noteq> \<box>"
    and s: "s = C\<langle>l\<cdot>\<sigma>\<rangle>" and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>" by auto
  show "(Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> nrrstep R"
  proof (rule nrrstepI[OF lr])
    show "More f bef C aft \<noteq> \<box>" by simp
  qed (insert s t, auto)
qed

lemma ctxt_closed_rtrancl: assumes 
  r: "ctxt_closed r" 
  shows "ctxt_closed (r^*)"
proof (rule one_imp_ctxt_closed)
  fix f bef s t aft
  assume "(s,t) \<in> r^*"
  thus "(Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> r^*"
  proof (induct rule: rtrancl_induct, simp)
    fix t u
    assume a: "(t,u) \<in> r" and
      b: "(Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> r^*"
    from ctxt_closed_one[OF r a]
    have "(Fun f (bef @ t # aft), Fun f (bef @ u # aft)) \<in> r" .
    with b show "(Fun f (bef @ s # aft), Fun f (bef @ u # aft)) \<in> r^*" by auto
  qed
qed

lemma ctxt_closed_relto: "ctxt_closed R \<Longrightarrow> ctxt_closed S \<Longrightarrow> ctxt_closed (relto R S)"
  by (auto intro: ctxt_closed_rtrancl ctxt_closed_comp)

definition all_ctxt_closed :: "'f sig \<Rightarrow> ('f, 'v) trs \<Rightarrow> bool" where
  "all_ctxt_closed F r \<longleftrightarrow> (\<forall>f ts ss. (f, length ss) \<in> F \<longrightarrow> length ts = length ss \<longrightarrow>
    (\<forall>i. i < length ts \<longrightarrow> (ts ! i, ss ! i) \<in> r) \<longrightarrow> (\<forall> i. i < length ts \<longrightarrow> funas_term (ts ! i) \<union> funas_term (ss ! i) \<subseteq> F) \<longrightarrow> (Fun f ts, Fun f ss) \<in> r) \<and> (\<forall> x. (Var x, Var x) \<in> r)"

lemma all_ctxt_closedE: "all_ctxt_closed F r \<Longrightarrow> (f,length ss) \<in> F \<Longrightarrow> length ts = length ss 
  \<Longrightarrow> \<lbrakk>\<And> i. i < length ts \<Longrightarrow> (ts ! i, ss ! i) \<in> r \<rbrakk>
  \<Longrightarrow> \<lbrakk>\<And> i. i < length ts \<Longrightarrow> funas_term (ts ! i) \<subseteq> F \<rbrakk> 
  \<Longrightarrow> \<lbrakk>\<And> i. i < length ts \<Longrightarrow> funas_term (ss ! i) \<subseteq> F \<rbrakk> 
  \<Longrightarrow> (Fun f ts, Fun f ss) \<in> r"
  unfolding all_ctxt_closed_def by auto

lemma all_ctxt_closed_sig_reflE: assumes all: "all_ctxt_closed F r"
  shows "funas_term t \<subseteq> F \<Longrightarrow> (t,t) \<in> r"
proof (induct t)
  case (Var x)
  from all[unfolded all_ctxt_closed_def] show ?case by auto
next
  case (Fun f ts)
  show ?case
    by (rule all_ctxt_closedE[OF all _ _ Fun(1)], insert Fun(2), force+)
qed

lemma all_ctxt_closed_reflE: assumes all: "all_ctxt_closed UNIV r"
  shows "(t,t) \<in> r"
  by (rule all_ctxt_closed_sig_reflE[OF all], auto)

lemma all_ctxt_closed_subst_step_sig:
  fixes r :: "('f, 'v) trs" and t :: "('f, 'v) term"
  assumes all: "all_ctxt_closed F r"
    and sig: "funas_term t \<subseteq> F"
    and steps: "\<And> x. x \<in> vars_term t \<Longrightarrow> (\<sigma> x, \<tau> x) \<in> r"
    and sig_subst: "\<And> x. x \<in> vars_term t \<Longrightarrow> funas_term (\<sigma> x) \<union> funas_term (\<tau> x) \<subseteq> F"
  shows "(t \<cdot> \<sigma>, t \<cdot> \<tau>) \<in> r"
  using sig steps sig_subst
proof (induct t)
  case (Var x)
  thus ?case by auto
next
  case (Fun f ts)
  {
    fix t
    assume t: "t \<in> set ts"
    with Fun(2-3) have "funas_term t \<subseteq> F" "\<And> x. x \<in> vars_term t \<Longrightarrow> (\<sigma> x, \<tau> x) \<in> r" by auto
    from Fun(1)[OF t this Fun(4)] t have step: "(t \<cdot> \<sigma>, t \<cdot> \<tau>) \<in> r" by auto
    from Fun(4) t have "\<And> x. x \<in> vars_term t \<Longrightarrow> funas_term (\<sigma> x) \<union> funas_term (\<tau> x) \<subseteq> F" by auto
    with `funas_term t \<subseteq> F` have "funas_term (t \<cdot> \<sigma>) \<union> funas_term (t \<cdot> \<tau>) \<subseteq> F" unfolding funas_term_subst by auto
    note step this
  }
  hence steps: "\<And> i. i < length ts \<Longrightarrow> (ts ! i \<cdot> \<sigma>, ts ! i \<cdot> \<tau>) \<in> r \<and> funas_term (ts ! i \<cdot> \<sigma>) \<union> funas_term (ts ! i \<cdot> \<tau>) \<subseteq> F" unfolding set_conv_nth by auto
  with all_ctxt_closedE[OF all, of f "map (\<lambda> t. t \<cdot> \<tau>) ts" "map (\<lambda> t. t \<cdot> \<sigma>) ts"] Fun(2)
  show ?case by auto
qed

lemma all_ctxt_closed_subst_step:
  fixes r :: "('f, 'v) trs" and t :: "('f, 'v) term"
  assumes all: "all_ctxt_closed UNIV r"
    and steps: "\<And> x. x \<in> vars_term t \<Longrightarrow> (\<sigma> x, \<tau> x) \<in> r"
  shows "(t \<cdot> \<sigma>, t \<cdot> \<tau>) \<in> r"
  by (rule all_ctxt_closed_subst_step_sig[OF all _ steps], auto)

lemma all_ctxt_closed_ctxtE: assumes all: "all_ctxt_closed F R" 
  and Fs: "funas_term s \<subseteq> F" 
  and Ft: "funas_term t \<subseteq> F" 
  and step: "(s,t) \<in> R"
  shows "funas_ctxt C \<subseteq> F \<Longrightarrow> (C \<langle> s \<rangle>, C \<langle> t \<rangle>) \<in> R"
proof(induct C)
  case Hole
  from step show ?case by auto
next
  case (More f bef C aft)
  let ?n = "length bef"
  let ?m = "Suc (?n + length aft)"
  show ?case unfolding ctxt_apply.simps
  proof (rule all_ctxt_closedE[OF all])
    fix i
    let ?t = "\<lambda> s. (bef @ C \<langle> s \<rangle> # aft) ! i"
    assume "i < length (bef @ C\<langle>s\<rangle> # aft)"
    hence i: "i < ?m" by auto
    hence mem: "\<And> s. ?t s \<in> set (bef @ C \<langle> s \<rangle> # aft)" unfolding set_conv_nth by auto
    from mem[of s] More Fs show "funas_term (?t s) \<subseteq> F" by auto
    from mem[of t] More Ft show "funas_term (?t t) \<subseteq> F" by auto
    from More have step: "(C \<langle> s \<rangle>, C \<langle> t \<rangle>) \<in> R" by auto
    {
      fix s
      assume "s \<in> set bef \<union> set aft"
      with More have "funas_term s \<subseteq> F" by auto
      from all_ctxt_closed_sig_reflE[OF all this] have "(s,s) \<in> R" by auto
    } note steps = this
    show "(?t s, ?t t) \<in> R" 
    proof (cases "i = ?n")
      case True
      thus ?thesis using step by auto
    next
      case False
      show ?thesis
      proof (cases "i < ?n")
        case True
        thus ?thesis unfolding append_Cons_nth_left[OF True] using steps by auto
      next
        case False
        with `i \<noteq> ?n` i have "\<exists> k. k < length aft \<and> i = Suc ?n + k" by presburger
        then obtain k where k: "k < length aft" and i: "i = Suc ?n + k" by auto
        from k show ?thesis using steps unfolding i by (auto simp: nth_append)
      qed
    qed
  qed (insert More, auto)
qed 

lemma trans_ctxt_sig_imp_all_ctxt_closed: assumes tran: "trans r" 
  and refl: "\<And> t. funas_term t \<subseteq> F \<Longrightarrow> (t,t) \<in> r"
  and ctxt: "\<And> C s t. funas_ctxt C \<subseteq> F \<Longrightarrow> funas_term s \<subseteq> F \<Longrightarrow> funas_term t \<subseteq> F \<Longrightarrow> (s,t) \<in> r \<Longrightarrow> (C \<langle> s \<rangle>, C \<langle> t \<rangle>) \<in> r" 
  shows "all_ctxt_closed F r" 
unfolding all_ctxt_closed_def
proof (rule, intro allI impI)
  fix f ts ss
  assume f: "(f,length ss) \<in> F" and
         l: "length ts = length ss" and
         steps: "\<forall> i < length ts. (ts ! i, ss ! i) \<in> r" and
         sig: "\<forall> i < length ts. funas_term (ts ! i) \<union>  funas_term (ss ! i) \<subseteq> F"
  from sig have sig_ts: "\<And> t. t \<in> set ts \<Longrightarrow> funas_term t \<subseteq> F"  unfolding set_conv_nth by auto
  from sig have sig_ss: "\<And> t. t \<in> set ss \<Longrightarrow> funas_term t \<subseteq> F"  unfolding l set_conv_nth by auto
  let ?n = "length ts"
  let ?us = "\<lambda> i. take i ss @ (take (?n - i) (List.drop i ts))"
  have "\<And> i. i \<le> ?n \<Longrightarrow> (Fun f ts, Fun f (?us i)) \<in> r"
  proof -
    fix i
    assume "i \<le> ?n"
    thus "(Fun f ts, Fun f (?us i)) \<in> r"
    proof (induct i)
      case 0 
      have "(Fun f ts, Fun f ts) \<in> r"
        by (rule refl, insert f sig_ts sig_ss l, auto)
      thus ?case by simp
    next
      case (Suc i)
      hence i: "i < length ts" by auto
      from Suc have ind: "(Fun f ts, Fun f (?us i)) \<in> r" by auto
      from Suc steps have step: "(ts ! i, ss ! i) \<in> r" by auto
      let ?bef = "take i ss"
      let ?aft = "take (?n - Suc i) (List.drop (Suc i) ts)"
      let ?C = "More f ?bef Hole ?aft"
      from i l have id: "Suc (min (length ss) i + (length ts - Suc i)) = length ss" by auto
      have "funas_ctxt ?C = {(f,length ss)} \<union> (\<Union> (funas_term ` (set (take i ss) \<union> set (drop (Suc i) ts))))" by (auto simp: id)
      also have "set (take i ss) \<union> set (drop (Suc i) ts) \<subseteq> set ss \<union> set ts" 
        by (metis set_drop_subset set_take_subset sup_mono)
      finally have "funas_ctxt ?C \<subseteq> {(f,length ss)} \<union> (\<Union> (funas_term ` (set ss \<union> set ts)))" by auto
      also have "\<dots> \<subseteq> F" using f sig_ss sig_ts by auto
      finally have C: "funas_ctxt ?C \<subseteq> F" .
      from i sig have "funas_term (ts ! i) \<subseteq> F" "funas_term (ss ! i) \<subseteq> F" by auto
      from ctxt[OF C this step]
      have cstep: "(Fun f (?bef @ ts ! i # ?aft), Fun f (?bef @ ss ! i # ?aft)) \<in> r" (is "(?l,?r) \<in> r") by auto
      have lId: "Fun f (?us i) = ?l" using l i by (auto simp: drop_Suc_conv_tl)
      have rId: "Fun f (?us (Suc i)) = ?r" using l i by (auto simp: take_Suc_conv_app_nth)
      from cstep have "(Fun f (?us i), Fun f (?us (Suc i))) \<in> r" by (simp only: lId rId)
      with  ind show ?case using tran unfolding trans_def by best
    qed
  qed   
  hence "(Fun f ts, Fun f (?us ?n)) \<in> r" by force
  thus "(Fun f ts, Fun f ss) \<in> r" using l by  auto
qed (insert refl, auto)

lemma trans_ctxt_imp_all_ctxt_closed: assumes tran: "trans r" 
  and refl: "refl r"
  and ctxt: "ctxt_closed r" 
  shows "all_ctxt_closed F r"
  by (rule trans_ctxt_sig_imp_all_ctxt_closed[OF tran _ ctxt_closedE[OF ctxt]], insert refl[unfolded refl_on_def], auto)

lemmas ctxt_closed_intros[intro] =
  trans_rtrancl
  refl_rtrancl
  ctxt_closed_rtrancl
  ctxt_closed_rstep
  ctxt_closed_union
  ctxt_closed_comp
  ctxt_closed_relto
  ctxt_closed_nrrstep
  ctxt_closed_trancl

lemma rsteps_all_ctxt_closed[intro]: "all_ctxt_closed F ((rstep r)^*)"
  by (blast intro: trans_ctxt_imp_all_ctxt_closed)

lemma subst_rsteps_imp_rsteps:
  fixes \<sigma> :: "('f, 'v) subst"
  assumes "\<forall>x\<in>vars_term t. (\<sigma> x, \<tau> x) \<in> (rstep R)\<^sup>*"
  shows "(t \<cdot> \<sigma>, t \<cdot> \<tau>) \<in> (rstep R)\<^sup>*"
  by (rule all_ctxt_closed_subst_step)
     (insert assms, auto)

lemma rtrancl_trancl_into_trancl:
  assumes len: "length ts = length ss"
    and steps: "\<forall> i < length ts. (ts ! i, ss ! i) \<in> R^*"
    and i: "i < length ts"
    and step: "(ts ! i, ss ! i) \<in> R^+"
    and ctxt: "ctxt_closed R"
  shows "(Fun f ts, Fun f ss) \<in> R^+"
proof -
  from ctxt have ctxt_rt: "ctxt_closed (R^*)" by blast
  from ctxt have ctxt_t: "ctxt_closed (R^+)" by blast
  from id_take_nth_drop[OF i] have ts: "ts = take i ts @ ts ! i # drop (Suc i) ts" (is "_ = ?ts") by auto
  from id_take_nth_drop[OF i[simplified len]] have ss: "ss = take i ss @ ss ! i # drop (Suc i) ss" (is "_ = ?ss") by auto
  let ?mid = "take i ss @ ts ! i # drop (Suc i) ss"
  from trans_ctxt_imp_all_ctxt_closed[OF trans_rtrancl refl_rtrancl ctxt_rt] have all: "all_ctxt_closed UNIV (R^*)" .
  from ctxt_closed_one[OF ctxt_t step] have "(Fun f ?mid, Fun f ?ss) \<in> R^+" .
  hence part1: "(Fun f ?mid, Fun f ss) \<in> R^+" unfolding ss[symmetric] . 
  from ts have lents: "length ts =  length ?ts" by simp
  have "(Fun f ts, Fun f ?mid) \<in> R^*"
  proof (rule all_ctxt_closedE[OF all], simp, simp only: lents, simp add: len)
    fix j
    assume jts: "j < length ts"
    from i len have i: "i < length ss" by auto
    show "(ts ! j, ?mid ! j) \<in> R^*"
    proof (cases "j < i")
      case True
      with i have j: "j < length ss" by auto
      with True have id: "?mid ! j = ss ! j" by (simp add: nth_append)
      from steps len j have "(ts ! j, ss ! j) \<in> R^*" by auto 
      thus ?thesis using id by simp
    next
      case False
      show ?thesis 
      proof (cases "j = i")
        case True
        hence "?mid ! j = ts ! j" using i by (simp add: nth_append)
        thus ?thesis by simp
      next
        case False
        from i have min: "min (length ss) i = i" by simp
        from False `\<not> j < i` have "j > i" by arith
        then obtain k where k: "j - i = Suc k" by (cases "j - i", auto)
        hence j: "j = Suc (i+k)" by auto
        with jts len have ss: "Suc i + k \<le> length ss" and jlen: "j < length ts" by auto
        hence "?mid ! j = ss ! j" using j i by (simp add: nth_append min `\<not> j < i` nth_drop[OF ss])
        with steps jlen show ?thesis by auto
      qed
    qed
  qed auto
  with part1 show ?thesis by auto
qed

lemma SN_ctxt_apply_imp_SN_ctxt_to_term_list_gen: 
  assumes ctxt: "ctxt_closed r"
  assumes SN: "SN_on r {C\<langle>t\<rangle>}"
  shows "SN_on r (set (ctxt_to_term_list C))"
proof -
  {
    fix u
    assume "u \<in> set (ctxt_to_term_list C)"
    from ctxt_to_term_list_supt[OF this, of t] have "C\<langle>t\<rangle> \<unrhd> u"
      by (rule supt_imp_supteq)
    from SN_imp_SN_subt_gen[OF ctxt, OF SN this]
    have "SN_on r {u}" by auto
  }
  thus ?thesis
    unfolding SN_on_def by auto
qed


lemma trs_inclusion: 
  "ctxt_closed Rel \<Longrightarrow> subst_closed Rel \<Longrightarrow> R \<subseteq> Rel \<Longrightarrow> rstep R \<subseteq> Rel"
  by fast

lemma trancl_rstep_ctxt_closed: 
  "(s,t) \<in> (rstep R)^+ \<Longrightarrow> (C\<langle>s\<rangle>,C\<langle>t\<rangle>) \<in> (rstep R)^+"
  by (rule ctxt_closed_imp_ctxt, blast)

lemma args_steps_imp_steps_gen: assumes ctxt: "\<And> bef s t aft. (s,t) \<in> r (length bef) \<Longrightarrow>
  length ts = Suc (length bef + length aft) \<Longrightarrow>
  (Fun f (bef @ (s :: ('f,'v)term) # aft), Fun f (bef @ t # aft)) \<in> R^*"
  and len: "length ss = length ts" and args: "\<And> i. i<length ts \<Longrightarrow> (ss!i,ts!i) \<in> (r i)^*"
  shows "(Fun f ss,Fun f ts) \<in> R^*"
proof -
  let ?tss = "\<lambda>i. take i ts @ drop i ss"
  {
    fix bef :: "('f,'v)term list" and s t and  aft :: "('f,'v)term list"
    assume "(s,t) \<in> (r (length bef))^*" and len: "length ts = Suc (length bef + length aft)"
    hence "(Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> R^*"
    proof (induct)
      case (step t u)
      from step(3)[OF len] ctxt[OF step(2) len] show ?case by auto
    qed simp
  }
  note one = this
  have a:"\<forall>i \<le> length ts. (Fun f ss,Fun f (?tss i)) \<in> R^*"
  proof (intro allI impI)
    fix i assume "i \<le> length ts" thus "(Fun f ss,Fun f (?tss i)) \<in> R^*"
    proof (induct i)
      case 0
      thus ?case by simp
    next
      case (Suc i)
      hence IH: "(Fun f ss,Fun f (?tss i)) \<in> R^*" 
        and i: "i < length ts" by auto
      have si: "?tss (Suc i) = take i ts @ ts ! i # drop (Suc i) ss"
        unfolding take_Suc_conv_app_nth[OF i] by simp
      have ii: "?tss i = take i ts @ ss ! i # drop (Suc i) ss"
        unfolding nth_drop'[OF i[unfolded len[symmetric]]] ..
      from i have i': "length (take i ts) < length ts"  and len': "length (take i ts) = i" by auto
      from len i have len'': "length ts = Suc (length (take i ts) + length (drop (Suc i) ss))" by simp
      from one[OF args[OF i'] len''] IH
      show ?case unfolding si ii len' by auto
    qed
  qed
  from this[THEN spec, THEN mp[OF _ le_refl]]
  show ?thesis using len by auto
qed


lemma args_steps_imp_steps: assumes ctxt: "ctxt_closed R"
  and len: "length ss = length ts" and args: "\<forall>i<length ss. (ss!i,ts!i) \<in> R^*"
  shows "(Fun f ss,Fun f ts) \<in> R^*"
proof (rule args_steps_imp_steps_gen[OF _ len])
  fix i
  assume "i < length ts" thus "(ss ! i, ts ! i) \<in> R^*" using args len by auto
qed (insert ctxt_closed_one[OF ctxt], auto)  

lemmas args_rsteps_imp_rsteps = args_steps_imp_steps [OF ctxt_closed_rstep]

lemma replace_at_subst_rsteps:
  fixes \<sigma> \<tau> :: "('f, 'v) subst"
  assumes *: "\<And>x. (\<sigma> x, \<tau> x) \<in> (rstep R)\<^sup>*"
    and "p \<in> poss t"
    and "t |_ p = Var x"
  shows "(replace_at (t \<cdot> \<sigma>) p (\<tau> x), t \<cdot> \<tau>) \<in> (rstep R)\<^sup>*"
using assms(2-)
proof (induction t arbitrary: p)
  case (Fun f ts)
  then obtain i q where [simp]: "p = i <# q" and i: "i < length ts"
    and q: "q \<in> poss (ts ! i)" and [simp]: "ts ! i |_ q = Var x" by (cases p) auto
  let ?C = "ctxt_of_pos_term q (ts ! i \<cdot> \<sigma>)"
  let ?ts = "map (\<lambda>t. t \<cdot> \<tau>) ts"
  let ?ss = "take i (map (\<lambda>t. t \<cdot> \<sigma>) ts) @ ?C\<langle>\<tau> x\<rangle> # drop (Suc i) (map (\<lambda>t. t \<cdot> \<sigma>) ts)"
  have "\<forall>j<length ts. (?ss ! j, ?ts ! j) \<in> (rstep R)\<^sup>*"
  proof (intro allI impI)
    fix j
    assume j: "j < length ts"
    moreover
    { assume [simp]: "j = i"
      have "?ss ! j = ?C\<langle>\<tau> x\<rangle>" using i by (simp add: nth_append_take)
      with Fun.IH [of "ts ! i" q]
      have "(?ss ! j, ?ts ! j) \<in> (rstep R)\<^sup>*" using q and i by simp }
    moreover
    { assume "j < i"
      with i have "?ss ! j = ts ! j \<cdot> \<sigma>"
        and "?ts ! j = ts ! j \<cdot> \<tau>" by (simp_all add: nth_append_take_is_nth_conv)
      then have "(?ss ! j, ?ts ! j) \<in> (rstep R)\<^sup>*" by (auto simp: * subst_rsteps_imp_rsteps) }
    moreover
    { assume "j > i"
      with i and j have "?ss ! j = ts ! j \<cdot> \<sigma>"
        and "?ts ! j = ts ! j \<cdot> \<tau>" by (simp_all add: nth_append_drop_is_nth_conv)
      then have "(?ss ! j, ?ts ! j) \<in> (rstep R)\<^sup>*" by (auto simp: * subst_rsteps_imp_rsteps) }
    ultimately show "(?ss ! j, ?ts ! j) \<in> (rstep R)\<^sup>*" by arith
  qed
  moreover have "i < length ts" by fact
  ultimately show ?case
    by (simp add: args_rsteps_imp_rsteps)
qed simp

lemma substs_rsteps:
  assumes "\<And>x. (\<sigma> x, \<tau> x) \<in> (rstep R)\<^sup>*"
  shows "(t \<cdot> \<sigma>, t \<cdot> \<tau>) \<in> (rstep R)\<^sup>*"
proof (induct t)
  case (Var y)
  show ?case using assms by simp_all
next
  case (Fun f ts)
  then have "\<forall>i<length (map (\<lambda>t. t \<cdot> \<sigma>) ts).
    (map (\<lambda>t. t \<cdot> \<sigma> ) ts ! i, map (\<lambda>t. t \<cdot> \<tau>) ts ! i) \<in> (rstep R)\<^sup>*" by auto
  from args_rsteps_imp_rsteps [OF _ this] show ?case by simp
qed

lemma nrrstep_Fun_imp_arg_rstep:
 fixes ss :: "('f,'v)term list"
 assumes "(Fun f ss,Fun f ts) \<in> nrrstep R" (is "(?s,?t) \<in> nrrstep R")
 shows "\<exists>C i. i < length ss \<and> (ss!i,ts!i) \<in> rstep R \<and> C\<langle>ss!i\<rangle> = Fun f ss \<and> C\<langle>ts!i\<rangle> = Fun f ts"
proof -
 from `(?s,?t) \<in> nrrstep R`
 obtain l r j ps \<sigma> where A: "let C = ctxt_of_pos_term (j<#ps) ?s in (j<#ps) \<in> poss ?s \<and> (l,r) \<in> R \<and> C\<langle>l\<cdot>\<sigma>\<rangle> = ?s \<and> C\<langle>r\<cdot>\<sigma>\<rangle> = ?t" unfolding nrrstep_def rstep_r_p_s_def by force
 let ?C = "ctxt_of_pos_term (j<#ps) ?s"
 from A have "(j<#ps) \<in> poss ?s" and "(l,r) \<in> R" and "?C\<langle>l\<cdot>\<sigma>\<rangle> = ?s" and "?C\<langle>r\<cdot>\<sigma>\<rangle> = ?t" using Let_def by auto
 have C: "?C = More f (take j ss) (ctxt_of_pos_term ps (ss!j)) (drop (Suc j) ss)" (is "?C = More f ?ss1 ?D ?ss2") by auto
 from `?C\<langle>l\<cdot>\<sigma>\<rangle> = ?s` have "?D\<langle>l\<cdot>\<sigma>\<rangle> = (ss!j)" by (auto simp: take_drop_imp_nth)
 from `(l,r) \<in> R` have "(l\<cdot>\<sigma>,r\<cdot>\<sigma>) \<in> (CS R)" by auto
 hence "(?D\<langle>l\<cdot>\<sigma>\<rangle>,?D\<langle>r\<cdot>\<sigma>\<rangle>) \<in> (CC(CS R))" and "(?C\<langle>l\<cdot>\<sigma>\<rangle>,?C\<langle>r\<cdot>\<sigma>\<rangle>) \<in> (CC(CS R))" by (auto simp only: ctxt_closure.ctxt_closureI)
 hence D_rstep: "(?D\<langle>l\<cdot>\<sigma>\<rangle>,?D\<langle>r\<cdot>\<sigma>\<rangle>) \<in> rstep R" and "(?C\<langle>l\<cdot>\<sigma>\<rangle>,?C\<langle>r\<cdot>\<sigma>\<rangle>) \<in> rstep R"
   by (auto simp: rstep_CC_CS)
 from `?C\<langle>r\<cdot>\<sigma>\<rangle> = ?t` and C have "?t = Fun f (take j ss @ ?D\<langle>r\<cdot>\<sigma>\<rangle> # drop (Suc j) ss)" by auto
 hence ts: "ts = (take j ss @ ?D\<langle>r\<cdot>\<sigma>\<rangle> # drop (Suc j) ss)" by auto
 from `j<#ps \<in> poss ?s` have r0: "j < length ss" by simp
 hence "(take j ss @ ?D\<langle>r\<cdot>\<sigma>\<rangle> # drop (Suc j) ss) ! j = ?D\<langle>r\<cdot>\<sigma>\<rangle>" by (auto simp: nth_append_take)
 with ts have "ts!j = ?D\<langle>r\<cdot>\<sigma>\<rangle>" by auto
 let ?C' = "More f (take j ss) \<box> (drop (Suc j) ss)"
 from D_rstep have r1: "(ss!j,ts!j) \<in> rstep R" unfolding `ts!j = ?D\<langle>r\<cdot>\<sigma>\<rangle>` `?D\<langle>l\<cdot>\<sigma>\<rangle> = ss!j` by simp
 have "?s = ?C\<langle>l\<cdot>\<sigma>\<rangle>" unfolding `?C\<langle>l\<cdot>\<sigma>\<rangle> = ?s` by simp
 also have "\<dots> = ?C'\<langle>?D\<langle>l\<cdot>\<sigma>\<rangle>\<rangle>" unfolding C by simp
 finally have r2:"?C'\<langle>ss!j\<rangle> = ?s" unfolding `?D\<langle>l\<cdot>\<sigma>\<rangle> = ss!j` by simp
 have "?t = ?C\<langle>r\<cdot>\<sigma>\<rangle>" unfolding `?C\<langle>r\<cdot>\<sigma>\<rangle> = ?t` by simp
 also have "\<dots> = ?C'\<langle>?D\<langle>r\<cdot>\<sigma>\<rangle>\<rangle>" unfolding C by simp
 finally have r3:"?C'\<langle>ts!j\<rangle> = ?t" unfolding `ts!j = ?D\<langle>r\<cdot>\<sigma>\<rangle>` by simp
 from r0 r1 r2 r3 show ?thesis by best
qed


lemma pair_fun_eq[simp]:
 fixes f :: "'a \<Rightarrow> 'b" and g :: "'b \<Rightarrow> 'a"
 shows "((\<lambda>(x,y). (x,f y)) \<circ> (\<lambda>(x,y). (x,g y))) = (\<lambda>(x,y). (x,(f \<circ> g) y))" (is "?f = ?g")
proof (rule ext)
 fix ab :: "'c * 'b"
 obtain a b where "ab = (a,b)" by force
 have "((\<lambda>(x,y). (x,f y))\<circ>(\<lambda>(x,y). (x,g y))) (a,b) = (\<lambda>(x,y). (x,(f\<circ>g) y)) (a,b)" by simp
 thus "?f ab = ?g ab" unfolding `ab = (a,b)` by simp
qed


lemma restrict_singleton:
  assumes "x \<in> subst_domain \<sigma>" shows "\<exists>t. \<sigma> |s {x} = (\<lambda>y. if y = x then t else Var y)"
proof -
  have "\<sigma> |s {x} = (\<lambda>y. if y = x then \<sigma> y else Var y)" by (simp add: subst_restrict_def)
  hence "\<sigma> |s {x} = (\<lambda>y. if y = x then \<sigma> x else Var y)" by (simp cong: if_cong)
  thus ?thesis by (rule exI[of _ "\<sigma> x"])
qed

definition rstep_r_c_s :: "('f,'v)rule \<Rightarrow> ('f,'v)ctxt \<Rightarrow> ('f,'v)subst \<Rightarrow> ('f,'v)term rel"
where "rstep_r_c_s r C \<sigma> = {(s,t) | s t. s = C\<langle>fst r \<cdot> \<sigma>\<rangle> \<and> t = C\<langle>snd r \<cdot> \<sigma>\<rangle>}"

lemma rstep_iff_rstep_r_c_s: "((s,t) \<in> rstep R) = (\<exists> l r C \<sigma>. (l,r) \<in> R \<and> (s,t) \<in> rstep_r_c_s (l,r) C \<sigma>)" (is "?left = ?right")
proof
  assume ?left
  then obtain l r p \<sigma> where step: "(s,t) \<in> rstep_r_p_s R (l,r) p \<sigma>"
    unfolding rstep_iff_rstep_r_p_s by blast
  obtain D where D: "D = ctxt_of_pos_term p s" by auto
  with step have Rrule: "(l,r) \<in> R" and s: "s = D\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = D\<langle>r \<cdot> \<sigma>\<rangle>" 
    unfolding rstep_r_p_s_def by (force simp: Let_def)+
  thus ?right unfolding rstep_r_c_s_def by auto
next
  assume ?right
  from this obtain l r C \<sigma> where "(l,r) \<in> R \<and> (s,t) \<in> rstep_r_c_s (l,r) C \<sigma>" by auto
  hence rule: "(l,r) \<in> R" and s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>"
    unfolding rstep_r_c_s_def by auto
  show ?left unfolding rstep_CC_CS
    by (simp add: s t, rule ctxt_closureI, rule subst_closureI, rule rule)
qed

lemma rstep_subset_characterization:
  "(rstep R \<subseteq> rstep S) = (\<forall> l r. (l,r) \<in> R \<longrightarrow> (\<exists> l' r' C \<sigma> . (l',r') \<in> S \<and> l = C\<langle>l' \<cdot> \<sigma>\<rangle> \<and> r = C\<langle>r' \<cdot> \<sigma>\<rangle>))" (is "?left = ?right")
proof
  assume ?right 
  show ?left
  proof (rule subsetI, simp add: split_paired_all)
    fix s t
    assume "(s,t) \<in> rstep R"
    then obtain l r C \<sigma> where step: "(l,r) \<in> R \<and> (s,t) \<in> rstep_r_c_s (l,r) C \<sigma>"
      unfolding rstep_iff_rstep_r_c_s by best
    hence Rrule: "(l,r) \<in> R" and s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>" 
      unfolding rstep_r_c_s_def by (force)+
    from Rrule `?right` obtain l' r' C' \<sigma>' where Srule: "(l',r') \<in> S" and l: "l = C'\<langle>l' \<cdot> \<sigma>'\<rangle>" and r: "r = C'\<langle>r' \<cdot> \<sigma>'\<rangle>"
      by (force simp: Let_def)+
    let ?D = "C \<circ>c (C' \<cdot>c \<sigma>)"
    let ?sig = "\<sigma>' \<circ>s \<sigma>"
    have s2: "s = ?D\<langle>l' \<cdot> ?sig\<rangle>" by (simp add: s l)
    have t2: "t = ?D\<langle>r' \<cdot> ?sig\<rangle>" by (simp add: t r)
    from s2 t2 have sStep: "(s,t) \<in> rstep_r_c_s (l',r') ?D ?sig" unfolding rstep_r_c_s_def by force
    with Srule show "(s,t) \<in> rstep S" by (simp only: rstep_iff_rstep_r_c_s, blast)
  qed
next
  assume ?left
  show ?right
  proof (rule ccontr)
    assume "\<not> ?right"
    from this obtain l r where "(l,r) \<in> R" and cond: "\<forall> l' r' C \<sigma>. (l',r') \<in> S \<longrightarrow> (l \<noteq> C\<langle>l' \<cdot> \<sigma>\<rangle> \<or> r \<noteq> C\<langle>r' \<cdot> \<sigma>\<rangle>)" by blast
    hence "(l,r) \<in> rstep R" by blast
    with `?left` have "(l,r) \<in> rstep S" by auto
    with cond show False by (simp only: rstep_iff_rstep_r_c_s, unfold rstep_r_c_s_def, force)
  qed
qed

lemma rstep_preserves_funas_terms:
  assumes "funas_trs R \<subseteq> F" and "funas_term s \<subseteq> F" and "(s,t) \<in> rstep R" and "wf_trs R"
  shows "funas_term t \<subseteq> F"
proof -
  from `(s,t) \<in> rstep R` obtain l r C \<sigma> where R: "(l,r) \<in> R"
    and s: "s = C\<langle>l\<cdot>\<sigma>\<rangle>" and t: "t = C\<langle>r\<cdot>\<sigma>\<rangle>" using rstep_imp_C_s_r[of s t R] by fast
  from `funas_trs R \<subseteq> F` and R have "funas_term r \<subseteq> F"
    unfolding funas_defs [abs_def] by force
  from `wf_trs R` and R have "vars_term r \<subseteq> vars_term l" unfolding wf_trs_def by simp
  have "funas_term s = funas_ctxt C \<union> funas_term(l\<cdot>\<sigma>)" unfolding s by simp
  hence funas_s: "funas_term s = funas_ctxt C \<union> funas_term l \<union> \<Union>(funas_term ` \<sigma> ` vars_term l)"
    using funas_term_subst by auto
  have "funas_term t = funas_ctxt C \<union> funas_term(r\<cdot>\<sigma>)" unfolding t by simp
  hence "funas_term t = funas_ctxt C \<union> funas_term r \<union> \<Union>(funas_term ` \<sigma> ` vars_term r)"
    using funas_term_subst by auto
  hence "funas_term t \<subseteq> funas_ctxt C \<union> funas_term r \<union> \<Union>(funas_term ` \<sigma> ` vars_term l)"
    using `vars_term r \<subseteq> vars_term l` by auto
  with `funas_term r \<subseteq> F` `funas_term s \<subseteq> F` have "funas_term t \<subseteq> F" unfolding funas_s by force
  thus ?thesis by auto
qed

lemma rsteps_preserve_funas_terms:
  assumes F: "funas_trs R \<subseteq> F" and s: "funas_term s \<subseteq> F"
    and steps: "(s,t) \<in> (rstep R)^*" and wf: "wf_trs R"
  shows "funas_term t \<subseteq> F"
  using steps 
proof (induct)
  case base thus ?case by (rule s)
next
  case (step t u)
  show ?case by (rule rstep_preserves_funas_terms[OF F step(3) step(2) wf])
qed

lemma no_Var_rstep [simp]:
  assumes "wf_trs R" and "(Var x, t) \<in> rstep R" shows "False"
  using assms unfolding wf_trs_def rstep_iff_rstep_r_p_s rstep_r_p_s_def Let_def by force

lemma lhs_wf:
  assumes R: "(l, r) \<in> R" and "funas_trs R \<subseteq> F"
  shows "funas_term l \<subseteq>  F"
  using assms by (force simp: funas_trs_def funas_rule_def)

lemma rhs_wf:
  assumes R: "(l, r) \<in> R" and "funas_trs R \<subseteq> F"
  shows "funas_term r \<subseteq> F"
  using assms by (force simp: funas_trs_def funas_rule_def)

lemma supt_map_term [intro]:
  assumes "t \<rhd> s"
  shows "map_term fg t \<rhd> map_term fg s"
using assms
proof (induct)
  case (arg s ss f)
  hence "map_term fg s \<in> set(map (map_term fg) ss)" by simp
  thus ?case unfolding map_term.simps by (rule supt.arg)
next
  case (subt s ss u f)
  hence "map_term fg s \<in> set(map (map_term fg) ss)" by simp
  with `map_term fg s \<rhd> map_term fg u` show ?case
    unfolding map_term.simps by (metis supt.subt)
qed

lemma nondef_root_imp_arg_step:
  assumes "(Fun f ss, t) \<in> rstep R"
    and wf: "\<forall>(l, r)\<in>R. is_Fun l"
    and "(f, length ss) \<notin> defs R"
  shows "\<exists>i<length ss. (ss ! i, t |_ <i>) \<in> rstep R
    \<and> t = Fun f (take i ss @ (t |_ <i>) # drop (Suc i) ss)"
proof -
  from assms obtain l r p \<sigma>
    where rstep_r_p_s: "(Fun f ss, t) \<in> rstep_r_p_s R (l, r) p \<sigma>"
    unfolding rstep_iff_rstep_r_p_s by auto
  let ?C = "ctxt_of_pos_term p (Fun f ss)"
  from rstep_r_p_s have "p \<in> poss (Fun f ss)" and "(l, r) \<in> R"
    and "?C\<langle>l \<cdot> \<sigma>\<rangle> = Fun f ss" and "?C\<langle>r \<cdot> \<sigma>\<rangle> = t" unfolding rstep_r_p_s_def Let_def by auto
  have "\<exists>i q. p = i<#q"
  proof (cases p)
    case PCons thus ?thesis by auto
  next
    case Empty
    have "?C = \<box>" unfolding Empty by simp
    with `?C\<langle>l\<cdot>\<sigma>\<rangle> = Fun f ss` have "l\<cdot>\<sigma> = Fun f ss" by simp
    have "\<forall>(l,r)\<in>R. \<exists>f ss. l = Fun f ss"
    proof (intro ballI impI)
      fix lr assume "lr \<in> R"
      with wf have "\<forall>x. fst lr \<noteq> Var x" by auto
      hence "\<exists>f ss. (fst lr) = Fun f ss" by (cases "fst lr") auto
      thus "(\<lambda>(l,r). \<exists>f ss. l = Fun f ss) lr" by auto
    qed
    with `(l,r) \<in> R` obtain g ts where "l = Fun g ts" unfolding wf_trs_def by best
    with `l\<cdot>\<sigma> = Fun f ss` have "f = g" by auto
    from `l\<cdot>\<sigma> = Fun f ss` and `l = Fun g ts` and `(l, r) \<in> R`
      have "(f, length ss) \<in> defs R" unfolding defs_def defined_def `f = g` by auto
    with `(f,length ss) \<notin> defs R` show ?thesis by simp
  qed
  then obtain i q where "p = i<#q" by auto
  from `p \<in> poss(Fun f ss)` have "i < length ss" and "q \<in> poss(ss!i)" unfolding `p = i<#q` by auto
  let ?D = "ctxt_of_pos_term q (ss!i)"
  have C: "?C = More f (take i ss) ?D (drop (Suc i) ss)" unfolding `p = i<#q` by auto
  from `?C\<langle>l\<cdot>\<sigma>\<rangle> = Fun f ss` have "take i ss@?D\<langle>l\<cdot>\<sigma>\<rangle>#drop (Suc i) ss = ss" unfolding C by auto
  hence "(take i ss@?D\<langle>l\<cdot>\<sigma>\<rangle>#drop (Suc i) ss)!i = ss!i" by simp
  with `i < length ss` have "?D\<langle>l\<cdot>\<sigma>\<rangle> = ss!i" using nth_append_take[where xs = ss and i = i] by auto
  have t: "t = Fun f (take i ss@?D\<langle>r\<cdot>\<sigma>\<rangle>#drop (Suc i) ss)" unfolding `?C\<langle>r\<cdot>\<sigma>\<rangle> = t`[symmetric] C by simp
  from `i < length ss` have "t|_<i> = ?D\<langle>r\<cdot>\<sigma>\<rangle>" unfolding t unfolding subt_at.simps using nth_append_take[where xs = ss and i = i] by auto
  from `q \<in> poss(ss!i)` and `(l,r) \<in> R` 
    and `?D\<langle>l\<cdot>\<sigma>\<rangle> = ss!i` and `t|_<i> = ?D\<langle>r\<cdot>\<sigma>\<rangle>`[symmetric]
    have "(ss!i,t|_<i>) \<in> rstep_r_p_s R (l,r) q \<sigma>" unfolding rstep_r_p_s_def Let_def by auto
  hence "(ss!i,t|_<i>) \<in> rstep R" unfolding rstep_iff_rstep_r_p_s by auto
  from `i < length ss` and this and t show ?thesis unfolding `t|_<i> = ?D\<langle>r\<cdot>\<sigma>\<rangle>`[symmetric] by auto
qed

lemma nondef_root_imp_arg_steps:
  assumes "(Fun f ss,t) \<in> (rstep R)^*"
      and wf: "\<forall>(l, r)\<in>R. is_Fun l"
      and "(f,length ss) \<notin> defs R"
  shows "\<exists>ts. length ts = length ss \<and> t = Fun f ts \<and> (\<forall>i<length ss. (ss!i,ts!i) \<in> (rstep R)^*)"
proof -
  from assms obtain n where "(Fun f ss,t) \<in> (rstep R)^^n"
    using rtrancl_imp_relpow by best
  thus ?thesis
  proof (induct n arbitrary: t)
    case 0 thus ?case by auto
  next
    case (Suc n)
    then obtain u where "(Fun f ss,u) \<in> (rstep R)^^n" and "(u,t) \<in> rstep R" by auto
    with Suc obtain ts where IH1: "length ts = length ss" and IH2: "u = Fun f ts" and IH3: "\<forall>i<length ss. (ss!i,ts!i) \<in> (rstep R)^*" by auto
    from `(u,t) \<in> rstep R` have "(Fun f ts,t) \<in> rstep R" unfolding `u = Fun f ts` .
    from nondef_root_imp_arg_step[OF this wf `(f,length ss) \<notin> defs R`[simplified IH1[symmetric]]]
      obtain j where "j < length ts"
      and "(ts!j,t|_<j>) \<in> rstep R"
      and B: "t = Fun f (take j ts@(t|_<j>)#drop (Suc j) ts)" (is "t = Fun f ?ts") by auto
    from `j < length ts` have "length ?ts = length ts" by auto
    hence A: "length ?ts = length ss" unfolding `length ts = length ss` .
    have C: "\<forall>i<length ss. (ss!i,?ts!i) \<in> (rstep R)^*"
    proof (intro allI, intro impI)
      fix i assume "i < length ss"
        from `i < length ss` and IH3 have "(ss!i,ts!i) \<in> (rstep R)^*" by auto
      have "i = j \<or> i \<noteq> j" by auto
      thus "(ss!i,?ts!i) \<in> (rstep R)^*"
      proof
        assume "i = j"
        from `j < length ts` have "j \<le> length ts" by simp
        from nth_append_take[OF this] have "?ts!j = t|_<j>" by simp
        from `(ts!j,t|_<j>) \<in> rstep R` have "(ts!i,t|_<i>) \<in> rstep R" unfolding `i = j` .
        with `(ss!i,ts!i) \<in> (rstep R)^*` show ?thesis unfolding `i = j` unfolding `?ts!j = t|_<j>` by auto
      next
        assume "i \<noteq> j"
        from `i < length ss` have "i \<le> length ts" unfolding `length ts = length ss` by simp
        from `j < length ts` have "j \<le> length ts" by simp
        from nth_append_take_drop_is_nth_conv[OF `i \<le> length ts` `j \<le> length ts` `i \<noteq> j`]
          have "?ts!i = ts!i" by simp
        with `(ss!i,ts!i) \<in> (rstep R)^*` show ?thesis by auto
      qed
    qed
    from A and B and C show ?case by blast
  qed
qed

lemma rstep_imp_nrrstep:
  assumes "is_Fun s" and "\<not> defined R (the (root s))" and "\<forall>(l,r)\<in>R. is_Fun l"
    and "(s, t) \<in> rstep R"
  shows "(s, t) \<in> nrrstep R"
proof -
  from `is_Fun s` obtain f ss where s: "s = Fun f ss" by (cases s) auto
  from assms have undef: "(f, length ss) \<notin> defs R" by (simp add: defs_def s)
  from assms have non_var: "\<forall>(l, r)\<in>R. is_Fun l" by auto
  from nondef_root_imp_arg_step[OF `(s, t) \<in> rstep R`[unfolded s] non_var undef]
    obtain i where "i < length ss" and step: "(ss ! i, t |_ <i>) \<in> rstep R"
      and t: "t = Fun f (take i ss @ (t |_ <i>) # drop (Suc i) ss)" by auto
  from step obtain C l r \<sigma> where "(l, r) \<in> R" and lhs: "ss ! i = C\<langle>l \<cdot> \<sigma>\<rangle>"
    and rhs: "t |_ <i> = C\<langle>r \<cdot> \<sigma>\<rangle>" by auto
  let ?C = "More f (take i ss) C (drop (Suc i) ss)"
  have "(l, r) \<in> R" by fact
  moreover have "?C \<noteq> \<box>" by simp
  moreover have "s = ?C\<langle>l \<cdot> \<sigma>\<rangle>"
  proof -
    have "s = Fun f (take i ss @ ss!i # drop (Suc i) ss)"
      using id_take_nth_drop[OF `i < length ss`] unfolding s by simp
    also have "\<dots> = ?C\<langle>l \<cdot> \<sigma>\<rangle>" by (simp add: lhs)
   finally show ?thesis .
  qed
  moreover have "t = ?C\<langle>r \<cdot> \<sigma>\<rangle>"
  proof -
    have "t = Fun f (take i ss @ t |_ <i> # drop (Suc i) ss)" by fact
    also have "\<dots> = Fun f (take i ss @ C\<langle>r \<cdot> \<sigma>\<rangle> # drop (Suc i) ss)" by (simp add: rhs)
    finally show ?thesis by simp
  qed
  ultimately show "(s, t) \<in> nrrstep R" unfolding nrrstep_def' by blast
qed

lemma rsteps_imp_nrrsteps:
  assumes "is_Fun s" and "\<not> defined R (the (root s))"
    and no_vars: "\<forall>(l, r)\<in>R. is_Fun l"
    and "(s, t) \<in> (rstep R)^*"
  shows "(s, t) \<in> (nrrstep R)^*"
using `(s, t) \<in> (rstep R)^*`
proof (induct)
  case base show ?case by simp
next
  case (step u v)
  from assms obtain f ss where s: "s = Fun f ss" by (induct s) auto
  from nrrsteps_preserve_root[OF `(s, u) \<in> (nrrstep R)^*`[unfolded s]]
    obtain ts where u: "u = Fun f ts" by auto
  from nrrsteps_equiv_num_args[OF `(s, u) \<in> (nrrstep R)^*`[unfolded s]]
    have len: "length ss = length ts" unfolding u by simp
  have "is_Fun u" by (simp add: u)
  have undef: "\<not> defined R (the (root u))"
    using `\<not> defined R (the (root s))`
    unfolding s u by (simp add: len)
  have "(u, v) \<in> nrrstep R"
    using rstep_imp_nrrstep[OF `is_Fun u` undef no_vars] step
    by simp
  with step show ?case by auto
qed

lemma left_var_imp_not_SN:
  fixes R :: "('f,'v)trs" and t :: "('f, 'v) term"
  assumes "(Var y, r) \<in> R" (is "(?y, _) \<in> _")
  shows "\<not> (SN_on (rstep R) {t})"
proof (rule steps_imp_not_SN_on)
  fix t :: "('f,'v)term"
  let ?yt = "subst y t"
  show "(t, r \<cdot> ?yt) \<in> rstep R" 
    by (rule rstepI[OF assms, where C = \<box> and \<sigma> = ?yt], auto simp: subst_def)
qed

lemma not_SN_subt_imp_not_SN:
  assumes ctxt: "ctxt_closed R" and SN: "\<not> SN_on R {t}" and sub: "s \<unrhd> t"
  shows "\<not> SN_on R {s}"
  using restrict_SN_imp_SN_subt[OF ctxt _ sub] SN
  by auto

lemma Some_root_imp_Fun: "root t = Some (f, n) \<Longrightarrow> \<exists>ss. length ss = n \<and> t = Fun f ss"
  by (induct t) auto

lemma map_rule_power:
  fixes f :: "'f \<Rightarrow> 'f" 
  shows "((map_rule f) ^^ n) = map_rule (f ^^ n)"
proof (rule sym, intro ext, clarify) 
  fix  l r :: "('f,'v)term"
  show "map_rule (f ^^ n) (l,r) = (map_rule f ^^ n) (l,r)"
  proof (induct n)
    case 0
    show ?case
      by (simp, unfold id_def[symmetric] map_term_ident, simp)
  next
    case (Suc n)
    have "map_rule (f ^^ Suc n) (l,r) = map_rule f (map_rule (f ^^ n) (l,r))" 
      by (simp add: map_term_comp)
    also have "\<dots> = map_rule f ((map_rule f ^^ n) (l,r))" unfolding Suc ..
    also have "\<dots> = (map_rule f ^^ (Suc n)) (l,r)" by simp
    finally show ?case .
  qed
qed

lemma map_trs_power: fixes f :: "'f \<Rightarrow> 'f" 
  shows "((map_trs f) ^^ n) = map_trs (f ^^ n)"
proof (rule sym, rule ext, unfold map_trs.simps) 
  fix R :: "('f,'v)trs"
  obtain g where g: "(map_rule f :: ('f,'v)rule \<Rightarrow> ('f,'v)rule) = g" by simp
  have "map_rule (f ^^ n) ` R = (map_rule f ^^ n) ` R" unfolding map_rule_power ..
  also have "\<dots> = ((\<lambda> R. map_trs f R) ^^ n) R" unfolding map_trs.simps g
    by (induct n, force+)
  finally show "map_rule (f ^^ n) ` R = (map_trs f ^^ n) R" .
qed


fun
  sharp_term :: "('f \<Rightarrow> 'f) \<Rightarrow> ('f, 'v)term \<Rightarrow> ('f,'v) term"
where
  "sharp_term shp (Var x) = Var x" |
  "sharp_term shp (Fun f ss) = Fun (shp f) ss"

definition
  DP :: "('f \<Rightarrow> 'f) \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs"
where
  "DP shp R = {(s,t). \<exists>l r h us. s = sharp_term shp l \<and> t = sharp_term shp (Fun h us)
     \<and> (l,r) \<in> R \<and> r \<unrhd> (Fun h us) \<and> defined R (h, length us) \<and> \<not> l \<rhd> Fun h us}"

lemma sharp_term_subst_distr[simp]:
  "sharp_term shp (Fun f ss \<cdot> \<sigma>) = sharp_term shp (Fun f ss) \<cdot> \<sigma>"
by simp
 
lemma nrrstep_imp_sharp_rstep:
  assumes "(s, t) \<in> nrrstep R"
  shows "(sharp_term shp s, sharp_term shp t) \<in> rstep R"
proof -
  from assms obtain l r i ps \<sigma> where "(s,t) \<in> rstep_r_p_s R (l,r) (i<#ps) \<sigma>" unfolding nrrstep_def by best
  hence "let C = ctxt_of_pos_term (i<#ps) s in (i<#ps) \<in> poss s \<and> (l,r) \<in> R \<and> C\<langle>l\<cdot>\<sigma>\<rangle> = s \<and> C\<langle>r\<cdot>\<sigma>\<rangle> = t"
  (is "let C = ?C in _") unfolding rstep_r_p_s_def by simp
  hence "(l,r) \<in> R" and s: "?C\<langle>l\<cdot>\<sigma>\<rangle> = s" and t: "?C\<langle>r\<cdot>\<sigma>\<rangle> = t" unfolding Let_def by auto
  from `(s,t) \<in> nrrstep R` obtain f ss where "s = Fun f ss" using nrrstep_imp_Fun by best
  hence C: "?C = More f (take i ss) (ctxt_of_pos_term ps (ss!i)) (drop (Suc i) ss)"
    (is "?C = More f ?ss1 ?D ?ss2") by simp
  let ?C' = "More (shp f) ?ss1 ?D ?ss2"
  from `(l,r) \<in> R` have "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> CS R" by (rule subst_closureI)
  hence "(?C'\<langle>l \<cdot> \<sigma>\<rangle>, ?C'\<langle>r \<cdot> \<sigma>\<rangle>) \<in> CC(CS R)"
    by (rule ctxt_closureI)
  hence R_step: "(?C'\<langle>l \<cdot> \<sigma>\<rangle>, ?C'\<langle>r \<cdot> \<sigma>\<rangle>) \<in> rstep R"
    by (simp add: rstep_CC_CS)
  from `?C\<langle>l\<cdot>\<sigma>\<rangle> = s` have "sharp_term shp s = sharp_term shp (?C\<langle>l \<cdot> \<sigma>\<rangle>)" by simp
  hence s: "sharp_term shp s = ?C'\<langle>l \<cdot> \<sigma>\<rangle>" using C by simp
  from `?C\<langle>r\<cdot>\<sigma>\<rangle> = t` have "sharp_term shp t = sharp_term shp (?C\<langle>r\<cdot>\<sigma>\<rangle>)" by simp
  hence t: "sharp_term shp t = ?C'\<langle>r \<cdot> \<sigma>\<rangle>" by (simp add: C)
  from R_step show ?thesis unfolding s t .
qed

lemma nrrsteps_imp_sharp_rsteps:
  "(s,t) \<in> (nrrstep R)^* \<Longrightarrow> (sharp_term shp s, sharp_term shp t) \<in> (rstep R)^*"
proof (induct rule: rtrancl_induct)
  case base thus ?case by simp
next
  case (step a b)
  from `(a,b) \<in> nrrstep R` have "(sharp_term shp a, sharp_term shp b) \<in> rstep R"
    by (rule nrrstep_imp_sharp_rstep)
   with step show ?case by auto
qed

lemma finiteR_imp_finiteDP:
  assumes "finite R"
  shows "finite (DP shp R)"
proof -
 from assms have fS:
  "finite{(l,r,u). (\<exists>h us. u = Fun h us \<and> (l,r) \<in> R \<and> r \<unrhd> u \<and> defined R (h, length us)\<and> \<not>(l \<rhd> u))}"
  (is "finite ?S")
  by (rule finiteR_imp_finiteDP')
 let ?f = "\<lambda>(x,y,z). (sharp_term shp x, sharp_term shp z)"
 have eq1: "(\<Union>y\<in>?S. {x. x = ?f y}) = ?f ` ?S" by blast
 with fS have "finite(?f ` ?S)" by auto
 have "DP shp R = ?f ` ?S" (is "?DP = ?T")
 proof
  show "?DP \<subseteq> ?T"
  proof
   fix x assume "x \<in> ?DP"
   then obtain l r h us where "fst x = sharp_term shp l"
     and "snd x = sharp_term shp (Fun h us)" "(l,r) \<in> R" "r \<unrhd> Fun h us" "defined R (h, length us)" " \<not>(l \<rhd> Fun h us)" unfolding DP_def split_def by auto
   hence "(l,r,Fun h us) \<in> ?S" by auto
   thus "x \<in> ?T" unfolding eq1[symmetric]
   proof (rule UN_I)
    have "x = (fst x, snd x)" by simp
    hence "x = (sharp_term shp l, sharp_term shp (Fun h us))"
      unfolding `fst x = sharp_term shp l` `snd x = sharp_term shp (Fun h us)` .
    thus "x \<in> {x. x = ?f (l,r,Fun h us)}" by auto
   qed
  qed
 next
  show "?T \<subseteq> ?DP"
  proof
   fix x assume "x \<in> ?T"
   hence "x \<in> (\<Union>y\<in>?S. {x. x = ?f y})" unfolding eq1 .
   then obtain y where "y \<in> ?S" and "x = ?f y" by fast
   then obtain l r h us where "y = (l,r,Fun h us)" and dp: "(l,r) \<in> R" "r \<unrhd> Fun h us" "defined R (h, length us)" "\<not>(l \<rhd> Fun h us)" by blast
   with `x = ?f y` have "x = (sharp_term shp l, sharp_term shp (Fun h us))" by auto
   thus "x \<in> ?DP" using dp unfolding DP_def by auto
  qed
 qed
 with fS show ?thesis by auto
qed

lemma vars_sharp_eq_vars[simp]: "vars_term (sharp_term shp t) = vars_term t"
  by (induct t, auto)

lemma wf_trs_imp_wf_DP:
 assumes "wf_trs R"
 shows "wf_trs (DP shp R)"
unfolding wf_trs_def
proof (intro allI impI)
 fix s t
 assume "(s,t) \<in> DP shp R"
 then obtain l r h us where "s = sharp_term shp l" "t = sharp_term shp (Fun h us)" "(l,r) \<in> R" "r \<unrhd> (Fun h us)" "\<not>(l \<rhd> Fun h us)"
  unfolding DP_def by auto
 from `wf_trs R` and `(l,r) \<in> R` have "\<exists>f ss. l = Fun f ss"
  "vars_term r \<subseteq> vars_term l" unfolding wf_trs_def by auto
 from `\<exists>f ss. l = Fun f ss` obtain f ss where "l = Fun f ss" by auto
 hence "s = Fun (shp f) ss" unfolding `s = sharp_term shp l` by simp
 hence "\<exists>f ss. s = Fun f ss" by auto
 from `r \<unrhd> Fun h us` have "vars_term(Fun h us) \<le> vars_term r" by (induct rule: supteq.induct) auto
 hence "vars_term t \<subseteq> vars_term s" unfolding `s = sharp_term shp l`
   and `t = sharp_term shp (Fun h us)` vars_sharp_eq_vars using `vars_term r \<le> vars_term l` by simp
 from `\<exists>f ss. s = Fun f ss` `vars_term t \<subseteq> vars_term s`
 show "(\<exists>f ss. s = Fun f ss) \<and> vars_term t \<subseteq> vars_term s" by simp
qed

lemma sharp_eq_imp_eq: assumes "inj shp" shows "sharp_term shp s = sharp_term shp t \<Longrightarrow> s = t"
proof (cases s)
  case (Var x) 
  assume "sharp_term shp s = sharp_term shp t" with Var show ?thesis by (induct t) auto
next
  case (Fun f ss)
  assume "sharp_term shp s = sharp_term shp t" 
  with Fun have "sharp_term shp (Fun f ss) = sharp_term shp t" by simp
  then obtain g ts where t: "t = Fun g ts" by (induct t) auto
  with `sharp_term shp s = sharp_term shp t` have "Fun (shp f) ss = Fun (shp g) ts" unfolding `s = Fun f ss` `t = Fun g ts`
    by simp
  hence "f = g" and "ss = ts" using `inj shp`[unfolded inj_on_def] by auto
  thus ?thesis unfolding Fun t by simp
qed


fun
  idx :: "(nat \<Rightarrow> nat) \<Rightarrow> nat \<Rightarrow> nat"
where
  "idx f 0 = 0" |
  "idx f (Suc i) = Suc (f (idx f i))"

(* TODO: this is subsumed by Tinf-lemma for relative rewriting *)
lemma SN_on_subterms_imp_SN_on_nrrstep:
  assumes "\<And>s. s \<lhd> t \<Longrightarrow> SN_on (rstep R) {s}"
  shows "SN_on (nrrstep R) {t}"
proof
  fix S assume "S 0 \<in> {t}" and "chain (nrrstep R) S"
  hence "\<forall>i. \<exists>f ss ts. S i = Fun f ss \<and> S (Suc i) = Fun f ts \<and> length ss = length ts
    \<and> (\<exists>j<length ss. (ss!j, ts!j) \<in> rstep R \<and> (\<forall>k<length ss. k \<noteq> j \<longrightarrow> ss!k = ts!k))"
    (is "\<forall>i. \<exists>f ss ts. ?P i f ss ts")
    using nrrstep_args by fast
  from choice[OF this] obtain f where "\<forall>i. \<exists>ss ts. ?P i (f i) ss ts" by best
  from choice[OF this] obtain ss where "\<forall>i. \<exists>ts. ?P i (f i) (ss i) ts" by best
  from choice[OF this] obtain ts where P: "\<forall>i. ?P i (f i) (ss i) (ts i)" by best
  moreover have "\<forall>i. f i = f (Suc i) \<and> ts i = ss (Suc i)"
  proof
    fix i
    from P have "S (Suc i) = Fun (f (Suc i)) (ss (Suc i))" by simp
    moreover from P have "S (Suc i) = Fun (f i) (ts i)" by blast+
    ultimately show "f i = f (Suc i) \<and> ts i = ss (Suc i)" by simp
  qed
  ultimately have "\<forall>i. ?P i (f i) (ss i) (ss (Suc i))" by auto
  hence "\<forall>i. length (ss i) = length (ss (Suc i))"
    and "\<forall>i. \<exists>j<length (ss i). (ss i ! j, ss (Suc i) ! j) \<in> rstep R
    \<and> (\<forall>k<length (ss i). k \<noteq> j \<longrightarrow> ss i ! k = ss (Suc i) ! k)" by blast+
  from choice[OF this(2)] obtain \<pi> where pi: "\<forall>i. \<pi> i < length (ss i)
    \<and> (ss i ! \<pi> i, ss (Suc i) ! \<pi> i) \<in> rstep R
    \<and> (\<forall>k<length (ss i). k \<noteq> \<pi> i \<longrightarrow> ss i ! k = ss (Suc i) ! k)" by blast+
  hence chain: "\<forall>i. (ss i ! \<pi> i, ss (Suc i) ! \<pi> i) \<in> rstep R" by simp
  have len: "\<forall>i. length (ss i) = length (ss 0)"
  proof
    fix i show "length (ss i) = length (ss 0)"
    proof (induct i)
      case 0 show ?case ..
    next
      case (Suc i) thus ?case using `\<forall>i. length (ss i) = length (ss (Suc i))` by simp
    qed
  qed
  have "\<forall>i. \<pi> i < length (ss 0)"
  proof
    fix i
    have "length (ss i) = length (ss 0)" using len by simp
    moreover from pi have "\<pi> i < length (ss i)" by simp
    ultimately show "\<pi> i < length (ss 0)" by simp
  qed
  hence "range \<pi> \<subseteq> {i. i < length (ss 0)}" by auto
  moreover have "finite {i. i < length (ss 0)}" by simp
  ultimately have "finite (range \<pi>)" by (rule finite_subset)
  from finite_range[OF this] obtain j where j: "\<forall>i. \<exists>k\<ge>i. \<pi> k = j" by (auto simp: INFM_nat_le)
  from choice[OF this] obtain q where q: "\<forall>i. q i \<ge> i \<and> \<pi> (q i) = j" by auto
  have j_len: "\<forall>i. j < length (ss i)"
  proof
    fix i
    from j obtain k where "k \<ge> i" and "\<pi> k = j" by best
    moreover from pi have "\<pi> k < length (ss k)" by simp
    ultimately have "j < length (ss k)" by simp
    moreover from len have "length (ss k) = length (ss 0)" by simp
    moreover from len have "length (ss i) = length (ss 0)" by simp
    ultimately show "j < length (ss i)" by simp
  qed
  from chain have step: "\<forall>i. (ss (q i) ! \<pi> (q i), ss (Suc (q i)) ! \<pi> (q i)) \<in> rstep R" by simp
  have refl_seq: "\<forall>i. (ss i ! j, ss (Suc i) ! j) \<in> (rstep R)^="
  proof
    fix i show "(ss i ! j, ss (Suc i) ! j) \<in> (rstep R)^="
    proof (cases "\<pi> i = j")
      assume "\<pi> i = j"
      moreover from pi have "(ss i ! \<pi> i, ss (Suc i) ! \<pi> i) \<in> rstep R" by simp
      ultimately show ?thesis by simp
    next
      assume "\<pi> i \<noteq> j"
      moreover from `\<forall>i. j < length (ss i)` have "j < length (ss i)" ..
      moreover from pi have "j < length (ss i) \<and> j \<noteq> \<pi> i \<longrightarrow> ss i ! j = ss (Suc i) ! j" by simp
      ultimately show ?thesis by simp
    qed
  qed
  let ?s = "\<lambda>i. ss (idx q i) ! j"
  have "\<forall>i. (ss i ! j, ss (Suc (q i)) ! j) \<in> (rstep R)^+"
  proof
    fix i
    from q have "q i \<ge> i" by simp
    from chain_imp_rtrancl[OF refl_seq this] have "(ss i ! j, ss (q i) ! j) \<in> (rstep R)^*" by simp
    moreover from step have "(ss (q i) ! j, ss (Suc (q i)) ! j) \<in> rstep R" by (simp add: q)
    ultimately show "(ss i ! j, ss (Suc (q i)) ! j) \<in> (rstep R)^+" by simp
  qed
  hence "\<forall>i. (?s i, ?s (Suc i)) \<in> (rstep R)^+" by simp
  hence "\<not> SN_on (rstep R) {?s 0}" using SN_on_trancl by best
  moreover have "?s 0 \<lhd> t"
  proof -
    from P have t: "t = Fun (f 0) (ss 0)" by (simp add: singletonD[OF `S 0 \<in> {t}`, symmetric])
    have "j < length (ss 0)" using j_len by simp
    hence "?s 0 \<in> set (ss 0)" by simp
    thus ?thesis unfolding t ..
  qed
  moreover from assms have "\<forall>s\<lhd>t. SN_on (rstep R) {s}" by simp
  ultimately show False by simp
qed

hide_const idx

lemma SN_args_imp_SN :
  assumes SN: "\<And>s. s \<in> set ss \<Longrightarrow> SN_on (rstep R) {s}"
    and nvar: "\<forall>(l, r)\<in>R. is_Fun l"
    and ndef: "(f, length ss) \<notin> defs R"
  shows "SN_on (rstep R) {Fun f ss}"
proof (rule ccontr)
  assume nSN: "\<not> SN_on (rstep R) {Fun f ss}"
  have SN: "\<And>s. Fun f ss \<rhd> s \<Longrightarrow> SN_on (rstep R) {s}"
  proof -
    fix s
    assume "Fun f ss \<rhd> s"
    then obtain C where "C \<noteq> \<box>" and "Fun f ss = C\<langle>s\<rangle>" ..
    then obtain bef D aft where "ss = bef @ D\<langle>s\<rangle> # aft" by (cases C, auto)
    hence "D\<langle>s\<rangle> \<in> set ss" by simp
    from SN[OF this] not_SN_subt_imp_not_SN[OF ctxt_closed_rstep _ ctxt_imp_supteq[of D s]]
    show "SN_on (rstep R) {s}" by auto
  qed
  let ?t = "Fun f ss"
  let ?T = "{t. \<not> SN_on (rstep R) {t} \<and> (\<forall>s\<lhd>t. SN_on (rstep R) {s})}"
  have t: "?t \<in> ?T" using nSN SN by best
  from nSN obtain S where "S 0 = ?t" and chain: "chain (rstep R) S" by auto
  have "\<exists>i. (S 0, S i) \<in> (nrrstep R)^* \<and> (S i, S (Suc i)) \<in> rrstep R"
  proof -
    from t have "SN_on (nrrstep R) {S 0}"
      by (simp add: SN_on_subterms_imp_SN_on_nrrstep `S 0 = ?t`)
    with chain chain_Un_SN_on_imp_first_step[of S "rrstep R" "nrrstep R"]
      obtain j where first_step: "(S j, S (Suc j)) \<in> rrstep R"
      and nrsteps: "\<forall>i<j. (S i, S (Suc i)) \<in> nrrstep R"
      by (auto simp: rstep_iff_rrstep_or_nrrstep)
    from nrsteps have "(S 0, S j) \<in> (nrrstep R)^*"
      by (induct j) auto
    with first_step show ?thesis by auto
  qed
  then obtain j where nrseq: "(S 0, S j) \<in> (nrrstep R)^*"
    and rrstep: "(S j, S (Suc j)) \<in> rrstep R" by auto
  from rrstep_imp_rule_subst[OF rrstep]
    obtain l r \<sigma> where "(l, r) \<in> R" and "l\<cdot>\<sigma> = S j" and "r\<cdot>\<sigma> = S (Suc j)" by auto
  from nrseq have "(?t, S j) \<in> (nrrstep R)^*" by (simp add: `S 0 = ?t`)
  moreover from nrrsteps_preserve_root[OF this] nrrsteps_equiv_num_args[OF this] obtain ls'
    where "S j = Fun f ls'" and "length ss = length ls'" by auto
  ultimately have "l\<cdot>\<sigma> = Fun f ls'"
    by (auto simp: `l\<cdot>\<sigma> = S j`) 
  from nvar `(l,r) \<in> R` obtain g ls 
    where "l = Fun g ls" by (cases l, force+)
  with `l\<cdot>\<sigma> = Fun f ls'` `(l,r) \<in> R` `length ss = length ls'` ndef have "(Fun f ls,r) \<in> R" and "(f, length ls) \<notin> defs R" by auto
  thus False unfolding defs_def defined_def by force
qed

lemma SN_on_imp_SN_on_sharp:
  assumes SN: "SN_on (rstep R) {s}"
    and s: "s = Fun f ss" and ndef: "\<not> defined R (shp f, length ss)"
  shows "SN_on (rstep R) {sharp_term shp s}" 
  unfolding s sharp_term.simps
proof (rule SN_args_imp_SN)
  fix t
  assume "t \<in> set ss" thus "SN_on (rstep R) {t}"
  using SN_imp_SN_arg[OF SN[unfolded s]] by auto
next
  {
    fix x r
    assume "(Var x,r) \<in> R"
    from left_var_imp_not_SN[OF this] SN have False ..
  }
  thus "\<forall>(l, r)\<in>R. is_Fun l" by force
next
  show "(shp f, length ss) \<notin> defs R" using ndef unfolding defs_def by simp
qed

lemma DP_step_in_R:
  fixes R::"('f,'v)trs" and v::"('f,'v)term \<Rightarrow> 'v" and shp :: "'f \<Rightarrow> 'f"
  assumes "(s,t) \<in> DP shp R" and inj: "inj shp" shows "\<exists>C. funas_ctxt C \<subseteq> (funas_trs R) \<and> (sharp_term (the_inv shp) s,C\<langle>sharp_term (the_inv shp) t\<rangle>) \<in> R"
proof -
  let ?us = "sharp_term (the_inv shp)"
  let ?s = "sharp_term shp"
  from assms obtain l r f ts
    where s: "s = sharp_term shp l" and t: "t = sharp_term shp (Fun f ts)"
    and R: "(l,r) \<in> R" and sub: "r \<unrhd> Fun f ts" unfolding DP_def supt_supteq_conv by auto
  from sub obtain C where r: "r = C\<langle>Fun f ts\<rangle>" by auto
  from rhs_wf[OF R subset_refl] have "funas_term r \<subseteq> funas_trs R" .
  hence "funas_term (C\<langle>Fun f ts\<rangle>) \<subseteq> funas_trs R" unfolding r .
  hence  "funas_ctxt C \<subseteq> funas_trs R" and "funas_term (Fun f ts) \<subseteq> funas_trs R" by auto
  from lhs_wf[OF R subset_refl] have "funas_term l \<subseteq> funas_trs R" .
  have us: "?us s = l" unfolding s by (cases l, auto simp: the_inv_f_f[OF inj])
  have ut: "?us t = Fun f ts" unfolding t by (simp add: the_inv_f_f[OF inj])
  from R have "(?us s,C\<langle>?us t\<rangle>) \<in> R" unfolding us ut r .
  with `funas_ctxt C \<subseteq> funas_trs R` show ?thesis by best
qed

lemma sharp_rrstep_imp_rstep:
  assumes rrstep: "(sharp_term shp s,sharp_term shp t) \<in> CS(DP shp R)" and "inj shp" and "wf_trs R"
  shows "\<exists>C. (s,C\<langle>t\<rangle>) \<in> rstep R"
proof -
  from `wf_trs R` have "wf_trs (DP shp R)" by (rule wf_trs_imp_wf_DP)
  from rrstep obtain l r \<sigma> where "(l,r) \<in> DP shp R" and ss: "sharp_term shp s = l\<cdot>\<sigma>" and st: "sharp_term shp t = r\<cdot>\<sigma>"
   by (induct, auto)
  from `(l,r) \<in> DP shp R` obtain l' r' h' us' where l: "l = sharp_term shp l'" and r: "r = sharp_term shp (Fun h' us')" (is "r = sharp_term shp ?u'")
  and l'r': "(l',r') \<in> R" and "r' \<unrhd> (Fun h' us')" and "\<not> l' \<rhd> Fun h' us'"
  unfolding DP_def by auto
  from `wf_trs R` and `(l',r') \<in> R` obtain f' ss' where l': "l' = Fun f' ss'" using wf_trs_imp_lhs_Fun by best
  from `r' \<unrhd> ?u'` obtain C where r': "r' = C\<langle>?u'\<rangle>" by best
  have ss: "sharp_term shp s = (Fun (shp f') ss') \<cdot> \<sigma>" unfolding ss l l' by simp
  then obtain f where s: "s = (Fun f ss') \<cdot> \<sigma>" by (cases s, auto)
  from s ss have "shp f = shp f'" by simp
  with `inj shp`[unfolded inj_on_def] have f: "f = f'" by simp
  have ts: "sharp_term shp t = (Fun (shp h') us') \<cdot> \<sigma>" unfolding st r by simp
  then obtain h where t: "t = (Fun h us') \<cdot> \<sigma>" by (cases t, auto)
  from t ts have "shp h = shp h'" by simp
  with `inj shp`[unfolded inj_on_def] have h: "h = h'" by simp
  show ?thesis
   by (rule exI[of _ "C \<cdot>c \<sigma>"], unfold s t f h, rule rstepI[OF l'r', of _ \<box> \<sigma>], unfold l' r', simp, simp) 
qed
 
lemma reverseTrs [iff]: "(rstep R)^-1 = rstep (R^-1)"
proof - 
  have oneDir: "\<And>R. (rstep R)^-1 \<subseteq> rstep (R^-1)" 
  proof
    fix R :: "('f,'v)trs" and st
    show "(st \<in> (rstep R)^-1) \<Longrightarrow> (st \<in> rstep (R^-1))"
    proof (cases st)
      case (Pair s t)
      assume "st \<in> (rstep R)^-1"
      with Pair have "(t,s) \<in> rstep R" by simp
      hence "(s,t) \<in> rstep (R^-1)" 
      proof
        fix l r C \<sigma> assume in_R: "(l, r) \<in> R" and t: "t = C\<langle>l\<cdot>\<sigma>\<rangle>" and s: "s = C\<langle>r\<cdot>\<sigma>\<rangle>"
        from in_R show "(s,t) \<in> rstep (R^-1)" unfolding s t by auto
      qed
      with Pair show ?thesis by simp
    qed
  qed
  hence one: "(rstep R)^-1 \<subseteq> rstep (R^-1)" by blast
  from oneDir have "(rstep (R^-1))^-1 \<subseteq> rstep ((R^-1) ^-1)" by blast
  with one show ?thesis by force
qed


lemma ctxt_of_pos_term_map_term_conv[iff]:
 assumes "p \<in> poss s"
 shows "map_ctxt fg (ctxt_of_pos_term p s) = (ctxt_of_pos_term p (map_term fg s))"
using assms proof (induct s arbitrary: p)
 case (Var x) thus ?case by simp
next
 case (Fun f ss) thus ?case
 proof (cases p)
  case Empty thus ?thesis by simp
 next
  case (PCons i q)
  with `p \<in> poss(Fun f ss)` have "i < length ss" and "q \<in> poss(ss!i)" unfolding PCons poss.simps by auto
  hence "ss!i \<in> set ss" by simp
  with Fun and `q \<in> poss(ss!i)`
  have IH: "map_ctxt fg(ctxt_of_pos_term q (ss!i)) = ctxt_of_pos_term q (map_term fg (ss!i))" by simp
  have "map_ctxt fg(ctxt_of_pos_term p (Fun f ss)) = map_ctxt fg(ctxt_of_pos_term (i<#q) (Fun f ss))" unfolding PCons by simp
  also have "\<dots> = map_ctxt fg(More f (take i ss) (ctxt_of_pos_term q (ss!i)) (drop (Suc i) ss))" by simp
  also have "\<dots> = More (fg f) (map (map_term fg) (take i ss)) (map_ctxt fg(ctxt_of_pos_term q (ss!i))) (map (map_term fg) (drop (Suc i) ss))" by simp
  also have "\<dots> = More (fg f) (map (map_term fg) (take i ss)) (ctxt_of_pos_term q (map_term fg (ss!i))) (map (map_term fg) (drop (Suc i) ss))" unfolding IH by simp
  also have "\<dots> = More (fg f) (take i (map (map_term fg) ss)) (ctxt_of_pos_term q (map (map_term fg) ss!i)) (drop (Suc i) (map (map_term fg) ss))" unfolding nth_map[OF `i < length ss`,symmetric] take_map drop_map nth_map by simp
  finally show ?thesis unfolding PCons by simp
 qed
qed

lemma var_rewrite_imp_not_SN:
  assumes sn: "SN_on (rstep R) {u}" and step: "(t, s) \<in> rstep R" 
  shows "is_Fun t" 
using assms
proof (cases t)
  case (Fun f ts) thus ?thesis by simp
next
  case (Var x)
  from step obtain l r p \<sigma> where "(Var x, s) \<in> rstep_r_p_s R (l,r) p \<sigma>" unfolding Var rstep_iff_rstep_r_p_s by best
  hence "l \<cdot> \<sigma> = Var x" and rule: "(l,r) \<in> R" unfolding rstep_r_p_s_def by (auto simp: Let_def)
  from this obtain y where "l = Var y" (is "_ = ?y") by (cases l, auto)
  with rule have "(?y, r) \<in> R" by auto
  hence "\<not> (SN_on (rstep R) {u})" by (rule left_var_imp_not_SN)
  with sn show ?thesis by blast
qed

lemma rstep_empty[simp]: "rstep {} = {}"
proof (rule set_eqI, simp, rule notI, simp only: split_paired_all)
  fix l r :: "('f,'v)term"
  assume "(l,r) \<in> rstep {}"
  thus False by (simp only: rstep_iff_rstep_r_p_s rstep_r_p_s_def, auto)
qed

lemma rstep_union: "rstep (R \<union> S) = rstep R \<union> rstep S"
proof (rule set_eqI, simp only: split_paired_all)
  fix s t
  show "((s,t) \<in> rstep (R \<union> S)) = ((s,t) \<in> rstep R \<union> rstep S)"
    by (simp only: Un_iff, simp only: rstep_iff_rstep_r_p_s, unfold rstep_r_p_s_def, force simp: Let_def)
qed

lemma rstep_id: "rstep Id = Id"
  by (rule set_eqI, simp only: split_paired_all, simp only: rstep_iff_rstep_r_p_s, unfold rstep_r_p_s_def, auto simp: Let_def, 
    rule exI, rule exI[where x = \<epsilon>], auto, rule exI[where x = Var], auto)


lemma map_rule_id[simp]: "map_rule id = id"
  by (intro ext, auto)

lemma map_trs_id[simp]: "map_trs id = id"
  by (intro ext, auto simp: map_trs.simps)

definition sig_step :: "'f sig \<Rightarrow> ('f, 'v) trs \<Rightarrow> ('f, 'v) trs" where
  "sig_step F R = {(a, b). (a, b) \<in> R \<and> funas_term a \<subseteq> F \<and> funas_term b \<subseteq> F}"

lemma sig_stepI[intro]: "(a,b) \<in> R \<Longrightarrow> funas_term a \<subseteq> F \<Longrightarrow> funas_term b \<subseteq> F \<Longrightarrow> (a,b) \<in> sig_step F R" unfolding sig_step_def by auto

lemma sig_stepE[elim,consumes 1]: "(a,b) \<in> sig_step F R \<Longrightarrow> \<lbrakk>(a,b) \<in> R \<Longrightarrow> funas_term a \<subseteq> F \<Longrightarrow> funas_term b \<subseteq> F \<Longrightarrow> P\<rbrakk> \<Longrightarrow> P" unfolding sig_step_def by auto

lemma all_ctxt_closed_sig_rsteps[intro]: fixes R :: "('f,'v)trs"
  shows "all_ctxt_closed F ((sig_step F (rstep R))^*)" (is "all_ctxt_closed _ (?R^*)")
proof (rule trans_ctxt_sig_imp_all_ctxt_closed)
  fix C :: "('f,'v)ctxt" and s t :: "('f,'v)term"
  assume C: "funas_ctxt C \<subseteq> F"
  and s: "funas_term s \<subseteq> F"
  and t: "funas_term t \<subseteq> F"
  and steps: "(s,t) \<in> ?R^*"
  from steps 
  show "(C \<langle> s \<rangle>, C \<langle> t \<rangle>) \<in> ?R^*"
  proof (induct)
    case (step t u)
    from step(2) have tu: "(t,u) \<in> rstep R" and t: "funas_term t \<subseteq> F" and u: "funas_term u \<subseteq> F" by auto
    have "(C \<langle> t \<rangle>, C \<langle> u \<rangle>) \<in> ?R" by (rule sig_stepI[OF rstep.ctxt[OF tu]], insert C t u, auto)   
    with step(3) show ?case by auto
  qed auto
qed auto
  
lemma wf_loop_imp_sig_ctxt_rel_not_SN:
  assumes R: "(l,C\<langle>l\<rangle>) \<in> R" and wf_l: "funas_term l \<subseteq> F"
  and wf_C: "funas_ctxt C \<subseteq> F"
  and ctxt: "ctxt_closed R"
  shows "\<not> SN_on (sig_step F R) {l}"
proof -
  let ?t = "\<lambda> i. (C^i)\<langle>l\<rangle>"
  have "\<forall>i. funas_term (?t i) \<subseteq> F"
  proof
    fix i show "funas_term (?t i) \<subseteq> F" unfolding funas_term_ctxt_apply
      by (rule Un_least[OF _ wf_l], induct i, insert wf_C, auto)
  qed
  moreover have "\<forall>i. (?t i,?t(Suc i)) \<in> R"
  proof
    fix i
    show "(?t i, ?t (Suc i)) \<in> R"
    proof (induct i)
      case 0 with R show ?case by auto
    next
      case (Suc i)
      from ctxt_closed_imp_ctxt[OF ctxt Suc, of C]
      show ?case by simp
    qed
  qed
  ultimately have steps: "\<forall>i. (?t i,?t(Suc i)) \<in> sig_step F R" unfolding sig_step_def by blast
  show ?thesis unfolding SN_defs 
    by (simp, intro exI[of _ ?t], simp only: steps, simp)
qed

lemma lhs_var_imp_sig_step_not_SN_on:
  assumes x: "(Var x, r) \<in> R" and F: "funas_trs R \<subseteq> F"
  shows "\<not> SN_on (sig_step F (rstep R)) {Var x}"  
proof -
  let ?\<sigma> = "(\<lambda>x. r)"
  let ?t = "\<lambda>i. (?\<sigma> ^^ i) x"
  obtain t where t: "t = ?t" by auto
  from rhs_wf[OF x F] have wf_r: "funas_term r \<subseteq> F" .
  {
    fix i
    have "funas_term (?t i) \<subseteq> F"
    proof (induct i)
      case 0 show ?case using wf_r by auto
    next
      case (Suc i)
      have "?t (Suc i) = ?t i \<cdot> ?\<sigma>" unfolding subst_power_Suc subst_compose_def by simp
      also have "funas_term ... \<subseteq> F" unfolding funas_term_subst[of "?t i"]
        using Suc wf_r by auto
      finally show ?case .
    qed
  } note wf_t = this
  { 
    fix i
    have "(t i, t (Suc i)) \<in> (sig_step F (rstep R))" unfolding t
      by (rule sig_stepI[OF rstepI[OF x, of _ \<box> "?\<sigma> ^^ i"] wf_t wf_t], auto simp: subst_compose_def)
  } note steps = this
  have x: "t 0 = Var x" unfolding t by simp
  with steps show ?thesis unfolding SN_defs not_not 
    by (intro exI[of _ t], auto)
qed

lemma rhs_free_vars_imp_sig_step_not_SN:
  assumes R: "(l,r) \<in> R" and free: "\<not> vars_term r \<subseteq> vars_term l"
  and F: "funas_trs R \<subseteq> F"
  shows "\<not> SN_on (sig_step F (rstep R)) {l}"
proof -
  from free obtain x where x: "x \<in> vars_term r - vars_term l" by auto
  hence "x \<in> vars_term r" by simp
  from supteq_Var[OF this] have "r \<unrhd> Var x" .
  then obtain C where r: "C\<langle>Var x\<rangle> = r" by auto
  let ?\<sigma> = "\<lambda>y. if y = x then l else Var y"
  let ?t = "\<lambda>i. ((C \<cdot>c ?\<sigma>)^i)\<langle>l\<rangle>"
  from rhs_wf[OF R] F have wf_r: "funas_term r \<subseteq> F" by fast
  from lhs_wf[OF R] F have wf_l: "funas_term l \<subseteq> F" by fast
  from wf_r[unfolded r[symmetric]]
  have wf_C: "funas_ctxt C \<subseteq> F" by simp
  from x have neq: "\<forall>y\<in>vars_term l. y \<noteq> x" by auto
  have "l\<cdot>?\<sigma> = l \<cdot> Var" 
    by (rule term_subst_eq, insert neq, auto)
  hence l: "l \<cdot> ?\<sigma> = l" by simp
  have wf_C: "funas_ctxt (C \<cdot>c ?\<sigma>) \<subseteq> F" using wf_C wf_l
    by simp
  have rsigma: "r\<cdot>?\<sigma> = (C \<cdot>c ?\<sigma>)\<langle>l\<rangle>" unfolding r[symmetric] by simp
  from R have lr: "(l \<cdot> ?\<sigma>, r \<cdot> ?\<sigma>) \<in> rstep R" by auto
  hence lr: "(l,(C \<cdot>c ?\<sigma>)\<langle>l\<rangle>) \<in> rstep R" unfolding l unfolding rsigma .
  show ?thesis
    by (rule wf_loop_imp_sig_ctxt_rel_not_SN[OF lr wf_l wf_C ctxt_closed_rstep])
qed

lemma lhs_var_imp_rstep_not_SN: assumes "(Var x,r) \<in> R" shows "\<not> SN(rstep R)"
using lhs_var_imp_sig_step_not_SN_on[OF assms subset_refl] unfolding sig_step_def SN_defs by blast

lemma rhs_free_vars_imp_rstep_not_SN:
  assumes "(l,r) \<in> R" and "\<not> vars_term r \<subseteq> vars_term l"
  shows "\<not> SN_on (rstep R) {l}"
using rhs_free_vars_imp_sig_step_not_SN[OF assms subset_refl] unfolding sig_step_def SN_defs by blast

lemma free_right_rewrite_imp_not_SN:
  assumes step: "(t,s) \<in> rstep_r_p_s R (l,r) p \<sigma>" 
  and vars: "\<not> vars_term l \<supseteq> vars_term r" 
  shows "\<not> SN_on (rstep R) {t}" 
proof
  assume SN: "SN_on (rstep R) {t}"
  let ?C = "ctxt_of_pos_term p t"
  from step have left: "?C\<langle>l \<cdot> \<sigma>\<rangle> = t" (is "?t = t") and right: "?C\<langle>r \<cdot> \<sigma>\<rangle> = s" and pos: "p \<in> poss t"
    and rule: "(l,r) \<in> R"
    unfolding rstep_r_p_s_def by (auto simp: Let_def)
  from rhs_free_vars_imp_rstep_not_SN[OF rule vars] have nSN:"\<not> SN_on (rstep R) {l}" by simp
  from SN_imp_SN_subt[OF SN ctxt_imp_supteq[of ?C "l \<cdot> \<sigma>", simplified left]]
  have SN: "SN_on (rstep R) {l \<cdot> \<sigma>}" .
  from SNinstance_imp_SN[OF SN] nSN show False by simp
qed

lemma not_SN_on_rstep_subst_apply_term[intro]:
  assumes "\<not> SN_on (rstep R) {t}" shows "\<not> SN_on (rstep R) {t \<cdot> \<sigma>}"
  using assms unfolding SN_on_def by best

lemma SN_rstep_imp_wf_trs: assumes "SN (rstep R)" shows "wf_trs R"
proof (rule ccontr)
  assume "\<not> wf_trs R"
  then obtain l r where R: "(l,r) \<in> R"
    and not_wf: "(\<forall>f ts. l \<noteq> Fun f ts) \<or> \<not>(vars_term r \<subseteq> vars_term l)" unfolding wf_trs_def
    by auto
  from not_wf have "\<not> SN (rstep R)"
  proof
    assume free: "\<not> vars_term r \<subseteq> vars_term l"
    from rhs_free_vars_imp_rstep_not_SN[OF R free] show ?thesis unfolding SN_defs by auto
  next
    assume "\<forall>f ts. l \<noteq> Fun f ts"
    then obtain x where l:"l = Var x" by (cases l) auto
    with R have "(Var x,r) \<in> R" unfolding l by simp
    from lhs_var_imp_rstep_not_SN[OF this] show ?thesis by simp
  qed
  with assms show False by blast
qed
  
lemma SN_sig_step_imp_wf_trs: assumes SN: "SN (sig_step F (rstep R))" and F: "funas_trs R \<subseteq> F" shows "wf_trs R"
proof (rule ccontr)
  assume "\<not> wf_trs R"
  then obtain l r where R: "(l,r) \<in> R"
    and not_wf: "(\<forall>f ts. l \<noteq> Fun f ts) \<or> \<not>(vars_term r \<subseteq> vars_term l)" unfolding wf_trs_def
    by auto
  from not_wf have "\<not> SN (sig_step F (rstep R))"
  proof
    assume free: "\<not> vars_term r \<subseteq> vars_term l"
    from rhs_free_vars_imp_sig_step_not_SN[OF R free F] show ?thesis unfolding SN_on_def by auto
  next
    assume "\<forall>f ts. l \<noteq> Fun f ts"
    then obtain x where l:"l = Var x" by (cases l) auto
    with R have "(Var x,r) \<in> R" unfolding l by simp
    from lhs_var_imp_sig_step_not_SN_on[OF this F] show ?thesis 
      unfolding SN_on_def by auto
  qed
  with assms show False by blast
qed

lemma rstep_cases'[consumes 1, case_names root nonroot]:
assumes rstep: "(s, t) \<in> rstep R"
  and root: "\<And>l r \<sigma>. (l, r) \<in> R \<Longrightarrow> l \<cdot> \<sigma> = s \<Longrightarrow> r \<cdot> \<sigma> = t \<Longrightarrow> P"
  and nonroot: "\<And>f ss1 u ss2 v. s = Fun f (ss1 @ u # ss2) \<Longrightarrow> t = Fun f (ss1 @ v # ss2) \<Longrightarrow> (u, v) \<in> rstep R \<Longrightarrow> P"
shows "P"
proof -
  from rstep_imp_C_s_r[OF rstep] obtain C \<sigma> l r
    where R: "(l,r) \<in> R" and s: "C\<langle>l\<cdot>\<sigma>\<rangle> = s" and t: "C\<langle>r\<cdot>\<sigma>\<rangle> = t" by fast
  show ?thesis proof (cases C)
    case Hole
    from s t have "l\<cdot>\<sigma> = s" and "r\<cdot>\<sigma> = t" by (auto simp: Hole)
    with R show ?thesis by (rule root)
  next
    case (More f ss1 D ss2)
    let ?u = "D\<langle>l\<cdot>\<sigma>\<rangle>"
    let ?v  = "D\<langle>r\<cdot>\<sigma>\<rangle>"
    have "s = Fun f (ss1 @ ?u # ss2)" by (simp add: More s[symmetric])
    moreover have "t = Fun f (ss1 @ ?v # ss2)" by (simp add: More t[symmetric])
    moreover have "(?u,?v) \<in> rstep R" by (intro rstep.intros) (rule R)
    ultimately show ?thesis by (rule nonroot)
  qed
qed

lemma NF_Var: assumes wf: "wf_trs R" shows "(Var x, t) \<notin> rstep R"
proof
  assume "(Var x, t) \<in> rstep R"
  from rstep_imp_C_s_r[OF this] obtain C l r \<sigma>
    where R: "(l,r) \<in> R" and lhs: "Var x = C\<langle>l\<cdot>\<sigma>\<rangle>" by fast
  from lhs have "Var x = l\<cdot>\<sigma>" by (induct C) auto
  then obtain y where l: "l = Var y" by (induct l) auto
  from wf R obtain f ss where "l = Fun f ss" unfolding wf_trs_def by best
  with l show False by simp
qed

lemma rstep_cases_Fun'[consumes 2, case_names root nonroot]:
assumes wf: "wf_trs R"
  and rstep: "(Fun f ss,t) \<in> rstep R"
  and root': "\<And>ls r \<sigma>. (Fun f ls,r) \<in> R \<Longrightarrow> map (\<lambda>t. t\<cdot>\<sigma>) ls = ss \<Longrightarrow> r\<cdot>\<sigma> = t \<Longrightarrow> P"
  and nonroot': "\<And>i u. i < length ss \<Longrightarrow> t = Fun f (take i ss@u#drop (Suc i) ss) \<Longrightarrow> (ss!i,u) \<in> rstep R \<Longrightarrow> P"
shows "P"
using rstep proof (cases rule: rstep_cases')
  case (root l r \<sigma>)
  with wf obtain g ls where l: "l = Fun g ls" unfolding wf_trs_def by best
  from root have [simp]: "g = f" unfolding l by simp
  from root have "(Fun f ls,r) \<in> R" and "map (\<lambda>t. t\<cdot>\<sigma>) ls = ss" and "r\<cdot>\<sigma> = t" unfolding l by auto
  thus ?thesis by (rule root')
next
  case (nonroot g ss1 u ss2 v)
  hence [simp]: "g = f" and args: "ss = ss1 @ u # ss2" by auto
  let ?i = "length ss1"
  from args have ss1: "take ?i ss = ss1" by simp
  from args have "drop ?i ss = u # ss2" by simp
  hence "drop (Suc 0) (drop ?i ss) = ss2" by simp
  hence ss2: "drop (Suc ?i) ss = ss2" by simp
  from args have len: "?i < length ss" by simp
  from id_take_nth_drop[OF len] have "ss = take ?i ss @ ss!?i # drop (Suc ?i) ss" by simp
  hence u: "ss!?i = u" unfolding args unfolding ss1[unfolded args] ss2[unfolded args] by simp
  from nonroot have "t = Fun f (take ?i ss@v#drop (Suc ?i) ss)" unfolding ss1 ss2 by simp
  moreover from nonroot have "(ss!?i,v) \<in> rstep R" unfolding u by simp
  ultimately show ?thesis by (rule nonroot'[OF len])
qed

lemma rstep_preserves_undefined_root:
  assumes "wf_trs R" and "\<not> defined R (f, length ss)" and "(Fun f ss, t) \<in> rstep R"
  shows "\<exists>ts. length ts = length ss \<and> t = Fun f ts"
proof -
  from `wf_trs R` and `(Fun f ss, t) \<in> rstep R` show ?thesis
  proof (cases rule: rstep_cases_Fun')
    case (root ls r \<sigma>)
    hence "defined R (f, length ss)" by (auto simp: defined_def)
    with `\<not> defined R (f, length ss)` show ?thesis by simp
  next
    case (nonroot i u) thus ?thesis by simp
  qed
qed

lemma funas_DP_subset:
  "funas_trs(DP shp R)
     \<subseteq> funas_trs R \<union> {(shp f,n) | f n. (f,n) \<in> funas_trs R}"
     (is "?F \<subseteq> ?H \<union> ?I")
proof (rule subrelI)
  fix f n 
  assume "(f,n) \<in> ?F"
  then obtain s t where st: "(s,t) \<in> DP shp R" and "(f,n) \<in> funas_rule (s,t)" 
    unfolding funas_trs_def by auto
  then obtain u where fn: "(f,n) \<in> funas_term u" and u: "u = s \<or> u = t" unfolding funas_rule_def
    by auto
  from st[unfolded DP_def] obtain l r uu where lr: "(l,r) \<in> R" and s: "s = sharp_term shp l" and t: "t = sharp_term shp uu" and uu: "r \<unrhd> uu"
    by force
  from fn u[unfolded s t] obtain v where fn: "(f,n) \<in> funas_term (sharp_term shp v)" and v: "v = l \<or> v = uu" by auto
  from fn have fn: "(f,n) \<in> funas_term v \<union> {(shp f,n) | f n. (f,n) \<in> funas_term v}"
    by (cases v, auto)
  from uu obtain C where r: "r = C\<langle>uu\<rangle>" ..  
  have "funas_term uu \<subseteq> funas_term r" unfolding r by simp
  with v have "funas_term v \<subseteq> funas_rule (l,r)" unfolding funas_rule_def by auto
  hence subset: "funas_term v \<subseteq> funas_trs R" using lr unfolding funas_trs_def by auto
  with fn show "(f,n) \<in> ?H \<union> ?I" by auto
qed

lemma rstep_ctxt_imp_nrrstep: assumes step: "(s,t) \<in> rstep R" and C: "C \<noteq> \<box>" shows "(C\<langle>s\<rangle>,C\<langle>t\<rangle>) \<in> nrrstep R"
proof -
  from step[unfolded rstep_iff_rstep_r_c_s rstep_r_c_s_def]
  obtain l r D \<sigma> where lr: "(l,r) \<in> R" and s: "s = D\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = D\<langle>r \<cdot> \<sigma>\<rangle>" by auto
  show ?thesis
    by (unfold nrrstep_def', rule, clarify, rule exI[of _ l], rule exI[of _ r], rule exI[of _ "C \<circ>c D"], rule exI[of _ \<sigma>], simp add: lr s t, cases C, auto simp: C)
qed

lemma rsteps_ctxt_imp_nrrsteps: assumes steps: "(s,t) \<in> (rstep R)^*" and C: "C \<noteq> \<box>" shows "(C\<langle>s\<rangle>,C\<langle>t\<rangle>) \<in> (nrrstep R)^*"
  using steps
proof (induct)
  case base show ?case by simp
next
  case (step t u)
  from rstep_ctxt_imp_nrrstep[OF step(2) C] step(3) show ?case by simp
qed

lemma nrrstep_mono:
  assumes "R \<subseteq> R'"
  shows "nrrstep R \<subseteq> nrrstep R'"
  using assms by (force simp: nrrstep_def rstep_r_p_s_def Let_def)

lemma singleton_subst_restrict [simp]:
  "subst x s |s {x} = subst x s"
  unfolding subst_def subst_restrict_def by (rule ext) simp

lemma singleton_subst_map [simp]:
  "f \<circ> subst x s = (f \<circ> Var)(x := f s)" by (intro ext, auto simp: subst_def)

lemma subst_restrict_vars [simp]:
  "(\<lambda>z. if z \<in> V then f z else g z) |s V = f |s V"
unfolding subst_restrict_def
proof (intro ext)
  fix x
  show "(if x \<in> V then if x \<in> V then f x else g x else Var x)
    = (if x \<in> V then f x else Var x)" by simp
qed

lemma subst_restrict_restrict [simp]:
  assumes "V \<inter> W = {}"
  shows "((\<lambda>z. if z \<in> V then f z else g z) |s W) = g |s W"
unfolding subst_restrict_def
proof (intro ext)
  fix x
  show "(if x \<in> W then if x \<in> V then f x else g x else Var x)
    = (if x \<in> W then g x else Var x)" using assms by auto
qed

lemma rstep_rstep: "rstep (rstep R) = rstep R"
proof -
  have "CC (CS (rstep R)) = rstep R" by (simp only: subst_closed_rstep_equal_rstep ctxt_closed_rstep_id)
  thus ?thesis unfolding rstep_CC_CS .
qed

lemma rstep_trancl_distrib: "rstep (R^+) \<subseteq> (rstep R)^+"
proof
  fix s t
  assume "(s,t) \<in> rstep (R^+)"
  thus "(s,t) \<in> (rstep R)^+"
  proof
    fix l r C \<sigma>
    assume lr: "(l,r) \<in> R^+" and s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>"
    from lr have "(C\<langle>l \<cdot> \<sigma>\<rangle>, C\<langle>r \<cdot> \<sigma>\<rangle>) \<in> (rstep R)^+" 
    proof(induct)
      case (base r)
      thus ?case by auto
    next
      case (step r rr)
      from step(2) have "(C\<langle>r \<cdot> \<sigma>\<rangle>, C\<langle>rr \<cdot> \<sigma>\<rangle>) \<in> (rstep R)" by auto
      with step(3) show ?case by auto
    qed
    thus "(s,t) \<in> (rstep R)^+" unfolding s t .
  qed 
qed

lemma rsteps_closed_subst:
  assumes "(s, t) \<in> (rstep R)\<^sup>*"
  shows "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (rstep R)\<^sup>*"
  using assms and subst_closed_rtrancl [OF subst_closed_rstep] by (auto simp: subst_closed_def)

lemma join_subst [intro]:
  assumes "(s, t) \<in> (rstep R)\<^sup>\<down>"
  shows "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (rstep R)\<^sup>\<down>"
proof -
  from assms obtain u where "(s, u) \<in> (rstep R)\<^sup>*" and "(t, u) \<in> (rstep R)\<^sup>*" by auto
  then have "(s \<cdot> \<sigma>, u \<cdot> \<sigma>) \<in> (rstep R)\<^sup>*" and "(t \<cdot> \<sigma>, u \<cdot> \<sigma>) \<in> (rstep R)\<^sup>*" by (auto intro: rsteps_closed_subst)
  then show ?thesis by blast
qed

lemma join_ctxt [intro]:
  assumes "(s, t) \<in> (rstep R)\<^sup>\<down>"
  shows "(C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> (rstep R)\<^sup>\<down>"
proof -
  from assms obtain u where "(s, u) \<in> (rstep R)\<^sup>*" and "(t, u) \<in> (rstep R)\<^sup>*" by auto
  then have "(C\<langle>s\<rangle>, C\<langle>u\<rangle>) \<in> (rstep R)\<^sup>*" and "(C\<langle>t\<rangle>, C\<langle>u\<rangle>) \<in> (rstep R)\<^sup>*" by (auto intro: rsteps_closed_ctxt)
  then show ?thesis by blast
qed

lemma rstep_simps:
  "rstep (R\<^sup>=) = (rstep R)\<^sup>="
  "rstep (rstep R) = rstep R"
  "rstep (R \<union> S) = rstep R \<union> rstep S"
  "rstep Id = Id"
  "rstep (R\<^sup>\<leftrightarrow>) = (rstep R)\<^sup>\<leftrightarrow>"
  by auto

lemma rstep_rtrancl_idemp [simp]:
  "rstep ((rstep R)\<^sup>*) = (rstep R)\<^sup>*"
proof -
  { fix s t
    assume "(s, t) \<in> rstep ((rstep R)\<^sup>*)"
    then have "(s, t) \<in> (rstep R)\<^sup>*"
    by (induct) (metis rsteps_closed_ctxt rsteps_closed_subst) }
  then show ?thesis by auto
qed

definition instance_rule :: "('f, 'v) rule \<Rightarrow> ('f, 'w) rule \<Rightarrow> bool" where
  [code del]: "instance_rule lr st \<longleftrightarrow> (\<exists> \<sigma>. fst lr = fst st \<cdot> \<sigma> \<and> snd lr = snd st \<cdot> \<sigma>)"
(* instance rule has implementation in Substitution.thy *)

definition eq_rule_mod_vars :: "('f, 'v) rule \<Rightarrow> ('f, 'v) rule \<Rightarrow> bool" where
  "eq_rule_mod_vars lr st \<longleftrightarrow> instance_rule lr st \<and> instance_rule st lr"

notation eq_rule_mod_vars ("(_/ =v _)" [51,51] 50)

lemma instance_rule_var_cond: assumes eq: "instance_rule (s,t) (l,r)"
  and vars: "vars_term r \<subseteq> vars_term l"
  shows "vars_term t \<subseteq> vars_term s"
proof -
  from eq[unfolded instance_rule_def]
  obtain \<tau> where s: "s = l \<cdot> \<tau>" and t: "t = r \<cdot> \<tau>" by auto
  show ?thesis
  proof
    fix x 
    assume "x \<in> vars_term t"
    from this[unfolded t] have "x \<in> vars_term (l \<cdot> \<tau>)" using vars unfolding vars_term_subst by auto
    thus "x \<in> vars_term s" unfolding s by auto
  qed
qed

lemma instance_rule_rstep: assumes step: "(s,t) \<in> rstep {lr}"
  and bex: "Bex R (instance_rule lr)"
  shows "(s,t) \<in> rstep R"
proof -
  from bex obtain lr' where inst: "instance_rule lr lr'" and R: "lr' \<in> R" by auto
  obtain l r where lr: "lr = (l,r)" by force
  obtain l' r' where lr': "lr' = (l',r')" by force
  note inst = inst[unfolded lr lr']
  note R = R[unfolded lr']
  from inst[unfolded instance_rule_def] obtain \<sigma> where l: "l = l' \<cdot> \<sigma>" and r: "r = r' \<cdot> \<sigma>" by auto
  from step[unfolded lr] obtain C \<tau> where "s = C \<langle>l \<cdot> \<tau>\<rangle>" "t = C \<langle>r \<cdot> \<tau>\<rangle>" by auto
  with l r have s: "s = C\<langle>l' \<cdot> (\<sigma> \<circ>s \<tau>)\<rangle>" and t: "t = C\<langle>r' \<cdot> (\<sigma> \<circ>s \<tau>)\<rangle>" by auto
  from rstepI[OF R s t] show ?thesis .
qed

lemma eq_rule_mod_vars_var_cond: assumes eq: "(l,r) =v (s,t)"
  and vars: "vars_term r \<subseteq> vars_term l"
  shows "vars_term t \<subseteq> vars_term s"
  by (rule instance_rule_var_cond[OF _ vars], insert eq[unfolded eq_rule_mod_vars_def], auto)

lemma eq_rule_mod_varsE[elim]: assumes "(l,r) =v (s,t)"
  shows "\<exists> \<sigma> \<tau>. l = s \<cdot> \<sigma> \<and> r = t \<cdot> \<sigma> \<and> s = l \<cdot> \<tau> \<and> t = r \<cdot> \<tau> \<and> range \<sigma> \<subseteq> range Var \<and> range \<tau> \<subseteq> range Var"
proof -
  from assms[unfolded eq_rule_mod_vars_def instance_rule_def fst_conv snd_conv]
  obtain \<sigma> \<tau> where l: "l = s \<cdot> \<sigma>" and r: "r = t \<cdot> \<sigma>" and s: "s = l \<cdot> \<tau>" and t: "t = r \<cdot> \<tau>" by blast+
  let ?vst = "vars_term (Fun f [s,t])"
  let ?vlr = "vars_term (Fun f [l,r])"
  def \<sigma>' \<equiv> "\<lambda> x. if x \<in> ?vst then \<sigma> x else Var x"
  def \<tau>' \<equiv> "\<lambda> x. if x \<in> ?vlr then \<tau> x else Var x"
  show ?thesis
  proof (intro exI conjI)
    show l: "l = s \<cdot> \<sigma>'" unfolding l \<sigma>'_def
      by (rule term_subst_eq, auto)
    show r: "r = t \<cdot> \<sigma>'" unfolding r \<sigma>'_def
      by (rule term_subst_eq, auto)
    show s: "s = l \<cdot> \<tau>'" unfolding s \<tau>'_def
      by (rule term_subst_eq, auto)
    show t: "t = r \<cdot> \<tau>'" unfolding t \<tau>'_def
      by (rule term_subst_eq, auto)
    have "Fun f [s,t] \<cdot> Var = Fun f [l, r] \<cdot> \<tau>'" unfolding s t by simp
    also have "\<dots> = Fun f [s,t] \<cdot> (\<sigma>' \<circ>s \<tau>')" unfolding l r by simp
    finally have "Fun f [s,t] \<cdot> (\<sigma>' \<circ>s \<tau>') = Fun f [s,t] \<cdot> Var" by simp
    from term_subst_eq_rev[OF this] have vst: "\<And> x. x \<in> ?vst \<Longrightarrow> \<sigma>' x \<cdot> \<tau>' = Var x" unfolding subst_compose_def by auto
    have "Fun f [l,r] \<cdot> Var = Fun f [s, t] \<cdot> \<sigma>'" unfolding l r by simp
    also have "\<dots> = Fun f [l,r] \<cdot> (\<tau>' \<circ>s \<sigma>')" unfolding s t by simp
    finally have "Fun f [l,r] \<cdot> (\<tau>' \<circ>s \<sigma>') = Fun f [l,r] \<cdot> Var" by simp
    from term_subst_eq_rev[OF this] have vlr: "\<And> x. x \<in> ?vlr \<Longrightarrow> \<tau>' x \<cdot> \<sigma>' = Var x" unfolding subst_compose_def by auto
    {
      fix x
      have "\<sigma>' x \<in> range Var"
      proof (cases "x \<in> ?vst")
        case True
        from vst[OF this] show ?thesis by (cases "\<sigma>' x", auto)
      next
        case False
        thus ?thesis unfolding \<sigma>'_def by auto
      qed
    }
    thus "range \<sigma>' \<subseteq> range Var" by auto
    {
      fix x
      have "\<tau>' x \<in> range Var"
      proof (cases "x \<in> ?vlr")
        case True
        from vlr[OF this] show ?thesis by (cases "\<tau>' x", auto)
      next
        case False
        thus ?thesis unfolding \<tau>'_def by auto
      qed
    }
    thus "range \<tau>' \<subseteq> range Var" by auto
  qed
qed


definition
  left_linear_trs :: "('f, 'v) trs \<Rightarrow> bool"
where
  "left_linear_trs R \<longleftrightarrow> (\<forall>(l, r)\<in>R. linear_term l)"

lemma left_linear_mono: assumes "left_linear_trs S" and "R \<subseteq> S" shows "left_linear_trs R"
  using assms unfolding left_linear_trs_def by auto

lemma left_linear_map_trs[simp]:  "left_linear_trs (map_trs f R) = left_linear_trs R"
  unfolding left_linear_trs_def by (auto simp: map_trs.simps)

lemma left_linear_weak_match_rstep:
  assumes rstep: "(u, v) \<in> rstep R"
    and weak_match: "weak_match s u"
    and ll: "left_linear_trs R"
  shows "\<exists>t. (s, t) \<in> rstep R \<and> weak_match t v"
using weak_match 
proof (induct rule: rstep_induct[OF rstep])
  case (1 C sig l r)
  from 1(2) show ?case
  proof (induct C arbitrary: s)
    case (More f bef C aft s)
    let ?n = "Suc (length bef + length aft)" 
    let ?m = "length bef"
    from More(2) obtain ss where s: "s = Fun f ss" and lss: "?n = length ss" and wm: "(\<forall>i<length ss. weak_match (ss ! i) ((bef @ C\<langle>l \<cdot> sig\<rangle> # aft) ! i))"  by (cases s, auto)
    from lss wm[THEN spec, of ?m] have "weak_match (ss ! ?m) C\<langle>l \<cdot> sig\<rangle>" by auto
    from More(1)[OF this] obtain t where wmt:  "weak_match t C\<langle>r \<cdot> sig\<rangle>" and step: "(ss ! ?m,t) \<in> rstep R" by auto
    from lss have mss: "?m < length ss" by simp
    let ?tsi = "\<lambda> t. take ?m ss @ t # drop (Suc ?m) ss"
    let ?ts = "?tsi t"
    let ?ss = "?tsi (ss ! ?m)"
    from id_take_nth_drop[OF mss] 
    have lts: "length ?ts = ?n" using lss by auto
    show ?case
    proof (rule exI[of _ "Fun f ?ts"], intro conjI)
      have "weak_match (Fun f ?ts) (More f bef C aft)\<langle>r \<cdot> sig\<rangle> = 
        weak_match (Fun f ?ts) (Fun f (bef @ C\<langle>r \<cdot> sig\<rangle> # aft))" by simp
      also have "\<dots> = True" proof (simp only: weak_match.simps lts, simp, intro allI impI)
        fix i
        assume i: "i < ?n"
        show "weak_match (?ts ! i) ((bef @ C\<langle>r \<cdot> sig\<rangle> # aft) ! i)"
        proof (cases "i = ?m")
          case True
          have "weak_match (?ts ! i) ((bef @ C\<langle>r \<cdot> sig\<rangle> # aft) ! i) = weak_match t C\<langle>r \<cdot> sig\<rangle>"
            using True mss by (simp add: nth_append)
          thus ?thesis using wmt by simp
        next
          case False
          have eq: "?ts ! i = ss ! i \<and> (bef @ C\<langle>r \<cdot> sig\<rangle> # aft) ! i = (bef @ C\<langle>l \<cdot> sig\<rangle> # aft) ! i"
          proof (cases "i < ?m")
            case True
            thus ?thesis by (simp add: nth_append lss[symmetric])
          next
            case False
            with `i \<noteq> ?m` i have "\<exists> j. i = Suc (?m + j) \<and> j < length aft" by presburger
            then obtain j where i: "i = Suc (?m + j)" and j: "j < length aft" by auto
            hence id: " (Suc (length bef + j) - min (Suc (length bef + length aft)) (length bef)) = Suc j" by simp
            from j show ?thesis by (simp add: nth_append i id lss[symmetric])
          qed
          thus ?thesis using wm[THEN spec, of i] i[unfolded lss] by (simp) 
        qed
      qed
      finally show "weak_match (Fun f ?ts) (More f bef C aft)\<langle>r \<cdot> sig\<rangle>" by simp
    next
      have "s = Fun f ?ss" unfolding s using id_take_nth_drop[OF mss, symmetric] by simp
      also have "\<dots> = (More f (take ?m ss) \<box> (drop (Suc ?m) ss))\<langle>(ss ! ?m)\<rangle>" (is "_ = ?C\<langle>_\<rangle>") by simp
      finally have s: "s = ?C\<langle>ss ! ?m\<rangle>" .
      have t: "Fun f ?ts = ?C\<langle>t\<rangle>" by simp
      from rstep.ctxt[OF step]
      show "(s, Fun f ?ts) \<in> rstep R"
        unfolding s t .
    qed
  next
    case (Hole s)
    from ll 1(1) have "linear_term l" unfolding left_linear_trs_def by auto
    from linear_weak_match[OF this Hole[simplified] refl] obtain \<tau> where 
      "s = l \<cdot> \<tau>" and "(\<forall> x \<in> vars_term l . weak_match (Var x \<cdot> \<tau>) (Var x \<cdot> sig))"
      by auto
    then obtain tau where s: "s = l \<cdot> tau" and wm: "(\<forall> x \<in> vars_term l . weak_match (tau x) (Var x \<cdot> sig))"
      by (auto)
    let ?delta = "(\<lambda> x. if x \<in> vars_term l then tau x else Var x \<cdot> sig)"
    show ?case
    proof (rule exI[of _ "r \<cdot> ?delta"], rule conjI)
      have "s = l \<cdot> (tau |s (vars_term l))" unfolding s by (rule coincidence_lemma)
      also have "\<dots> = l \<cdot> (?delta  |s (vars_term l))" by simp
      also have "\<dots> = l \<cdot> ?delta" by (rule coincidence_lemma[symmetric])
      finally have s: "s = l \<cdot> ?delta" .
      from 1(1) have step: "(l \<cdot> ?delta, r \<cdot> ?delta) \<in> rstep R" by auto
      thus "(s, r \<cdot> ?delta) \<in> rstep R" unfolding s .
    next
      have "weak_match (r \<cdot> ?delta) (r \<cdot> sig)"
      proof (induct r)
        case (Fun f ss)
        from this[unfolded set_conv_nth]
        show ?case by (force)
      next
        case (Var x)
        show ?case
        proof (cases "x \<in> vars_term l")
          case True
          with wm Var show ?thesis by simp
        next
          case False
          show ?thesis by (simp add: Var False weak_match_refl)
        qed
      qed
      thus "weak_match (r \<cdot> ?delta) (\<box> \<langle>(r \<cdot> sig)\<rangle>)" by simp
    qed
  qed
qed


fun S where
  "S R s t 0 = s"
| "S R s t (Suc i) = (SOME u. (S R s t i,u) \<in> rstep R \<and> weak_match u (t(Suc i)))"

lemma weak_match_SN:
  assumes wm: "weak_match s t"
    and ll: "left_linear_trs R"
    and SN: "SN_on (rstep R) {s}"
  shows "SN_on (rstep R) {t}"
proof
  fix f
  assume t0: "f 0 \<in> {t}" and chain: "chain (rstep R) f"
  let ?s = "S R s f"
  let ?P = "\<lambda>i u. (?s i, u) \<in> rstep R \<and> weak_match u (f (Suc i))"
  have "\<forall>i. (?s i, ?s (Suc i)) \<in> rstep R \<and> weak_match (?s (Suc i)) (f (Suc i))"
  proof
    fix i show "(?s i, ?s (Suc i)) \<in> rstep R \<and> weak_match (?s (Suc i)) (f (Suc i))"
    proof (induct i)
      case 0
      from chain have ini: "(f 0, f (Suc 0)) \<in> rstep R" by simp
      hence "(t, f (Suc 0)) \<in> rstep R" unfolding singletonD[OF t0, symmetric] .
      from someI_ex[OF left_linear_weak_match_rstep[OF this wm ll]]
      show ?case by simp
    next
      case (Suc i)
      hence IH1: "(?s i, ?s (Suc i)) \<in> rstep R"
        and IH2: "weak_match (?s (Suc i)) (f (Suc i))" by auto
      from chain have nxt: "(f (Suc i), f (Suc (Suc i))) \<in> rstep R" by simp
      from someI_ex[OF left_linear_weak_match_rstep[OF this IH2 ll]]
      have "\<exists>u. ?P (Suc i) u" by auto
      from someI_ex[OF this] 
      show ?case by simp
    qed
  qed
  moreover have "?s 0 = s" by simp
  ultimately have "\<not> SN_on (rstep R) {s}" by best
  with SN show False by simp
qed

hide_const S

lemma lhs_notin_NF_rstep: "(l, r) \<in> R \<Longrightarrow> l \<notin> NF (rstep R)" by auto

lemma NF_instance:
  assumes "(t \<cdot> \<sigma>) \<in> NF (rstep R)" shows "t \<in> NF (rstep R)"
using assms by auto

lemma NF_subterm:
  assumes "t \<in> NF (rstep R)" and "t \<unrhd> s"
  shows "s \<in> NF (rstep R)"
proof (rule ccontr)  
  assume "\<not> ?thesis"
  then obtain u where "(s, u) \<in> rstep R" by auto
  from `t \<unrhd> s` obtain C where "t = C\<langle>s\<rangle>" by auto
  with `(s, u) \<in> rstep R` have "(t, C\<langle>u\<rangle>) \<in> rstep R" by auto
  hence "t \<notin> NF (rstep R)" by auto
  with assms show False by simp
qed

abbreviation
  lhss :: "('f, 'v) trs \<Rightarrow> ('f, 'v) term set"
where
  "lhss R \<equiv> fst ` R"

abbreviation
  rhss :: "('f, 'v) trs \<Rightarrow> ('f, 'v) term set"
where
  "rhss R \<equiv> snd ` R"


definition map_trs_wa :: "('f \<Rightarrow> nat \<Rightarrow> 'g) \<Rightarrow> ('f, 'v) trs \<Rightarrow> ('g, 'v) trs" where
  "map_trs_wa fg R = (\<lambda>(l, r). (map_term_wa fg l, map_term_wa fg r)) ` R"

lemma map_trs_wa_union: "map_trs_wa fg (R \<union> S) = map_trs_wa fg R \<union> map_trs_wa fg S"
  unfolding map_trs_wa_def by auto

lemma map_term_wa_compose: "map_term_wa gh (map_term_wa fg t) = map_term_wa (\<lambda> f n. gh (fg f n) n) t"
  by (induct t, auto)

lemma map_trs_wa_compose: "map_trs_wa gh (map_trs_wa fg R) = map_trs_wa (\<lambda> f n. gh (fg f n) n) R" (is "?L = map_trs_wa ?fgh R")
proof -
  have "map_trs_wa ?fgh R = {(map_term_wa ?fgh l, map_term_wa ?fgh r)| l r. (l,r) \<in> R}" unfolding map_trs_wa_def by auto
  also have "\<dots> = {(map_term_wa gh (map_term_wa fg l), map_term_wa gh (map_term_wa fg r)) | l r. (l,r) \<in> R}" unfolding map_term_wa_compose ..
  finally show ?thesis unfolding map_trs_wa_def by force
qed

lemma map_trs_wa_id: assumes id: "\<And> f n. fg f n = f"
  shows "map_trs_wa fg R = R"
  unfolding map_trs_wa_def map_term_wa_id[OF id] by auto


lemma map_trs_wa_rstep: assumes step:"(s,t) \<in> rstep R"
  shows "(map_term_wa fg s,map_term_wa fg t) \<in> rstep (map_trs_wa fg R)"
using step
proof (induct)
  case (IH C \<sigma> l r)
  show ?case unfolding map_trs_wa_def
    by (rule rstepI[where l = "map_term_wa fg l" and r = "map_term_wa fg r" and C = "map_ctxt_wa fg C"], auto simp: IH)
qed

lemma rstep_ground: 
  assumes wf_trs: "\<And>l r. (l, r) \<in> R \<Longrightarrow> vars_term r \<subseteq> vars_term l"
  and ground: "ground s"
  and step: "(s, t) \<in> rstep R"
  shows "ground t"
using step ground
proof (induct)
  case (IH C \<sigma> l r)
  from wf_trs[OF IH(1)] IH(2)
  show ?case by auto
qed

definition locally_terminating :: "('f,'v)trs \<Rightarrow> bool"
  where "locally_terminating R \<equiv> \<forall> F. finite F \<longrightarrow> SN (sig_step F (rstep R))"

lemma supt_rstep_stable:
  assumes "(s, t) \<in> op \<rhd> \<union> rstep R"
  shows "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> op \<rhd> \<union> rstep R"
using assms proof
  assume "s \<rhd> t" show ?thesis
  proof (rule UnI1)
    from `s \<rhd> t` show "s \<cdot> \<sigma> \<rhd> t \<cdot> \<sigma>" by (rule supt_stable)
  qed
next
  assume "(s, t) \<in> rstep R" show ?thesis
  proof (rule UnI2)
    from `(s, t) \<in> rstep R` show "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> rstep R" ..
  qed
qed

lemma supt_rstep_trancl_stable:
  assumes "(s, t) \<in> (op \<rhd> \<union> rstep R)^+"
  shows "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (op \<rhd> \<union> rstep R)^+"
using assms proof (induct)
  case (base u)
  hence "(s \<cdot> \<sigma>, u \<cdot> \<sigma>) \<in> op \<rhd> \<union> rstep R" by (rule supt_rstep_stable)
  thus ?case ..
next
  case (step u v)
  from `(s \<cdot> \<sigma>, u \<cdot> \<sigma>) \<in> (op \<rhd> \<union> rstep R)^+`
    and supt_rstep_stable[OF `(u, v) \<in> op \<rhd> \<union> rstep R`, of "\<sigma>"]
    show ?case ..
qed

lemma supt_rsteps_stable:
  assumes "(s, t) \<in> (op \<rhd> \<union> rstep R)^*"
  shows "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (op \<rhd> \<union> rstep R)^*"
using assms proof (induct)
  case base thus ?case ..
next
  case (step u v)
  from `(s, u) \<in> (op \<rhd> \<union> rstep R)^*` and `(u, v) \<in> op \<rhd> \<union> rstep R`
    have "(s, v) \<in> (op \<rhd> \<union> rstep R)^+" by (rule rtrancl_into_trancl1)
  from trancl_into_rtrancl[OF supt_rstep_trancl_stable[OF this]]
    show ?case .
qed

lemma eq_rule_mod_vars_refl[simp]: "r =v r"
proof (cases r)
  case (Pair l r)
  {
    have "fst (l, r) = fst (l, r) \<cdot> Var \<and> snd (l, r) = snd (l, r) \<cdot> Var" by auto
  }
  thus ?thesis unfolding Pair eq_rule_mod_vars_def instance_rule_def by best
qed

lemma instance_rule_refl[simp]: "instance_rule r r"
  using eq_rule_mod_vars_refl[of r] unfolding eq_rule_mod_vars_def by simp

lemma is_Fun_Fun_conv: "is_Fun t = (\<exists>f ts. t = Fun f ts)" by auto

lemma wf_trs_def':
  "wf_trs R = (\<forall>(l, r)\<in>R. is_Fun l \<and> vars_term r \<subseteq> vars_term l)"
  by (rule iffI) (auto simp: wf_trs_def is_Fun_Fun_conv)

definition wf_rule :: "('f, 'v) rule \<Rightarrow> bool" where
  "wf_rule r \<longleftrightarrow> is_Fun (fst r) \<and> vars_term (snd r) \<subseteq> vars_term (fst r)"

definition wf_rules :: "('f, 'v) trs \<Rightarrow> ('f, 'v) trs" where
  "wf_rules R = {r. r \<in> R \<and> wf_rule r}"

lemma wf_trs_wf_rules[simp]: "wf_trs (wf_rules R)"
  unfolding wf_trs_def' wf_rules_def wf_rule_def split_def by simp

lemma wf_rules_subset[simp]: "wf_rules R \<subseteq> R"
  unfolding wf_rules_def by auto

fun wf_reltrs :: "('f, 'v) trs \<Rightarrow> ('f, 'v) trs  \<Rightarrow> bool" where
  "wf_reltrs R S = (
    wf_trs R \<and> (R \<noteq> {} \<longrightarrow> (\<forall>l r. (l, r) \<in> S \<longrightarrow> vars_term r \<subseteq> vars_term l)))"

lemma SN_rel_imp_wf_reltrs:
  assumes SN_rel: "SN_rel (rstep R) (rstep S)" 
  shows "wf_reltrs R S"
proof (rule ccontr)
  assume "\<not> ?thesis"
  then obtain l r where "\<not> wf_trs R \<or> R \<noteq> {} \<and> (l,r) \<in> S \<and> \<not> vars_term r \<subseteq> vars_term l" (is "_ \<or> ?two") by auto
  thus False
  proof
    assume "\<not> wf_trs R"
    with SN_rstep_imp_wf_trs[OF SN_rel_imp_SN[OF assms]]
    show False by simp
  next
    assume ?two
    then obtain ll rr x where lr: "(l,r) \<in> S" and llrr: "(ll,rr) \<in> R" and x: "x \<in> vars_term r" and nx: "x \<notin> vars_term l" by auto
    obtain f and \<sigma>
      where sigma: "\<sigma> = (\<lambda>y. if x = y then Fun f [ll,l] else Var y)" by auto
    have id: "\<sigma> |s (vars_term l) = Var" unfolding sigma
      by (simp add: subst_restrict_def, rule ext, auto simp: nx)
    have l: "l = l \<cdot> \<sigma>"  by (simp add: coincidence_lemma[of l \<sigma>] id)
    have "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> rstep S" using lr by auto
    with l have sstep: "(l, r \<cdot> \<sigma>) \<in> rstep S" by simp
    from supteq_stable[OF supteq_Var[OF x], of \<sigma>] have 
       "r \<cdot> \<sigma> \<unrhd> Fun f [ll,l]" unfolding sigma by auto
    then obtain C where "C\<langle>Fun f [ll, l]\<rangle> = r \<cdot> \<sigma>" by auto
    with sstep have sstep: "(l,C\<langle>Fun f [ll, l]\<rangle>) \<in> rstep S" by simp
    obtain r where r: "r = relto (rstep R) (rstep S) \<union> op \<rhd>" by auto
    have "(C\<langle>Fun f [ll,l]\<rangle>, C\<langle>Fun f [rr,l]\<rangle>) \<in> rstep R"
      unfolding rstep_iff_rstep_r_c_s
      by (rule exI, rule exI, rule exI[of _ "C \<circ>c More f [] \<box> [l]"], rule exI[of _ Var], intro conjI, rule llrr, unfold rstep_r_c_s_def, simp)
    with sstep have relto: "(l,C\<langle>Fun f [rr,l]\<rangle>) \<in> r" unfolding  r by auto
    have "C\<langle>Fun f [rr,l]\<rangle> \<unrhd> Fun f [rr,l]" using ctxt_imp_supteq by auto
    also have "Fun f [rr,l] \<rhd> l" by auto
    finally have supt: "(C\<langle>Fun f [rr,l]\<rangle>, l) \<in> op \<rhd>" unfolding supt_def by simp 
    hence "(C\<langle>Fun f [rr,l]\<rangle>, l) \<in> r" unfolding r by auto
    with relto have loop: "(l, l) \<in> r^+" by auto
    have "SN r" unfolding r 
      by (rule SN_imp_SN_union_supt[OF SN_rel[unfolded SN_rel_defs]], blast)
    hence "SN (r^+)" by (rule SN_imp_SN_trancl)
    with loop show False unfolding SN_on_def by auto
  qed
qed

lemmas rstep_wf_rules_subset = rstep_mono[OF wf_rules_subset]

definition map_vars_trs :: "('v \<Rightarrow> 'w) \<Rightarrow> ('f, 'v) trs \<Rightarrow> ('f, 'w) trs" where
  "map_vars_trs f R = (\<lambda> (l, r). (map_vars f l, map_vars f r)) ` R"

lemma map_vars_rstep:
  assumes "(s, t) \<in> rstep (map_vars_trs f R)" (is "_ \<in> rstep ?R")
  shows "(s \<cdot> \<tau>, t \<cdot> \<tau>) \<in> rstep R"
  using assms
proof    
  fix ml mr C \<sigma>
  assume mem: "(ml,mr) \<in> ?R" and s: "s = C\<langle>ml \<cdot> \<sigma>\<rangle>" and t: "t = C\<langle>mr \<cdot> \<sigma>\<rangle>"
  let ?m = "map_vars f"
  from mem obtain l r where mem: "(l,r) \<in> R" and id: "ml = ?m l" "mr = ?m r" 
    unfolding map_vars_trs_def by auto
  have id: "s \<cdot> \<tau> = (C \<cdot>c \<tau>)\<langle>?m l \<cdot> \<sigma> \<circ>s \<tau>\<rangle>" "t \<cdot> \<tau> = (C \<cdot>c \<tau>)\<langle>?m r \<cdot> \<sigma> \<circ>s \<tau>\<rangle>" unfolding s t unfolding id by auto
  thus "(s \<cdot> \<tau>, t \<cdot> \<tau>) \<in> rstep R" unfolding id map_subst_domain 
    using mem by auto
qed

lemma map_vars_rsteps: assumes "(s,t) \<in> (rstep (map_vars_trs f R))^*" (is "_ \<in> (rstep ?R)^*")
  shows "(s \<cdot> \<tau>, t \<cdot> \<tau>) \<in> (rstep R)^*"
  using assms
proof (induct)   
  case base thus ?case by simp
next
  case (step t u)
  from map_vars_rstep[OF step(2), of \<tau>] step(3) show ?case by auto
qed

lemma rsteps_subst_closed: "(s,t) \<in> (rstep R)^+ \<Longrightarrow> (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (rstep R)^+"
proof -
  let ?R = "rstep R"
  assume steps: "(s,t) \<in> ?R^+"
  have subst: "subst_closed (?R^+)" by (rule subst_closed_trancl[OF subst_closed_rstep])
  from this[unfolded subst_closed_def] steps show ?thesis by auto
qed

lemma supteq_rtrancl_supt: "(R^+ O op \<unrhd>) \<subseteq> (op \<rhd> \<union> R)^+" (is "?l \<subseteq> ?r")
proof
  fix x z
  assume "(x,z) \<in> ?l"
  then obtain y where xy: "(x,y) \<in> R^+" and yz: "(y,z) \<in> op \<unrhd>" by auto
  from xy have xy: "(x,y) \<in> ?r" by (rule trancl_mono, simp)
  show "(x,z) \<in> ?r"
  proof (cases "y = z")
    case True
    with xy show ?thesis by simp
  next
    case False
    with yz have yz: "(y,z) \<in> op \<rhd> \<union> R" by auto
    with xy have xz: "(x,z) \<in> ?r O (op \<rhd> \<union> R)" by auto
    show ?thesis
      by (rule set_mp[OF _ xz], regexp)
  qed
qed

lemma rrstepI[intro]: "(l, r) \<in> R \<Longrightarrow> s = l \<cdot> \<sigma> \<Longrightarrow> t = r \<cdot> \<sigma> \<Longrightarrow> (s, t) \<in> rrstep R" 
  unfolding rrstep_def' by auto

lemma CS_rrstep_conv: "CS = rrstep"
  by (intro ext, unfold rrstep_def', rule, insert subst_closed_decompose, blast, auto)


subsection {* Rewrite steps at a fixed position *}

inductive_set rstep_pos :: "('f, 'v) trs \<Rightarrow> pos \<Rightarrow> ('f, 'v) term rel" for R and p
where
  rule [intro]:"(l, r) \<in> R \<Longrightarrow> p \<in> poss s \<Longrightarrow> s |_ p = l \<cdot> \<sigma> \<Longrightarrow>
    (s, replace_at s p (r \<cdot> \<sigma>)) \<in> rstep_pos R p"

lemma rstep_pos_subst:
  assumes "(s, t) \<in> rstep_pos R p"
  shows "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> rstep_pos R p"
using assms
proof (cases)
  case (rule l r \<sigma>')
  with rstep_pos.intros [OF this(2), of p "s \<cdot> \<sigma>" "\<sigma>' \<circ>s \<sigma>"]
    show ?thesis by (auto simp: ctxt_of_pos_term_subst)
qed

lemma rstep_pos_rule:
  assumes "(l, r) \<in> R"
  shows "(l, r) \<in> rstep_pos R \<epsilon>"
  using rstep_pos.intros [OF assms, of \<epsilon> l Var] by simp

lemma rstep_pos_rstep_r_p_s_conv:
  "rstep_pos R p = {(s, t) | s t r \<sigma>. (s, t) \<in> rstep_r_p_s R r p \<sigma>}"
  by (auto simp: rstep_r_p_s_def Let_def subt_at_ctxt_of_pos_term
           intro: replace_at_ident
           elim!: rstep_pos.cases)

lemma rstep_rstep_pos_conv:
  "rstep R = {(s, t) | s t p. (s, t) \<in> rstep_pos R p}"
  by (force simp: rstep_pos_rstep_r_p_s_conv rstep_iff_rstep_r_p_s)

lemma rstep_pos_supt:
  assumes "(s, t) \<in> rstep_pos R p"
    and q: "q \<in> poss u" and u: "u |_ q = s"
  shows "(u, (ctxt_of_pos_term q u)\<langle>t\<rangle>) \<in> rstep_pos R (q <#> p)"
using assms
proof (cases)
  case (rule l r \<sigma>)
  with q and u have "(q <#> p) \<in> poss u" and "u |_ (q <#> p) = l \<cdot> \<sigma>" by auto
  with rstep_pos.rule [OF rule(2) this] show ?thesis
    unfolding rule by (auto simp: ctxt_of_pos_term_append u)
qed

lemma rrstep_rstep_pos_conv:
  "rrstep R = rstep_pos R \<epsilon>"
  by (auto simp: rrstep_def rstep_pos_rstep_r_p_s_conv)

lemma rrstep_imp_rstep:
  assumes "(s, t) \<in> rrstep R"
  shows "(s, t) \<in> rstep R"
  using assms by (auto simp: rrstep_def rstep_iff_rstep_r_p_s)

lemma not_NF_rstep_imp_subteq_not_NF_rrstep:
  assumes "s \<notin> NF (rstep R)"
  shows "\<exists>t \<unlhd> s. t \<notin> NF (rrstep R)"
proof -
  from assms obtain u where "(s, u) \<in> rstep R" by auto
  then obtain l r C \<sigma> where "(l, r) \<in> R" and s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and u: "u = C\<langle>r \<cdot> \<sigma>\<rangle>" by auto
  then have "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> rrstep R" and "l \<cdot> \<sigma> \<unlhd> s" by auto
  then show ?thesis by blast
qed

lemma all_subt_NF_rrstep_iff_all_subt_NF_rstep:
  "(\<forall>s \<lhd> t. s \<in> NF (rrstep R)) \<longleftrightarrow> (\<forall>s \<lhd> t. s \<in> NF (rstep R))"
  by (auto dest: rrstep_imp_rstep supt_supteq_trans not_NF_rstep_imp_subteq_not_NF_rrstep)

lemma not_in_poss_imp_NF_rstep_pos [simp]:
  assumes "p \<notin> poss s"
  shows "s \<in> NF (rstep_pos R p)"
  using assms by (auto simp: NF_def elim: rstep_pos.cases)

lemma Var_rstep_imp_rstep_pos_Empty:
  assumes "(Var x, t) \<in> rstep R"
  shows "(Var x, t) \<in> rstep_pos R \<epsilon>"
  using assms by (metis Var_supt nrrstep_subt rrstep_rstep_pos_conv rstep_cases)

lemma rstep_args_NF_imp_rrstep:
  assumes "(s, t) \<in> rstep R"
    and "\<forall>u \<lhd> s. u \<in> NF (rstep R)"
  shows "(s, t) \<in> rrstep R"
  using assms by (metis NF_iff_no_step nrrstep_subt rstep_cases)

lemma rstep_pos_imp_rstep_pos_Empty:
  assumes "(s, t) \<in> rstep_pos R p"
  shows "(s |_ p, t |_ p) \<in> rstep_pos R \<epsilon>"
  using assms by (cases) (auto simp: replace_at_subt_at intro: rstep_pos_rule rstep_pos_subst)

lemma rstep_pos_arg:
  assumes "(s, t) \<in> rstep_pos R p"
    and "i < length ss" and "ss ! i = s" 
  shows "(Fun f ss, (ctxt_of_pos_term <i> (Fun f ss))\<langle>t\<rangle>) \<in> rstep_pos R (i <# p)"
using assms
apply (cases)
apply (auto)
by (metis args.simps(2) ctxt_apply.simps(2) ctxt_of_pos_term.simps(2) poss_PCons_poss rstep_pos.intros subt_at.simps(2))

lemma rstep_imp_max_pos:
  assumes "(s, t) \<in> rstep R"
  shows "\<exists>u. \<exists>p\<in>poss s. (s, u) \<in> rstep_pos R p \<and> (\<forall>v \<lhd> s |_ p. v \<in> NF (rstep R))"
using assms
proof (induction s arbitrary: t)
  case (Var x)
  from Var_rstep_imp_rstep_pos_Empty [OF this] show ?case by auto
next
  case (Fun f ss)
  show ?case
  proof (cases "\<forall>v \<lhd> Fun f ss |_ \<epsilon>. v \<in> NF (rstep R)")
    case True
    moreover with Fun.prems
      have "(Fun f ss, t) \<in> rstep_pos R \<epsilon>"
      by (auto dest: rstep_args_NF_imp_rrstep simp: rrstep_rstep_pos_conv)
    ultimately show ?thesis by auto
  next
    case False
    then obtain v where "v \<lhd> Fun f ss" and "v \<notin> NF (rstep R)" by auto
    then obtain s and w where "s \<in> set ss" and "s \<unrhd> v" and "(s, w) \<in> rstep R"
      by (auto simp: NF_def) (metis NF_iff_no_step NF_subterm supt_Fun_imp_arg_supteq)
    from Fun.IH [OF this(1, 3)] obtain u and p
      where "p \<in> poss s" and *: "(s, u) \<in> rstep_pos R p"
      and **: "\<forall>v \<lhd> s |_ p. v \<in> NF (rstep R)" by blast
    from `s \<in> set ss` obtain i
      where "i < length ss" and [simp]:"ss ! i = s" by (auto simp: in_set_conv_nth)
    with `p \<in> poss s` have "i <# p \<in> poss (Fun f ss)" by auto
    moreover with ** have "\<forall>v \<lhd> Fun f ss |_ (i <# p). v \<in> NF (rstep R)" by auto
    moreover from rstep_pos_arg [OF * `i < length ss` `ss ! i = s`]
      have "(Fun f ss, (ctxt_of_pos_term <i> (Fun f ss))\<langle>u\<rangle>) \<in> rstep_pos R (i <# p)" .
    ultimately show ?thesis by blast
  qed
qed

lemma rhs_free_vars_imp_rstep_not_SN':
  assumes "(l, r) \<in> R" and "\<not> vars_term r \<subseteq> vars_term l"
  shows "\<not> SN (rstep R)"
  using rhs_free_vars_imp_rstep_not_SN [OF assms] by (auto simp: SN_defs)

lemma SN_imp_variable_condition:
  assumes "SN (rstep R)"
  shows "\<forall>(l, r) \<in> R. vars_term r \<subseteq> vars_term l"
  using assms and rhs_free_vars_imp_rstep_not_SN' [of _ _ R] by blast

end

